#include "CalcPaths.h"

using namespace std;


void CalcPaths::buildGraph
          ( const map< int, vector< CorrespondenceMatrix* > >& allCorrespondences,
            vector< vector< Node* > >& allNodes, 
            float sigma_dist )
{
   bool testing = false;

   cout << "initializing the nodes" << endl;
   //
   //  make lists of nodes in every slice
   //
   //vector< vector< Node* > > allNodes;
   map< int, vector< CorrespondenceMatrix* > >::const_iterator corrIter;
   for( corrIter = allCorrespondences.begin();
        corrIter != allCorrespondences.end();
        ++corrIter )
   {
      //vector< CorrespondenceMatrix > corrMatrix = corrIter->second;
      // 
      // Get the correlation for two slices
      // only need the first correlation to get the nodes in the slice 
      CorrespondenceMatrix* corrMatrix = (corrIter->second)[0];
      vector< LabelType > sliceALabels;
      vector< LabelType > sliceBLabels;
      corrMatrix->getLabels( sliceALabels, sliceBLabels );

      //
      // make all the root nodes
      //
      if( corrIter == allCorrespondences.begin() )
      {
         //cout << "blah: " << corrMatrix.size() << endl ;
         vector< Node* > nodes;
         // loop over all the correlations and just get the label
         //for( unsigned int n = 0; n < corrMatrix->size(); n++ )
         for( unsigned int n = 0; n < sliceALabels.size(); n++ )
         {
            Node* N = new Node();
            N->label = sliceALabels[n]; //corrMatrix->getLabel( n );
            N->slice = corrMatrix->getSliceA(); //corrMatrix->getSlice( n );
            //int x, y;
            //corrMatrix->get( n ).xy( x, y );
            //N->x = x; N->y = y;
            //cout << "adding " << N->label << " with " << x << "," << y << endl ;
            nodes.push_back( N );
         }
         allNodes.push_back( nodes );
         //cout << "*allNodes size: " << allNodes.size() << endl ;
      }

      // handle all other nodes
      vector< Node* > nodes;
      // get the set of correspondences for the first node/region (should be the
      // same set of labels for all the nodes)
      //const CorrespondenceList& corrList = corrMatrix->get( 0 );
      //for( unsigned int n = 0; n < corrList.size(); n++ )
      for( unsigned int n = 0; n < sliceBLabels.size(); n++ )
      {
         Node* N = new Node();
         N->label = sliceBLabels[n]; //corrList.get( n ).labelB();
         N->slice = corrMatrix->getSliceB(); //corrList.get( n ).sliceB();
         //int x, y;
         //corrList.get( n ).xyB( x, y );
         //N->x = x; N->y = y;
         //cout << "adding " << N->label << endl ;
         nodes.push_back( N );
      }
      allNodes.push_back( nodes );
      //   cout << "allNodes size: " << allNodes.size() << endl ;
   }

   cout << "connecting nodes" << endl ;
   //
   // connect up the nodes - give each node a list of child nodes and a
   // cost
   //
   // loop over each vector of correspondences for each slice
   //
   int index = 0;
   for( corrIter = allCorrespondences.begin();
        corrIter != allCorrespondences.end();
        ++corrIter )
   {
      int slice = corrIter->first;
      //
      // connect the nodes in allNodes[i] with allNodes[i+1], ..allNodes[i+l]
      // loop over the number of correltations for each slice (including
      // skipped slices;
      //
      for( unsigned int l = 1; l <= corrIter->second.size(); l++ )
      {
         unsigned int I = index; //slice;
         unsigned int J = index + l; //slice + l;
         cout << "connecting " << I << " and " << J << endl ;
         cout << "connecting " << slice << " and " << slice + l << endl ;

         // do some timing
         clock_t clock_start;
         clock_start = clock();

         // get the correspondences between the current slice and slice+l
         CorrespondenceMatrix* corrMatrix = (corrIter->second)[l-1];

         for( unsigned int i = 0; i < allNodes[I].size(); i++ )
         {
            Node* node_i = allNodes[I][i];
            for( unsigned int j = 0; j < allNodes[J].size(); j++ )
            {
               Node* node_j = allNodes[J][j];

               const Correspondence* C = corrMatrix->getCorrespondence
                            ( node_i->label, node_i->slice, 
                              node_j->label, node_j->slice );

               // if we have a correlation for this node...
               //if( C->exits() )
               if( C != NULL )
               {
                  // lets not add in nodes that are larger than a certain
                  // distance
                  if( C->dist() < sigma_dist )
                  {
                     node_i->childrenNodes.push_back( node_j );
                     //node_j->parentNodes.push_back( node_i );

                     // set the xy values for each node;
                     int x, y;
                     C->xyA( x, y ); node_i->x = x; node_i->y = y;
                     C->xyB( x, y ); node_j->x = x; node_j->y = y;

                     if( testing )
                        node_i->childrenCosts.push_back( C->compute() );
                     else
                     {
                        double cost = C->compute( sigma_dist, J-I );
                        //cout << "I,J,i,j: " << I << "," << J << "," << i << "," << j 
                        //     << ": " << cost << endl ;
                        node_i->childrenCosts.push_back( cost );
                        //node_j->parentCosts.push_back( cost );
                     }
                  }
               }
            }
         }

         // more timing
         int secs = (clock() - clock_start) / (double)CLOCKS_PER_SEC;
         int mins = secs / 60;
         printf("time: %.i seconds, or %i minutes \n", secs, mins );

      }
      index++;
   }

   //
   // print the nodes
   //
   /*
   cout << "Node vectors: " << endl ;
   for( unsigned int i = 0; i < allNodes.size(); i++ )
   {
      for( unsigned int j = 0; j < allNodes[i].size(); j++ )
      {
         cout << allNodes[i][j]->label << " ( ";
         for( unsigned int c = 0; c < allNodes[i][j]->childrenNodes.size(); c++ )
            cout << allNodes[i][j]->childrenNodes[c]->label << " ";
         cout << ") ";

      }
      cout << endl ;
   }
   */
}


void CalcPaths::calcPaths( vector< vector< Node* > >& allNodes,
                          map< float, Path >& allPaths )
{
   for( unsigned int n = 0; n < allNodes[0].size(); n++ )
   {
      if( allNodes[0][n] != NULL )
      {
         //cout << "num children from root: " << allNodes[0][n]->label
         //   << ", " << allNodes[0][n]->slice << " " <<
         //   allNodes[0][n]->childrenNodes.size() << endl ;
         //
         //  compute the shortest path
         //
         Dijkstra dijkstra( allNodes[0][n] );
         //dijkstra.calculateLeastCostPath();

         Path P = extractPath( allNodes );
         // need to check if dijkstra made it to the end of the
         // graph
         if( P.size() > 1 )
         {
            allPaths[ P.getCost() ] = P; 
            //cout << "cost: " << P.getCost() << endl ;
         }

         dijkstra.resetCosts( allNodes[0][n] );
      }
   }
}

void CalcPaths::calcPathsMod( vector< vector< Node* > >& allNodes,
                              map< float, Path >& allPaths,
                              vector< int > boundingBox,
                              ListOfPaths& initPaths )
{
   // only do this once...
   if ( allNodes[0].size() > 1 )
   {
      cout << "adding main ROOT node" << endl ;
      // make top root node to point to all the other nodes
      vector< Node* > rootVector;
      Node* ROOT = new Node();
      ROOT->slice = -1;
      ROOT->slice = -1;
      ROOT->x = -1; ROOT->y = -1;
      for( unsigned int n = 0; n < allNodes[0].size(); n++ )
      {
         if( allNodes[0][n] != NULL )  // some root nodes may be null
                                       // because they were removed from the graph earlier
         {
            if( boundingBox.size() > 0 )  // add only nodes within this bounding box
            {
               if( allNodes[0][n]->withinBoundingBox( boundingBox ) )
               {
                  cout << "adding node (label,slice,x,y)" 
                       << allNodes[0][n]->label << ", " 
                       << allNodes[0][n]->slice << ", "
                       << allNodes[0][n]->x << ", "
                       << allNodes[0][n]->y << endl;
                  ROOT->childrenNodes.push_back( allNodes[0][n] );
                  ROOT->childrenCosts.push_back( 0.0 );
               }
               
            }
            else if( initPaths.size() > 0 )  // add only nodes from previous paths
            {
               /*
               for( unsigned int z = 0; z < initPaths.size(); z++ )
               {
                  Path* P = initPaths.getPathByIndex( z );
                  if( ( P->getX( slice ) == allNodes[0][n]->x ) &&
                      ( P->getY( slice ) == allNodex[0][n]->y ) )
                  {
                     ROOT->childrenNodes.push_back( allNodes[0][n] );
                     ROOT->childrenCosts.push_back( 0.0 );
                  }

               }
               */
            }
            else  // just add all the nodes
            {
               ROOT->childrenNodes.push_back( allNodes[0][n] );
               ROOT->childrenCosts.push_back( 0.0 );
            }
         }
      }
      rootVector.push_back( ROOT );
      allNodes.insert( allNodes.begin(), rootVector );
   }

   //
   //  compute the shortest path
   //
   Dijkstra dijkstra( allNodes[0][0] );  // always point to the ROOT node
   dijkstra.calculateLeastCostPath();

   Path P = extractPath( allNodes );
   // need to check if dijkstra made it to the end of the
   // graph
   if( P.size() > 1 )
   {
      allPaths[ P.getCost() ] = P; 
      //cout << "cost: " << P.getCost() << endl ;
   }

   dijkstra.resetCosts( allNodes[0][0] );
}

void CalcPaths::removePath( const Path& P,  
                 vector< vector< Node* > >& allNodes )
{
  for( unsigned int i = 0; i < P.size(); i++ )
  {
     int slice = P.getSlice( i );
     int label = P.getLabel( i );
     //cout << "removing (slice,label): " << slice << "," << label << endl;

     //
     // remove the node's children and costs:
     // inefficiently loop over all the nodes in the graph and find each
     // node and remove it's pointer to the node
     //
     for( unsigned int m = 0; m < allNodes.size(); m++ )
     {
        for( unsigned int n = 0; n < allNodes[m].size(); n++ )
        {
           Node* N = allNodes[m][n];
           if( N != NULL )
           {
              //Node** N = &allNodes[m][n];
              //Node*& N = allNodes[m][n];
              bool done = N->removeChild( slice, label );

              if( (N->slice == slice) && (N->label == label ) )
              {
                 //cout << "setting " << allNodes[m][n]<< " to NULL" << endl ;
                 allNodes[m][n] = NULL;  // remove this node
              }
           }
        }
     }


  }
     
}

Path CalcPaths::extractPath( vector< vector< Node* > >& allNodes )
{
   // get the last row of nodes
   vector< Node* > lastRow = allNodes[ allNodes.size() - 1 ];

   // get the cheapest node
   Node* bestNode = NULL;
   float bestCost = numeric_limits< float >::max();
   for( unsigned int i = 0; i < lastRow.size(); i++ )
   {
      if(lastRow[i] != NULL )
      {
         //cout << i << ": " ; lastRow[i]->print();
         if (lastRow[i]->minCost < bestCost)
         {
            bestNode = lastRow[i];
            bestCost = bestNode->minCost;
         }
      }
   }

   Path P;
   if( bestNode != NULL )
   {
      //cout << "best node: " << bestNode->label << ", " << bestNode->minCost << endl;

      // follow that node back to the root
      Node* N = bestNode;
      //cout << "Path: (slice, label, cost)" << endl;
      while( N != NULL ) 
      {
         //cout << N->slice << ", " << N->label << ", " << N->minCost 
         //   << "(" << N->x << "," << N->y << ")" << endl;
         if( N == bestNode ) // need first node to init path
            P.initPath( N->slice, N->label, N->minCost, N->minCost, N->x, N->y );
         else
         {
            if( N->slice > -1 )
               P.addEnd( N->slice, N->label, N->minCost, N->minCost, N->x, N->y );
         }
         N = N->prevNode;
      }
      P.setCost( bestCost );  // need to do this at the end

      //cout << "Path: " ; P.print();
   }

   return P;
}

