#ifndef CMDLINEOPTS_H
#define CMDLINEOPTS_H

#include <vector>
#include <string>

#include "tclap/CmdLine.h"

using namespace std;

class CmdLineOpts 
{
   private:

      TCLAP::CmdLine* p_cmd;

      vector< string > p_rawFiles;
      vector< string > p_segFiles;
      vector< string > p_segEditFiles;
      vector< string > p_waterFiles;
      vector< string > p_matrixFiles;
      vector< string > p_polyDataFiles;
      vector< string > p_corFiles;
      string p_vtkPath;
      vector< string > p_pathFile;
      string p_mergeFile;
      int p_window;
      int p_overlap;
      vector< float > p_spacing ;
      bool p_rebuildPaths;
      int p_downSample;

   public:

      CmdLineOpts( int argc, char** argv )
      {
         p_cmd = new TCLAP::CmdLine("cmd line message", ' ', "1.0");

         processOptions( argc, argv );
      };

      ~CmdLineOpts()
      {
         delete p_cmd;
      }

      vector< string > getRawFileNames() { return p_rawFiles; };

      vector< string > getSegFileNames() { return p_segFiles; };

      vector< string > getSegEditFileNames() { return p_segEditFiles; };

      vector< string > getWaterFileNames() { return p_waterFiles; };

      vector< string > getMatrixFileNames() { return p_matrixFiles; };

      vector< string > getPolyDataFileNames() { return p_polyDataFiles; };

      vector< string > getCorrelationDataFileNames() { return p_corFiles; };

      string getVTKPath() { return p_vtkPath; };

      vector< string > getPathFileName() { return p_pathFile; };

      string getMergeFileName() { return p_mergeFile; };

      int getWindow() { return p_window; } ;
      int getOverlap() { return p_overlap; } ;

      vector< float > getSpacing() { return p_spacing; };

      bool getRebuildFlag() { return p_rebuildPaths; };

      int getDownSampleValue() { return p_downSample; };


   private:

      void processOptions( int argc, char** argv )
      {
         try{

            TCLAP::MultiArg< string > rawFileParam("r", "raw_filenames", "raw data files",
                  true,  "string");
            p_cmd->add( rawFileParam );

            TCLAP::MultiArg< string > segFileParam("s", "seg_filename",  "segmentation data file",
                  true,  "string");
            p_cmd->add( segFileParam );

            TCLAP::MultiArg< string > segEditFileParam("e", "seg_edit_filename",  
                  "segmentation data file", false,  "string");
            p_cmd->add( segEditFileParam );

            TCLAP::MultiArg< string > waterFileParam("w", "watershed_filename",  "watershed data file",
                  false,  "string");
            p_cmd->add( waterFileParam );

            TCLAP::MultiArg< string > corFileParam("c", "correlation_filename",  "correlation data file",
                  false,  "string");
            p_cmd->add( corFileParam );

            TCLAP::ValueArg< string > mergeFileParam("m",
                  "merge_region_filename",  "file of regions to merge",
                  false,  "", "string");
            p_cmd->add( mergeFileParam );

            TCLAP::UnlabeledMultiArg< string > matrixFileParam
               ("Matrices_and_Labels", "input for finding paths", false,
                "all the matrix and label files", true);
            p_cmd->add( matrixFileParam );

            TCLAP::ValueArg< string > vtkFileParam("i",
                  "vtk_path",  "vtk isosurface file",
                  true,  "", "string");
            p_cmd->add( vtkFileParam );

            TCLAP::MultiArg< string > pathFileParam("p",
                  "path_filename",  "path file",
                  false, "string");
            p_cmd->add( pathFileParam );

            TCLAP::MultiArg< string > polyDataParam("v",
                  "vtk_filename",  "vtk poly data file",
                  false, "string");
            p_cmd->add( polyDataParam );

            TCLAP::ValueArg< int > windowFileParam("b", "window_buffer",
                  "window of slices we view/process at a time",
                  false,  10, "int");
            p_cmd->add( windowFileParam );

            TCLAP::ValueArg< int > downSampleParam("d", "ds",
                  "downasample value for rendering",
                  false,  2, "int");
            p_cmd->add( downSampleParam );

            TCLAP::ValueArg< int > overlapFileParam("o", "window_overlap",  "overlap of window",
                  false,  1, "int");
            p_cmd->add( overlapFileParam );

            TCLAP::MultiArg< float > spacingParam("k", "image_spacing", "voxel spacing",
                  false,  "float");
            p_cmd->add( spacingParam );

            TCLAP::SwitchArg rebuildPathsParam("x", "rebuild_paths",
                  "rebuild isosurfaces from paths.xml", false );
            p_cmd->add( rebuildPathsParam );

            p_cmd->parse( argc, argv );

            p_rawFiles = rawFileParam.getValue();
            p_segFiles = segFileParam.getValue();
            p_segEditFiles = segEditFileParam.getValue();
            p_waterFiles = waterFileParam.getValue();
            p_corFiles = corFileParam.getValue();
            p_mergeFile = mergeFileParam.getValue();
            p_matrixFiles = matrixFileParam.getValue();
            p_vtkPath = vtkFileParam.getValue();
            p_pathFile = pathFileParam.getValue();
            p_polyDataFiles = polyDataParam.getValue();
            p_window = windowFileParam.getValue();
            p_overlap = overlapFileParam.getValue();
            p_spacing = spacingParam.getValue();
            p_rebuildPaths = rebuildPathsParam.getValue();
            p_downSample = downSampleParam.getValue();

         }catch(TCLAP::ArgException &e){
            cerr << "Error: " << e.error() << " for arg " << e.argId() << endl; 
            exit(1);
         }
      }


     
};

#endif 

