#include "ItkVtkConvert.h"

#include "itkImageFileWriter.h"
#include "itkImageRegionIteratorWithIndex.h"
#include "itkRegion.h"

#include "vtkMetaImageWriter.h"

//#define DEBUG 1

typedef itk::ImageRegionIteratorWithIndex< ImageType > ImageIteratorWithIndex;
typedef itk::ImageRegionIteratorWithIndex< ImageType3D > Image3DIteratorWithIndex;

ImageType::Pointer ItkVtkConvert::ConvertToItkImage(vtkImageData *vtkImage)
{
  ImageType::Pointer itkImg = ImageType::New();
  ImageType::IndexType start;
  ImageType::SizeType size;
  ImageType::SpacingType spacing;
  ImageType::PointType origin;

  // get the vtkImageData extent
  int extent[4];
  vtkImage->GetExtent( extent );
  //cout << "get extent: " ;
  //for(int i = 0; i < 4; i++ )
  //   cout << extent[i] << " ";
  //cout << endl ;

  start[0] = extent[0];
  start[1] = extent[2];
  size[0] = extent[1]-extent[0]+1;
  size[1] = extent[3]-extent[2]+1;

  double doubleSpacing[2];
  double doubleOrigin[2];
  vtkImage->GetSpacing(doubleSpacing);
  vtkImage->GetOrigin(doubleOrigin);
  for(int i=0;i<2;i++){
    spacing[i] = doubleSpacing[i];
    origin[i] = doubleOrigin[i];
  }
  origin[0] = origin[1] = 0;
  spacing[0] = spacing[1] = 1;
  // allocate the itk::Image
  ImageType::RegionType region;
  region.SetIndex(start);
  region.SetSize(size);
  itkImg->SetRegions(region);
  //itkImg->SetBufferedRegion( region );
  itkImg->SetOrigin(origin);
  itkImg->SetSpacing(spacing);
  itkImg->Allocate();

  // copy the data
  ImageIteratorWithIndex it(itkImg, itkImg->GetRequestedRegion());
  for(it.GoToBegin(); !it.IsAtEnd(); ++it){
    ImageType::IndexType curIdx = it.GetIndex();
    it.Set(vtkImage->GetScalarComponentAsFloat(curIdx[0], curIdx[1], 0, 0));
  }
  itkImg->Update();

#ifdef DEBUG
   typedef itk::ImageFileWriter<ImageType> ImageWriterType;
   ImageWriterType::Pointer sliceWriter = ImageWriterType::New();
   sliceWriter->SetInput(itkImg);
   sliceWriter->SetFileName("ConvertToITKTest.mha");
   sliceWriter->Write();
#endif

   itkImg->Register();
  
  return itkImg;

}

ImageType3D::Pointer ItkVtkConvert::ConvertToItkImage3D( vtkImageData *vtkImage )
{
  ImageType3D::Pointer itkImg = ImageType3D::New();
  ImageType3D::IndexType start;
  ImageType3D::SizeType size;
  ImageType3D::SpacingType spacing;
  ImageType3D::PointType origin;

  // get the vtkImageData extent
  int extent[6];
  vtkImage->GetExtent( extent );
  for( int i = 0; i < 6; i++ )
     cout << extent[i] << " ";
  cout << endl;
  start[0] = extent[0];
  start[1] = extent[2];
  start[2] = extent[4];
  size[0] = extent[1]-extent[0]+1;
  size[1] = extent[3]-extent[2]+1;
  size[2] = extent[5]-extent[4]+1;


  double doubleSpacing[3];
  double doubleOrigin[3];
  vtkImage->GetSpacing(doubleSpacing);
  vtkImage->GetOrigin(doubleOrigin);
  for(int i=0;i<3;i++){
    spacing[i] = doubleSpacing[i];
    origin[i] = doubleOrigin[i];
  }
  // allocate the itk::Image
  ImageType3D::RegionType region;
  region.SetIndex(start);
  region.SetSize(size);
  itkImg->SetRegions(region);
  itkImg->SetOrigin(origin);
  itkImg->SetSpacing(spacing);
  itkImg->Allocate();

  // copy the data
  Image3DIteratorWithIndex it(itkImg, itkImg->GetRequestedRegion());
  for(it.GoToBegin(); !it.IsAtEnd(); ++it){
    ImageType3D::IndexType curIdx = it.GetIndex();
    //cout << curIdx << endl;
    it.Set(vtkImage->GetScalarComponentAsFloat(curIdx[0], curIdx[1],
             curIdx[2], 0 ));
             //extent[4], 0));
  }

#ifdef DEBUG
   typedef itk::ImageFileWriter<ImageType3D> ImageWriterType;
   ImageWriterType::Pointer sliceWriter = ImageWriterType::New();
   sliceWriter->SetInput(itkImg);
   sliceWriter->SetFileName("ConvertToITKTest3D.mha");
   sliceWriter->Write();
#endif

   itkImg->Register();
  
  return itkImg;
  
   /*
   int dims[6];
   vtkImage->GetExtent( dims );

   int mins = dims[4];
   int maxs = dims[5];

   int region[6];
   region[0] = dims[0];
   region[1] = dims[1];
   region[2] = dims[2];
   region[3] = dims[3];
   region[4] = region[5] = 0; // default for now

   vtkExtractVOI* extractSlice = vtkExtractVOI::New();
   extractSlice->SetInput( vtkImage );

   //int labelsum = 0;

   ImageType3D::Pointer itkImageData = 
      new3d< ImageType3D >( dims[1], dims[3], dims[5] );

   for( int s = mins; s < maxs; s++ )
   {
      region[4] = region[5] = s;
      extractSlice->SetVOI( region );
      extractSlice->Update();

      cout << "converting" << endl ;

      itkImageType::Pointer itkSlice = ConvertToItkImage( extractSlice->GetOutput());
   }

   extractSlice->Delete();
   */
}

vtkImageData* ItkVtkConvert::ConvertToVtkImage(ImageType::Pointer itkImage)
{
  ImageType::SpacingType spacing = itkImage->GetSpacing();
  ImageType::PointType origin = itkImage->GetOrigin();
  ImageType::IndexType start = itkImage->GetRequestedRegion().GetIndex();
  ImageType::SizeType size = itkImage->GetRequestedRegion().GetSize();

  vtkImageData *vtkImage = vtkImageData::New();

  // get the vtkImageData extent
  int extent[6];
  extent[0] = start[0];
  extent[1] = start[0]+size[0]-1;
  extent[2] = start[1];
  extent[3] = start[1]+size[1]-1;
  extent[4] = 0;
  extent[5] = 0;
  vtkImage->SetExtent(extent);

  double doubleSpacing[3];
  double doubleOrigin[3];
  for(int i=0;i<2;i++){
    doubleSpacing[i] = spacing[i];
    doubleOrigin[i] = origin[i];
  }
  vtkImage->SetSpacing(doubleSpacing);
  vtkImage->SetOrigin(doubleOrigin);

  // allocate the vtkImage scalars
  vtkImage->SetScalarTypeToFloat();
  vtkImage->SetNumberOfScalarComponents(1);
  vtkImage->AllocateScalars();

#ifdef DEBUG
  std::cout << "Empty VTK Image: ----------------" << std::endl;
  vtkImage->PrintSelf(cout, vtkIndent());
#endif

  // copy the data
  ImageIteratorWithIndex it(itkImage, itkImage->GetRequestedRegion());
  for(it.GoToBegin(); !it.IsAtEnd(); ++it){
    ImageType::IndexType curIdx = it.GetIndex();
    vtkImage->SetScalarComponentFromFloat(curIdx[0], curIdx[1], 0, 0, it.Get());
  }

#ifdef DEBUG
  vtkMetaImageWriter *vtkWriter = vtkMetaImageWriter::New();
  vtkWriter->SetInput(vtkImage);
  vtkWriter->SetFileName("ConvertToVTKTest.mha");
  vtkWriter->Write();
#endif
  
  return vtkImage;

}

