#ifndef METAREADER_H
#define METAREADER_H

#include <string>
#include "itkImage.h"
#include "itkImageFileReader.h"

typedef itk::Image< double, 3 > Image;
typedef itk::ImageFileReader< Image > ImageReader;

//  
//  Read only the meta data associated with an mha/mhd image file
//
class MetaReader 
{
   private:

      unsigned int p_dims;
      int* p_size;
      int* p_origin;
      int* p_spacing;


   public:

      MetaReader() { p_dims = 0 ; };
         
      ~MetaReader()
      {
         delete[] p_size;
         delete[] p_origin;
         delete[] p_spacing;
      }

      void setData( std::string filename )
      {
         ImageReader::Pointer reader = ImageReader::New();
         reader->SetFileName( filename.c_str() );
         reader->UpdateOutputInformation();
         
         p_dims =
            reader->GetImageIO()->GetNumberOfDimensions();
         p_size = new int( p_dims );
         p_origin = new int[ p_dims ];
         p_spacing = new int[ p_dims ];
         for( unsigned int d = 0; d < p_dims; d++ )
         {
            p_size[ d ] = reader->GetImageIO()->GetDimensions( d );
            p_origin[ d ] = reader->GetImageIO()->GetOrigin( d );
            p_spacing[ d ] = reader->GetImageIO()->GetSpacing( d );
         }

      };

      unsigned int getDims() { return p_dims; } ;

      void getSize( int* size )
      {
         for( unsigned int d = 0; d < p_dims; d++ )
         {
            size[ d ] = p_size[ d ];
         }
      };

      void getOrigin( int* origin )
      {
         for( unsigned int d = 0; d < p_dims; d++ )
            origin[ d ] = p_origin[ d ];
      };

      void getSpacing( int* spacing )
      {
         for( unsigned int d = 0; d < p_dims; d++ )
            spacing[ d ] = p_spacing[ d ];
      };

     
};

#endif 

