
#include <qapplication.h>
#include <qfiledialog.h>
#include "qmenubar.h"
#include <QString>
#include <QList>


#include "ReconstructionViewer.h"
#include <vtkActor.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>
#include "vtkCylinderSource.h"
#include <vtkPolyDataMapper.h>
#include <vtkMetaImageReader.h>
#include "vtkExtractVOI.h"
#include "vtkTransform.h"
#include "vtkTexture.h"
#include "vtkPlaneSource.h"
#include "vtkTransformPolyDataFilter.h"
#include "vtkLight.h"
#include "vtkCellArray.h"
#include "vtkProperty.h"
#include "vtkImageData.h"
#include "vtkImageActor.h"
#include "vtkImageImport.h"
#include "vtkImageExport.h"
#include "vtkImageActor.h"
#include "vtkCallbackCommand.h"
#include "vtkWorldPointPicker.h"

void pickSliceCallback(vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);
   if( interactor->GetRepeatCount() > 0 )
   {
      //cout << "count: " << interactor->GetRepeatCount() << endl ;
      //cout << "pick slice callback called" << endl ;
      ReconstructionViewer* re = static_cast< ReconstructionViewer* >( clientdata );

      if( interactor )
      {
         int mouseX, mouseY;
         interactor->GetEventPosition( mouseX, mouseY );
         //cout << "mouse: " << mouseX << ", " << mouseY << endl;

         vtkRenderer* rend = interactor->FindPokedRenderer( mouseX, mouseY );

         vtkWorldPointPicker* picker = vtkWorldPointPicker::New();
         picker->Pick( mouseX, mouseY, 0, rend );
         double pickCoords[3];
         picker->GetPickPosition( pickCoords);
         //cout << "picked: " ;
         //for( int p = 0; p < 3; p++ )
         //   cout << pickCoords[p] << " " ;
         //cout << endl;

         re->goToSection( pickCoords[2] );
         
         picker->Delete(); 
      }
   }

}

// Constructor
ReconstructionViewer::ReconstructionViewer(QWidget* p) 
 : QMainWindow(p)
{
  setupUi(this);

  //
  //  File menu
  //
  QAction* a_fileOpen = new QAction(tr("&Open"), this);
  a_fileOpen->setShortcut(tr("Ctrl+N"));
  a_fileOpen->setStatusTip(tr("Create a new file"));
  connect(a_fileOpen, SIGNAL(triggered()), this, SLOT(fileOpen()));

  QAction* a_mergeOpen = new QAction(tr("&Import Merged Regions"), this);
  connect(a_mergeOpen, SIGNAL(triggered()), this, SLOT(openMergeRegions()));

  QAction* a_fileExit = new QAction(tr("&Exit"), this);
  a_fileExit->setShortcut(tr("Ctrl+Q"));
  a_fileExit->setStatusTip(tr("Exit"));
  connect(a_fileExit, SIGNAL(triggered()), this, SLOT(fileExit()));

  QMenu* file_menu = this->menuBar()->addMenu(tr("&File"));
  file_menu->addAction(a_fileOpen);
  file_menu->addAction(a_mergeOpen);
  file_menu->addAction(a_fileExit);

  //
  //  Draw menu
  //
  QAction* a_drawOption = new QAction(tr("&Draw"), this);
  a_drawOption->setStatusTip(tr("Draw on the image"));
  a_drawOption->setCheckable(true);
  connect(a_drawOption, SIGNAL(triggered()), this, SLOT(drawTrig()));

  QAction* a_eraseOption = new QAction(tr("&Erase"), this);
  a_eraseOption->setStatusTip(tr("Erase from the image"));
  a_eraseOption->setCheckable(true);
  connect(a_eraseOption, SIGNAL(triggered()), this, SLOT(eraseTrig()));

  QActionGroup* radioGroup = new QActionGroup( this );
  radioGroup->addAction( a_drawOption);
  radioGroup->addAction( a_eraseOption);
  a_drawOption->setChecked(true);

  QMenu* draw_menu = this->menuBar()->addMenu(tr("&Draw"));
  draw_menu->addAction(a_drawOption);
  draw_menu->addAction(a_eraseOption);

  //
  //  section scroll bar 
  //
  connect( sectionScrollBar, SIGNAL( valueChanged(int) ), this, SLOT( sectionChanged() ));
  connect( sectionScrollBar, SIGNAL( valueChanged(int) ), scrollLabel, SLOT( setNum(int) ));
  sectionScrollBar->setFocusPolicy( Qt::StrongFocus );

  QShortcut* sectionShortcutPlus = new QShortcut( QKeySequence( "." ), sectionScrollBar ) ;
  sectionShortcutPlus->setContext( Qt::ApplicationShortcut );
  connect( sectionShortcutPlus, SIGNAL( activated() ), this, SLOT( sectionChangedPlus() ));
  QShortcut* sectionShortcutMinus = new QShortcut( QKeySequence( "," ), sectionScrollBar ) ;
  sectionShortcutMinus->setContext( Qt::ApplicationShortcut );
  connect( sectionShortcutMinus, SIGNAL( activated() ), this, SLOT( sectionChangedMinus() ));

  //
  //  show image radio button
  //
  connect( showImageButton, SIGNAL( clicked() ), this, SLOT( showRawImage() ) );
  QShortcut* showImageShortcut = new QShortcut( QKeySequence( "a" ), showImageButton );
  showImageShortcut->setContext( Qt::ApplicationShortcut );
  connect( showImageShortcut, SIGNAL( activated() ), this, SLOT( showRawImageShortcut() ));

  //
  //  show membrane detection radio button
  //
  connect( showSegmentationButton, SIGNAL( clicked() ), this, SLOT( showSegImage() ) );
  QShortcut* showSegmentationShortcut = new QShortcut( QKeySequence( "s"), showSegmentationButton );
  showSegmentationShortcut->setContext( Qt::ApplicationShortcut );
  connect( showSegmentationShortcut, SIGNAL( activated() ), this, SLOT( showSegImageShortcut() ));

  //
  //  show segmentation radio button
  //
  connect( showWatershedButton, SIGNAL( clicked() ), this, SLOT( showWaterImage() ) );
  QShortcut* showWatershedShortcut = new QShortcut( QKeySequence( "w"), showWatershedButton );
  showWatershedShortcut->setContext( Qt::ApplicationShortcut );
  connect( showWatershedShortcut, SIGNAL( activated() ), this, SLOT( showWaterImageShortcut() ));

  //
  //  picking callback
  //
  connect( manualTrackButton, SIGNAL( clicked() ), this, SLOT( setPicking() ) );

  //
  //  render the path
  //
  connect( renderButton, SIGNAL( clicked() ), this, SLOT( render() ) );

  vtkWidget->GetRenderWindow()->
    AddRenderer( p_sectionEditor.getRenderer() );

  vtk3DWidget->GetRenderWindow()->
     AddRenderer( p_sectionEditor.get3DRenderer() );

  p_sectionEditor.addPickCallback( vtkWidget->GetInteractor() );

  vtkCallbackCommand *pickSliceCommand = vtkCallbackCommand::New();
  pickSliceCommand->SetCallback( pickSliceCallback );
  pickSliceCommand->SetClientData( (void*)this );
  vtkRenderWindowInteractor* waterInteractor =
     vtk3DWidget->GetInteractor();
  waterInteractor->AddObserver( vtkCommand::LeftButtonPressEvent, pickSliceCommand );

  //
  //  Watershed callbacks
  //
  thresholdText->setText( p_sectionEditor.getThresholdStr().c_str() );
  blurText->setText( p_sectionEditor.getBlurSigmaStr().c_str() );
  levelText->setText( p_sectionEditor.getWatershedLevelStr().c_str() );
  watershedThresholdText->setText( p_sectionEditor.getWatershedThresholdStr().c_str() );
  connect( applyToSliceButton, SIGNAL( clicked() ), this, SLOT( applyWatershedToSlice( ) ) );
  connect( applyToVolumeButton, SIGNAL( clicked() ), this, SLOT( applyWatershedToVolume() ) );

  //
  //  tracking buttons
  //
  connect( mergeRegionsButton, SIGNAL( clicked() ), this, SLOT( mergeRegions() ) );
  connect( correlateRegionsButton, SIGNAL( clicked() ), this, SLOT( correlateRegions() ) );
  connect( computePathsButton, SIGNAL( clicked() ), this, SLOT( computePaths() ) );
  connect( printPathsButton, SIGNAL( clicked () ), this, SLOT( printPaths() ) );
  connect( viewPathsButton, SIGNAL( clicked () ), this, SLOT( viewPathData() ) );

  //
  //  add table for paths
  //
  /*
  p_listBox = new QListBox( this );
  
  QHBox *hb = new QHBox( this );
  //p_lineEdit = new QLineEdit( hb );
  QPushButton *pbAdd = new QPushButton( "Add", hb );
  QPushButton *pbClear = new QPushButton( "Clear", hb );

  connect( pbAdd, SIGNAL(clicked()), this, SLOT(addItem()) );
  connect( pbClear, SIGNAL(clicked()), p_listBox, SLOT(clear()) );
  p_listBox->insertItem( "hello" );
  //m_lineEdit->setText( "" );
  */
  p_rebuildPaths = false;

  p_pathDialog = new PathDialogGUI();
  p_pathDialog->init();
  connect( p_pathDialog->newButton, SIGNAL( clicked() ), 
           this, SLOT( newPathButtonCB() ) );
  connect( p_pathDialog->pathWidget, SIGNAL( itemSelectionChanged() ), 
           this, SLOT( itemSelectionChangedCB() ) );
  connect( p_pathDialog->deleteButton, SIGNAL( clicked() ), 
           this, SLOT( deletePathButtonCB() ) );
  connect( p_pathDialog->editCheckBox, SIGNAL( clicked() ), 
           this, SLOT( editPathButtonCB() ) );
  connect( p_pathDialog->autoFillButton, SIGNAL( clicked() ), 
           this, SLOT( autoFillButtonCB() ) );
  connect( p_pathDialog->nextButton, SIGNAL( clicked() ), 
           this, SLOT( getPathButtonCB() ) );
  connect( p_pathDialog->mergeButton, SIGNAL( clicked() ), 
           this, SLOT( mergeButtonCB() ) );
}

ReconstructionViewer::~ReconstructionViewer()
{
   // delete p_pathDialog;
}

void ReconstructionViewer::setFileNames
         (vector< string > rawFileName, 
          vector< string > segFileName, 
          vector< string > watFileName)
{
  p_rawFileNames = rawFileName;
  p_segFileNames = segFileName;
  p_watFileNames = watFileName;
}

void ReconstructionViewer::setFileNames
         ( vector< string > rawFileName, vector< string > segFileName)
{
  p_rawFileNames = rawFileName;
  p_segFileNames = segFileName;
}

void ReconstructionViewer::setSegEditFileNames
         ( vector< string > fileNames )
{
  p_segEditFileNames = fileNames;
}

// 
//  main() for running the section editor
//
void ReconstructionViewer::fileOpen()
{
  //optionsBox->setEnabled(true);

  p_sectionEditor.loadRawImage( p_rawFileNames );
  p_sectionEditor.loadSegImage( p_segFileNames );
  showRawImageShortcut();
  showSegImageShortcut();

  if( p_watFileNames.size() > 0 )
  {
     p_sectionEditor.loadWaterImage( p_watFileNames );
     showWaterImageShortcut();
  }

  if( p_segEditFileNames.size() > 0 )
     p_sectionEditor.loadSegEditImage( p_segEditFileNames );

  p_sectionEditor.createVTKPipeline();
  
  // initialize our scroll bar
  sectionScrollBar->setMinimum( p_sectionEditor.getMinSection() );
  sectionScrollBar->setMaximum( p_sectionEditor.getMaxSection() );
  sectionScrollBar->setValue( 0 );

  p_sectionEditor.updateRenderWindow();

//  p_sectionEditor.loadVTKFiles( !p_rebuildPaths );
//  if( p_rebuildPaths )
//     p_sectionEditor.rebuildPaths();

  p_sectionEditor.initializeLiveWire();
  p_sectionEditor.updateLiveWire();

}

void ReconstructionViewer::openMergeRegions()
{
   QString s = QFileDialog::getOpenFileName(
                    ".",
                    "*.*",
                    this,
                    "open file dialog",
                    "Choose a file" );

   if(!s.isNull()){
      string filename = (const char*)s.toAscii();
      cout << "file: " << filename  << endl ;
      openMergeRegions( filename ); 
      //this->openFile(filename.toAscii());
   }
}

void ReconstructionViewer::openPathFile( const vector< string >& pathFiles )
{
   // only handle one xml path file at a time, right now
   p_sectionEditor.loadVTKFiles( pathFiles[0], !p_rebuildPaths );

   if( p_rebuildPaths )
      p_sectionEditor.rebuildPaths();

   p_pathDialog->addPaths( p_sectionEditor.getPaths() );

   //p_sectionEditor.openPathFile( pathFile );
}

void ReconstructionViewer::setCorrFiles( const vector< string >& files )
{
   p_sectionEditor.initCorrelation( files );
}

void ReconstructionViewer::openMergeRegions( string filename )
{
   p_sectionEditor.openMergeRegions( filename ); 
}

void ReconstructionViewer::openMatrixFiles( vector< string > files )
{
   vector<string> matrix_filenames;  //  matrix*
   vector<string> label_mapping_filenames; // slice*

   for(unsigned int a = 0; a < files.size(); a++)
   {
      string a_file( files[a] );
      cout << "found " << a_file << endl;
      if( a_file.find( "matrix", 0 ) != string::npos )
         matrix_filenames.push_back(a_file);
      else if( a_file.find( "_labels", 0 ) != string::npos )
         label_mapping_filenames.push_back(a_file);
      else
         cout << "file: " << a_file << " not identified." << endl ;
   }
   cout << matrix_filenames.size() << " matricies found" << endl ;
   cout << label_mapping_filenames.size() << " label files found" << endl ;

   p_sectionEditor.openMatrixFiles( matrix_filenames,
                                    label_mapping_filenames );

}

void ReconstructionViewer::openPolyDataFiles( vector< string > files )
{
   p_sectionEditor.openPolyDataFiles( files );

}

void ReconstructionViewer::fileExit() {
  qApp->exit();
}

void ReconstructionViewer::sectionChangedPlus( )
{
   int i = sectionScrollBar->value();
   sectionScrollBar->setValue( i + 1 );
   //p_sectionEditor.showSection( i );

} 

void ReconstructionViewer::sectionChangedMinus( )
{
   int i = sectionScrollBar->value();
   sectionScrollBar->setValue( i - 1 );
   //p_sectionEditor.showSection( i );
} 


void ReconstructionViewer::sectionChanged( )
{
   int i = sectionScrollBar->value();
   p_sectionEditor.showSection( i );

   //string l(i);
   //scrollLabel->setText( l.c_str() );

   //p_sectionEditor.updateLiveWire();
} 

void ReconstructionViewer::goToSection( int z )
{
   /*
   vtkWorldPointPicker* picker = vtkWorldPointPicker::New();
   picker->Pick( mx, my, 0, p_renWater );
   double pickCoords[3];
   picker->GetPickPosition( pickCoords);
   cout << "picked: " ;
   for( int p = 0; p < 3; p++ )
      cout << pickCoords[p] << " " ;
   cout << endl;
   */

   int slice = z / p_sectionEditor.getSpacing()[2] ;
   //cout << "going to slice " << slice << endl ;
   bool stat = p_sectionEditor.showSection( slice );

   // update everything else
   if( stat )
      sectionScrollBar->setValue( slice );

} 

void ReconstructionViewer::showRawImage()
{
   p_sectionEditor.showRaw();
}

void ReconstructionViewer::showRawImageShortcut()
{
   showImageButton->click();
}

void ReconstructionViewer::showSegImage()
{
   p_sectionEditor.showSeg();
}

void ReconstructionViewer::showSegImageShortcut()
{
   showSegmentationButton->click();
}

void ReconstructionViewer::showWaterImage()
{
   p_sectionEditor.showWater();
}

void ReconstructionViewer::showWaterImageShortcut()
{
   showWatershedButton->click(); 
}

void ReconstructionViewer::applyWatershedToSlice( )
{
   if( floodFillCheckBox->isChecked() )
   {
      int section = sectionScrollBar->value();
      p_sectionEditor.createFloodFill( section );
   }
   else
      applyWatershed( SLICE );
}
/*
   cout << "flag: " ;//<< flag << endl ;
   return;
   
   float level = atof ( levelText->text() );
   float thresh = atof ( watershedThresholdText->text() );
   int section = sectionScrollBar->value();
   p_sectionEditor.setWatershedLevel( level );
   p_sectionEditor.setWatershedThreshold( thresh );
   p_sectionEditor.createWatershedSegmentation( section );
   showWaterImage();
*/

void ReconstructionViewer::applyWatershedToVolume()
{
   applyWatershed( VOLUME );
}

void ReconstructionViewer::applyWatershed( int scope )
{
   //cout << "got scope: " << scope << endl ;
   int section = sectionScrollBar->value();

   if( thresholdCheckBox->isChecked() )
   {
      float thresh = atof( thresholdText->text() );
      p_sectionEditor.setThreshold( thresh );
      if( scope == SLICE )
         p_sectionEditor.applyThreshold( section );
      else
         p_sectionEditor.applyThreshold();
   }
   else if( blurCheckBox->isChecked() )
   {
      float sigma = atof( blurText->text() );
      p_sectionEditor.setBlurSigma( sigma );
      if( scope == SLICE )
         p_sectionEditor.applyBlur( section );
      else
         p_sectionEditor.applyBlur();
   }
   else
   {
      // do nothing
      cout << "need to check a box " << endl ;
   }

   if( watershedCheckBox->isChecked() )
   {
      float level = atof ( levelText->text() );
      float thresh = atof ( watershedThresholdText->text() );
      p_sectionEditor.setWatershedLevel( level );
      p_sectionEditor.setWatershedThreshold( thresh );
      if( scope == SLICE )
         p_sectionEditor.createWatershedSegmentation( section );
      else
         p_sectionEditor.createWatershedSegmentation();
   }
}

void ReconstructionViewer::setPicking()
{
   p_sectionEditor.setPick();
}

void ReconstructionViewer::setDownSampleValue( int ds )
{
   p_sectionEditor.setDownSampleValue( ds );
}

void ReconstructionViewer::render()
{
   p_sectionEditor.processPath();
   p_pathDialog->addPaths( p_sectionEditor.getPaths() );
   itemSelectionChangedCB();
}

void ReconstructionViewer::mergeRegions()
{
   p_sectionEditor.mergeRegions( 150 );
}

void ReconstructionViewer::correlateRegions()
{
   p_sectionEditor.correlateRegions( );
}

void ReconstructionViewer::computePaths()
{
   p_sectionEditor.computePaths();
}

void ReconstructionViewer::printPaths()
{
   p_sectionEditor.printPaths();
}

void ReconstructionViewer::viewPathData()
{
   p_pathDialog->show();
}

void ReconstructionViewer::newPathButtonCB()
{
   setPicking();
   Qt::CheckState cs = Qt::Checked;
   manualTrackButton->setCheckState( cs );
}

void ReconstructionViewer::getPathButtonCB()
{
   p_sectionEditor.calcNextPath();
   p_pathDialog->addPaths( p_sectionEditor.getPaths() );
   itemSelectionChangedCB();
}

void ReconstructionViewer::autoFillButtonCB()
{
   vector< int > ids_selected = getSelectedPaths();

   if( ids_selected.size() > 1 ||
         ids_selected.size() == 0 )
   {
      cout << "ERROR: Select one path. " << endl;
      return;
   }

   p_sectionEditor.autoFillPath( ids_selected[0] );
   //p_sectionEditor.rebuildPath();
}

void ReconstructionViewer::editPathButtonCB()
{
   vector< int > ids_selected = getSelectedPaths();

   if( p_pathDialog->editCheckBox->checkState() == Qt::Checked )
   {
      cout << "on" << endl ;
      if( ids_selected.size() > 1 ||
            ids_selected.size() == 0 )
      {
         cout << "ERROR: Select one path. " << endl;
         return;
      }
        
      newPathButtonCB(); // set the pick state
      p_sectionEditor.setPickPath( ids_selected[0] );
      //Qt::CheckState cs = Qt::Checked;
      //   manualTrackButton->setCheckState( cs );
   }
   else if( p_pathDialog->editCheckBox->checkState() == Qt::Unchecked )
   {
      cout << "off" << endl;
      //p_sectionEditor.deletePathGeometry( ids_selected[0] ); wha?
      p_sectionEditor.rebuildPath();
   }
}

void ReconstructionViewer::deletePathButtonCB()
{
   cout << "deleting " ;

   vector< int > ids_to_delete = getSelectedPaths();
   // remove the item from the widget
   QList< QListWidgetItem* > si = 
      p_pathDialog->pathWidget->selectedItems();
   QList< QListWidgetItem* >::const_iterator i;
   for( i = si.begin(); i != si.end(); i++ )
   {
      cout << (*i)->text().toStdString() << " ";
      p_pathDialog->pathWidget->removeItemWidget( *i );
      delete *i;
   }
   cout << endl ;


   // remove the vtk and path data structures
   p_sectionEditor.deletePaths( ids_to_delete );


}

void ReconstructionViewer::mergeButtonCB()
{
   vector< int > ids_to_merge = getSelectedPaths();
   if( ids_to_merge.size() != 2 )
   {
      cout << "need to select two paths to merge.  " 
         << ids_to_merge.size() << " selected" << endl ;
      return;
   }

   // else
   cout << "merging " << ids_to_merge[0] << " and "
      << ids_to_merge[1] << endl;
   p_sectionEditor.mergePaths( ids_to_merge[0], ids_to_merge[1] );
   /*
   p_sectionEditor.mergePaths( 0, 65 );
   p_sectionEditor.mergePaths( 1, 72 );
   p_sectionEditor.mergePaths( 2, 69 );
   p_sectionEditor.mergePaths( 5, 66 );
   p_sectionEditor.mergePaths( 7, 70 );
   p_sectionEditor.mergePaths( 8, 86 );
   p_sectionEditor.mergePaths( 9, 94 );
   p_sectionEditor.mergePaths( 10, 74 );
   p_sectionEditor.mergePaths( 11, 73 );
   p_sectionEditor.mergePaths( 13, 71 );
   p_sectionEditor.mergePaths( 16, 83 );
   p_sectionEditor.mergePaths( 17, 75 );
   p_sectionEditor.mergePaths( 19, 88 );
   p_sectionEditor.mergePaths( 20, 95 );
   p_sectionEditor.mergePaths( 22, 82 );
   p_sectionEditor.mergePaths( 23, 77 );
   p_sectionEditor.mergePaths( 27, 85 );
   p_sectionEditor.mergePaths( 51, 92 );
   p_sectionEditor.mergePaths( 59, 84 );
   p_sectionEditor.mergePaths( 61, 90 );
   p_sectionEditor.mergePaths( 62, 93 );
   
   p_sectionEditor.mergePaths( 0, 140 );
   p_sectionEditor.mergePaths( 1, 148 );
   p_sectionEditor.mergePaths( 2, 144 );
   p_sectionEditor.mergePaths( 5, 139 );
   p_sectionEditor.mergePaths( 7, 143 );
   p_sectionEditor.mergePaths( 10, 168 );
   p_sectionEditor.mergePaths( 11, 141 );
   p_sectionEditor.mergePaths( 13, 166 );
   p_sectionEditor.mergePaths( 17, 156 );
   p_sectionEditor.mergePaths( 19, 161 );
   p_sectionEditor.mergePaths( 20, 171 );
   p_sectionEditor.mergePaths( 22, 157 );
   p_sectionEditor.mergePaths( 23, 145 );
   p_sectionEditor.mergePaths( 27, 163 );
   p_sectionEditor.mergePaths( 51, 162 );
   p_sectionEditor.mergePaths( 61, 155 );
   p_sectionEditor.mergePaths( 67, 142 );
   p_sectionEditor.mergePaths( 68, 154 );
   p_sectionEditor.mergePaths( 76, 153 );
   p_sectionEditor.mergePaths( 78, 149 );
   p_sectionEditor.mergePaths( 79, 151 );
   p_sectionEditor.mergePaths( 87, 164 );
   p_sectionEditor.mergePaths( 89, 152 );
   p_sectionEditor.mergePaths( 91, 158 );
   p_sectionEditor.mergePaths( 107, 169 );
   p_sectionEditor.mergePaths( 108, 147 );
   p_sectionEditor.mergePaths( 112, 172 );
   p_sectionEditor.mergePaths( 125, 150 );
   p_sectionEditor.mergePaths( 127, 146 );
   p_sectionEditor.mergePaths( 132, 165 );
   */

   //
   //  delete the path from the widget
   //
   p_pathDialog->addPaths( p_sectionEditor.getPaths() );
   /*string st;
   st += ids_to_merge[1];
   QList< QListWidgetItem* > si = 
      p_pathDialog->pathWidget->selectedItems();
   QList< QListWidgetItem* >::const_iterator i;
   for( i = si.begin(); i != si.end(); i++ )
   {
      if( (*i)->text().toStdString().compare( st )
            == 0 )
      {
      p_pathDialog->pathWidget->removeItemWidget( *i );
      delete *i;
      }
   }
   cout << endl ;
   */

}

vector< int > ReconstructionViewer::getSelectedPaths()
{
   QList< QListWidgetItem* > si = 
      p_pathDialog->pathWidget->selectedItems();
   //cout << "items selected: " << si.size() << endl;
   vector< int > ids_to_show;
   QList< QListWidgetItem* >::const_iterator i;
   for( i = si.begin(); i != si.end(); i++ )
   {
      string s = (*i)->text().toStdString();
      ids_to_show.push_back( atoi( s.c_str() ) );
   }

   return ids_to_show;
}

void ReconstructionViewer::itemSelectionChangedCB()
{
   vector< int > ids_to_show = getSelectedPaths();
   p_sectionEditor.showIso( ids_to_show );
}

void ReconstructionViewer::drawTrig()
{
   //cout << "draw" << endl;
   p_sectionEditor.setDraw( true );
}

void ReconstructionViewer::eraseTrig()
{
   //cout << "RV: erase" << endl;
   p_sectionEditor.setDraw( false );
}

