#ifndef RECONSTRUCTIONVIEWER_H
#define RECONSTRUCTIONVIEWER_H

#include "qmainwindow.h"
#include "ui_uiSimpleView4.h"
#include "SectionEditor.h"
#include "PathDialogGUI.h"
#include <QShortcut>
#include <qlistview.h>
#include <qwindowsstyle.h>

#include <string>
#include <vector>

// Forward class declarations
class vtkCylinderSource;
class vtkPolyDataMapper;
class vtkExodusReader;
class vtkDataSetMapper;
class vtkActor;
class vtkRenderer;
class vtkMetaImageReader;

using namespace std;

void pickSliceCallback(vtkObject* caller, unsigned long event, void *calldata, void *);


class ReconstructionViewer : public QMainWindow, 
                             private Ui_MainWindow
{
   Q_OBJECT

   public:

      // Constructor/Destructor
      ReconstructionViewer(QWidget* parent = 0);
      ~ReconstructionViewer();

      void setFileNames( vector< string > rawFileName, 
                         vector< string > segFileName,
                         vector< string > watFileName );

      void setFileNames( vector< string > rawFileName, 
                         vector< string > segFileName ) ;

      void setSegEditFileNames( vector< string > fileName ) ;

      void setSpacing( vector< float > spacing )
      { p_sectionEditor.setSpacing( spacing ); };

      void setVTKPath( string vtkpath )
      { p_sectionEditor.setVTKPath( vtkpath ); };

      void openMergeRegions( string mergeFileName );
      void openMatrixFiles( vector< string > files ); 
      void openPathFile( const vector< string >& files );
      void openPolyDataFiles( vector< string > files ); 
      void setCorrFiles( const vector< string >& files ); 

      void setWindow( int w, int o ) 
         { p_sectionEditor.setWindow( w, o ); };

      void setBuildPaths( bool rebuildPaths ) 
         { p_rebuildPaths = rebuildPaths; };

      void setDownSampleValue( int ds );

      public slots:

         virtual void fileOpen();
         virtual void openMergeRegions();
         virtual void fileExit();
         virtual void drawTrig();
         virtual void eraseTrig();
         virtual void sectionChanged( );
         virtual void sectionChangedPlus( );
         virtual void sectionChangedMinus( );
         virtual void goToSection( int z );
         virtual void showRawImage();
         virtual void showRawImageShortcut();
         virtual void showSegImage();
         virtual void showSegImageShortcut();
         virtual void showWaterImage();
         virtual void showWaterImageShortcut();
         virtual void applyWatershedToSlice();
         virtual void applyWatershedToVolume();
         virtual void setPicking();
         virtual void render();
         virtual void mergeRegions();
         virtual void correlateRegions();
         virtual void computePaths();
         virtual void printPaths();
         virtual void viewPathData();
         virtual void newPathButtonCB();
         virtual void getPathButtonCB();
         virtual void deletePathButtonCB();
         virtual void editPathButtonCB();
         virtual void autoFillButtonCB();
         virtual void mergeButtonCB();
         virtual void itemSelectionChangedCB();

   protected:
         
      vector< string > p_rawFileNames;
      vector< string>  p_segFileNames;
      vector< string>  p_segEditFileNames;
      vector< string >  p_watFileNames;

      PathDialogGUI* p_pathDialog;

      protected slots:

   private:

         enum Scope { SLICE, VOLUME };

         SectionEditor p_sectionEditor;

         bool p_rebuildPaths;

         void applyWatershed( int );

         vector< int > getSelectedPaths();

};

#endif // SIMPLEVIEW_H

