#ifndef __REGION_H__
#define __REGION_H__

#include "alltypedefs.h"

#include "Correspondence.h"

class Region
{
   private:

      LabelType p_label;
      int p_sliceNum;
      LabelImageType::IndexType p_minIndex;
      LabelImageType::IndexType p_maxIndex;
      LabelImageType::IndexType p_centerIndex;

      ScalarImageType::Pointer p_regionImage;

      void calcRegionCenter();

   public:

      Region(){};

      Region( const Region& R );

      Region( LabelType l, LabelImageType::IndexType i1,
                           LabelImageType::IndexType i2,
                           int slice );

      void updateIndex( LabelImageType::IndexType index );

      void print();

      ScalarImageType::SizeType getScalarSize();
      LabelImageType::SizeType getLabelSize();

      ScalarImageType::IndexType getScalarStart();
      LabelImageType::IndexType getLabelStart();

      LabelType label() { return p_label; };
      int sliceNum() { return p_sliceNum; };
      LabelImageType::IndexType centerIndex() { return p_centerIndex; };

      void image( ScalarImageType::Pointer image );
      ScalarImageType::Pointer image(){ return p_regionImage; };

      Correspondence compare( Region& R, float dist_thresh );

      //ScalarImageType::Pointer pad( ScalarImageType::SizeType newSize );
};

class RegionList
{
   private:

      typedef map< LabelType, Region > RegionMapType;
      RegionMapType p_regionList;

   public:

      RegionList( LabelImageType::Pointer label_image,
                  ScalarImageType::Pointer raw_image,
                  int slice_num,
                  const vector< LabelType >& skip_these_labels );

      CorrespondenceMatrix compare( RegionList& rl, float dist_thresh );

      RegionMapType::iterator begin() { return p_regionList.begin(); };
      RegionMapType::iterator end() { return p_regionList.end(); };
};

#endif
