#include "SectionEditor.h"

#include "ItkVtkConvert.h"
#include "WatershedFilter.h"
//#include "CalcRegionCorrelations.h"
//#include "STrAND.h"

#include "vtkExtractVOI.h"
#include "vtkImageActor.h"
#include "vtkPointData.h"
#include "vtkDataArray.h"
#include "vtkUnsignedCharArray.h"
#include "vtkUnsignedShortArray.h"
#include "vtkImageData.h"
//#include "vtkImageFlip.h"
#include "vtkImageShiftScale.h"
#include "vtkImageThreshold.h"
#include "vtkImageGaussianSmooth.h"
#include "vtkImageAccumulate.h"
#include "vtkInteractorStyleImage.h"
#include "vtkInteractorStyleTrackballCamera.h"
#include "vtkRenderWindowInteractor.h"
#include <vtkMetaImageReader.h>
#include <vtkMetaImageWriter.h>
#include <vtkPolyDataWriter.h>
#include <vtkPolyDataReader.h>
#include <vtkPolyDataNormals.h>
#include <vtkTriangleFilter.h>
#include <vtkRenderWindow.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>
#include <vtkImageMarchingCubes.h>
#include <vtkPolyDataMapper.h>
#include <vtkProperty.h>
#include <vtkImageGaussianSmooth.h>
#include <vtkImageMapToWindowLevelColors.h>
#include <vtkImageMapToColors.h>
#include <vtkCallbackCommand.h>
#include <vtkColorTransferFunction.h>
#include <vtkCommand.h>
#include "vtkWorldPointPicker.h"
#include <vtkImageDataStreamer.h>
#include <vtkExtentTranslator.h>
#include <vtkPNGReader.h>
#include <vtkPNGWriter.h>
#include <vtkImageDataStreamer.h>
#include <vtkImageClip.h>
#include <vtkStringArray.h>
#include <vtkImageMathematics.h>
#include <vtkSphereSource.h>
#include <vtkSmartPointer.h>

#include "vtkImageEditor.h"

#include "itkImage.h"
//#include "itkCastImageFilter.h"
//#include "itkWatershedImageFilter.h"

#include "png.h"

#include "CalcCorrelations.h"
#include "CalcPaths.h"


typedef itk::Image< float, 2 > itkImageType;
typedef itk::Image< float, 3 > itkImage3DType;
//typedef itk::Image< unsigned long, 2 > itkLongImageType;
//typedef itk::WatershedImageFilter< itkImageType > WatershedFilterType;
//typedef itk::CastImageFilter< itkLongImageType, ImageType > CastImageFilter;


// Constructor
SectionEditor::SectionEditor(  )   :
  p_minSection(0),
  p_maxSection(0),
  p_curSection(0),
  p_ren(NULL),
  p_renWinInteractor(NULL),
  p_interactor(NULL),
  p_renWater(NULL),
  p_renWinInteractorWater(NULL),
  p_interactorWater(NULL),
  p_rawImageReader(NULL),
  p_rawImageStreamer(NULL),
  p_rawSliceExtractor(NULL),
  p_rawSectionActor(NULL),
  p_segImageReader(NULL),
  p_segImageStreamer(NULL),
  p_segSliceExtractor(NULL),
  p_segSectionActor(NULL),
  p_segColorMap(NULL),
  p_segImageBlend(NULL),
  p_segEditImageReader(NULL),
  p_segEditImageStreamer(NULL),
  p_segEditSliceExtractor(NULL),
  p_segEditSectionActor(NULL),
  p_segEditColorMap(NULL),
  p_waterImageReader(NULL),
  p_waterImageStreamer(NULL),
  p_waterSliceExtractor(NULL),
  p_waterSectionActor(NULL),
  p_waterColorMap(NULL),
  p_randomTransferFx(NULL),
  p_pathImageData(NULL)
{
  p_ren = vtkRenderer::New();
  p_ren->SetBackground( .4, .4, .4 );

  
  p_renWater = vtkRenderer::New();
  p_renWater->SetBackground( .8, .8, .8 );
  // p_interactor = vtkInteractorStyleImage::New();
 // p_renWinInteractor = p_ren->GetRenderWindow()->MakeRenderWindowInteractor();
 // p_renWinInteractor->SetInteractorStyle( p_interactor );

  num_paths = 0;
  p_numLabels = 150; // needs to be set by the image
  p_showRaw = true;
  p_showSeg = true;
  p_showWater = true;
  p_pick = false;

  p_watershedLevel = .5;
  p_watershedThreshold = 0.2;
  p_blurSigma = 5;
  p_threshold = .5;

  //p_strand = new STrAND();
  //p_strand->setSections( 0, 25 ); // HACK

  srand( time(NULL) ); //1 );

  p_pathInc = 0;

  p_dims[0] = 0;
  p_dims[1] = 0;
  p_dims[2] = 0;
  p_dims[3] = 0;
  p_dims[4] = 0;
  p_dims[5] = 0;

  p_spacing[0] = 1;
  p_spacing[1] = 1;
  p_spacing[2] = 1;

  p_path = NULL;

}

SectionEditor::~SectionEditor()
{
   /*
   if( p_ren )
      p_ren->Delete();
   if( p_renWater )
      p_renWater->Delete();
   if( p_rawImageData )
      p_rawImageData->Delete();
   if( p_rawShiftScale )
      p_rawShiftScale->Delete();
   if( p_rawSectionActor )
      p_rawSectionActor->Delete();
   if( p_segImageDataFloat )
      p_segImageDataFloat->Delete();
   if( p_segImageData )
      p_segImageData->Delete();
   if( p_segSliceExtractor )
      p_segSliceExtractor->Delete();
   if( p_segShiftScale )
      p_segShiftScale->Delete();
   if( p_segSectionActor )
      p_segSectionActor->Delete();
   if( p_waterImageData )
      p_waterImageData->Delete();
   if( p_waterSliceExtractor )
      p_waterSliceExtractor->Delete();
   if( p_waterColorMap )
      p_waterColorMap->Delete();
   if( p_waterSectionActor )
      p_waterSectionActor->Delete();
   if( p_satLut )
      p_satLut->Delete();
   if( p_randomTransferFx )
      p_randomTransferFx->Delete();
      */
}

vtkRenderer* SectionEditor::getRenderer()
{
   return p_ren;
}

vtkRenderer* SectionEditor::get3DRenderer()
{
   return p_renWater;
}

void SectionEditor::updateRenderWindow()
{
   p_ren->GetRenderWindow()->Render();
   p_renWater->GetRenderWindow()->Render();
}

bool SectionEditor::showSection( int s )
{
   bool status = false;

   if( ( s <= p_maxSection ) && ( s >= p_minSection ) )
   {
      status = true;
      p_curSection = s;
      cout << "p_curSection: " << p_curSection << endl;
      //std::cout << "setting z slice to " << p_curSection << std::endl;
      p_dims[4] = p_dims[5] = p_curSection;
      p_rawSliceExtractor->SetOutputWholeExtent(p_dims);
      p_rawSliceExtractor->Modified();
      p_segSliceExtractor->SetOutputWholeExtent(p_dims);
      p_segSliceExtractor->Modified();
      p_segEditSliceExtractor->SetOutputWholeExtent(p_dims);
      p_segEditSliceExtractor->Modified();
      //p_segEditImageStreamer->UpdateWholeExtent();
      //p_segEditImageStreamer->Update();
      p_segEditImageStreamer->Modified();
      if( p_waterSliceExtractor )
      {
         p_waterSliceExtractor->SetOutputWholeExtent(p_dims);
         p_waterSliceExtractor->Modified();
      }
   }
   
   p_liveWireEdit->save( p_segEditSliceExtractor );
   updateLiveWire();
   
   p_ren->ResetCameraClippingRange();
   p_ren->GetRenderWindow()->Render();
   p_renWater->ResetCameraClippingRange();
   p_renWater->GetRenderWindow()->Render();

   return status;
}

void SectionEditor::showIso( vector< int > isoIds )
{
   //
   // turn EVERYTHING off
   // .. first the isosurfaces, and then the spheres
   map< int, vtkActor* >::iterator iter;
   for( iter = p_isoActors.begin(); iter != p_isoActors.end(); ++iter )
   {
     iter->second->SetVisibility( false ); 
   }
   map< int, vector< vtkActor*> >::iterator itervec;
   for( itervec = p_isoSphereActors.begin(); itervec != p_isoSphereActors.end(); ++itervec )
   {
     vector< vtkActor* > vec = itervec->second; 
     for( unsigned int a = 0; a < vec.size(); a++ )
        vec[ a ]->SetVisibility( false );
   }
   
   //
   // turn on what we want to see
   //
   for( unsigned int i = 0; i < isoIds.size(); i++ )
   {
      int id = isoIds[i];
      //bool v = p_isoActors[ id ]->GetVisibility();
      //p_isoActors[ id ]->SetVisibility( !v );
      p_isoActors[ id ]->SetVisibility( true );

      for( unsigned int a = 0; a < p_isoSphereActors[ id ].size(); a++ )
         p_isoSphereActors[ id ][ a ]->SetVisibility( true );
   }
   p_renWater->GetRenderWindow()->Render();
   p_ren->GetRenderWindow()->Render();
   
}

void SectionEditor::showRaw()
{
   if( p_rawSectionActor )
   {
      p_showRaw = !p_showRaw;
      p_rawSectionActor->SetVisibility( p_showRaw );
      p_rawSectionActor2->SetVisibility( p_showRaw );
      p_ren->GetRenderWindow()->Render();
      p_renWater->GetRenderWindow()->Render();
   }
}

void SectionEditor::showSeg()
{
   if ( !p_segSectionActor )
      return;

   p_showSeg = !p_showSeg;
   p_segSectionActor->SetVisibility( p_showSeg );
   p_ren->GetRenderWindow()->Render();
}

void SectionEditor::showWater()
{
   if ( !p_waterSectionActor )
      return;

   p_showWater = !p_showWater;
   p_waterSectionActor->SetVisibility( p_showWater );
   p_ren->GetRenderWindow()->Render();
}

void SectionEditor::setPick()
{
   p_pick = !p_pick;
}

void SectionEditor::setPickPath( int id )
{
   p_path = p_listOfPaths.getPathById( id );
   cout << "editing path: ";
   p_path->print();
}

//
//  automatically fill in the missing path data for a single path
//
void SectionEditor::autoFillPath( int id )
{
   p_path = p_listOfPaths.getPathById( id );
   cout << "before path: ";
   p_path->print();

   // find the missing slices...
   // get the list of all slices
   vector< int > allSlices = p_path->getSlices();
   // they might be out of order, so let's sort them
   sort( allSlices.begin(), allSlices.end() );
   int size = (int)allSlices.size();
   // how many slices are we missing?
   int lastSlice = allSlices[ allSlices.size()-1 ];
   cout << "missing " << lastSlice - size + 1 << " slices." << endl; 
   //cout << "missing " << allSlices.size() - allSlices[allSlices.size()-1] << " slices." << endl; 
   vector< int > missingSlices;
   cout << "slices: " ;
   for( unsigned int i = 0; i < allSlices.size()-1; i++ )
   {
      int ms = allSlices[i] + 1;
      if( ms != allSlices[i+1] )
      {
         missingSlices.push_back( allSlices[i] + 1 );
         cout << allSlices[i] + 1 << " ";

         //
         // make up the missing slice
         //
         int nx1 = p_path->getX( allSlices[i] );
         int ny1 = p_path->getY( allSlices[i] );
         int nx2 = p_path->getX( allSlices[i] + 2 );
         int ny2 = p_path->getY( allSlices[i] + 2 );
         int x = (nx1 + nx2) / 2;
         int y = (ny1 + ny2) / 2;
         // make sure there's a label at this x,y location
         unsigned short label = 1; // junk for now 
         //   getPixelValue( x, p_dims[3] - y, ms ); 
         cout << "adding slice " << ms 
              << " x,y: " << x << "," << p_dims[3] - y 
              << " label: " << label << endl;
         if( label != 0 )
            p_path->add( ms, label, 0, 0, x, y );
      }
   }
   cout << endl ;

   cout << "after path: ";
   p_path->print();

   //
   //  delete just the actors before we rebuild them
   //
   deletePathGeometry( id );
  
   rebuildPath();
   // make the new isosurface and spheres visible
   p_isoActors[ id ]->SetVisibility( true );
   for( unsigned int a = 0; a < p_isoSphereActors[ id ].size(); a++ )
   {
       //p_isoSphereActors[ id ][ a ]->SetVisibility( true );
       p_isoSphereActors[ id ][ a ]->GetProperty()->SetColor
          ( p_isoActors[ id ]->GetProperty()->GetColor() );
   }

   p_renWater->GetRenderWindow()->Render();
   p_ren->GetRenderWindow()->Render();
}

//
//  delete the vtk renderer objects from the scene
//
void SectionEditor::deletePathGeometry( int id )
{
   // remove the iso actor for this neuron
   map< int, vtkActor* >::iterator actorIter;
   actorIter = p_isoActors.find( id );
   // remove the actor
   vtkActor* A = (*actorIter).second;
   p_renWater->RemoveActor( A );
   A->Delete();
   // TODO: need to delete the mappers and reads associated with the
   // actor, too.
   p_isoActors.erase( actorIter );

   // remove the spheres for this actor 
   // TODO: need to delete the spheres too
   /* not going to delete for now, just change the color (below)
   for( unsigned int a = 0; a < p_isoSphereActors[ id ].size(); a++ )
   {
      p_ren->RemoveActor( p_isoSphereActors[ id ][ a ] );
      p_isoSphereActors[ id ][ a ]->Delete();
   }
   p_isoSphereActors[ p_path->getID() ].clear();
   */

   // write out the new path - done in rebuildPath
   //p_listOfPaths.writeXMLPaths( p_pathFile.c_str() );
}

void SectionEditor::deletePaths( vector< int > deleteIds )
{
   // make a backup of the current path file
   string bkup = p_pathFile + ".bkup";
   p_listOfPaths.writeXMLPaths( bkup.c_str() );

   for( unsigned int i = 0; i < deleteIds.size(); i++ )
   {
      int did = deleteIds[i];

      map< int, vtkActor* >::iterator actorIter;
      actorIter = p_isoActors.find( did );
      // remove the actors
      if( actorIter != p_isoActors.end() )
      {
         vtkActor* A = (*actorIter).second;
         p_renWater->RemoveActor( A );
         A->Delete();
         // TODO: need to delete the mappers and reads associated with the
         // actor, too.
         p_isoActors.erase( actorIter );
         
      }
      else
         cout << "ERROR: could not remove actor, id: " << did << endl ;

      // remove paths from the listofpaths
      p_listOfPaths.deletePaths( deleteIds );
      
   }
   p_renWater->GetRenderWindow()->Render();

   // write out the new path file
   p_listOfPaths.writeXMLPaths( p_pathFile.c_str() );

}

void SectionEditor::mergePaths( int p1, int p2 )
{  
   Path* path1 = p_listOfPaths.getPathById( p1 );
   Path* path2 = p_listOfPaths.getPathById( p2 );
   // merge path2 with path1
   path1->merge( path2 );
   // remove path2 from the listofpaths
   vector< int > del; del.push_back( p2 );
   p_listOfPaths.deletePaths( del );

   // delete both actors for these paths
   del.push_back( p1 );
   for( unsigned int i = 0; i < del.size(); i++ )
   {
      int did = del[i];

      map< int, vtkActor* >::iterator actorIter;
      actorIter = p_isoActors.find( did );
      // remove the actors
      if( actorIter != p_isoActors.end() )
      {
         vtkActor* A = (*actorIter).second;
         p_renWater->RemoveActor( A );
         A->Delete();
         // TODO: need to delete the mappers and reads associated with the
         // actor, too.
         p_isoActors.erase( actorIter );
         
      }
      else
         cout << "ERROR: could not remove actor, id: " << did << endl ;

   }

   p_path = path1;
   rebuildPath();

}

//
//  rebuild an isosurface for a single single path
//
void SectionEditor::rebuildPath()
{
   cout << "rebuildPath: " ; p_path->print();
   string vtkfile = processPath( p_path, p_path->getID() );
   //
   // write this out every time so that if rebuildPaths fails, we
   // don't have to edit the xml file to add in all the vtk filenames.
   //
   p_listOfPaths.writeXMLPaths( p_pathFile.c_str() );

   for( unsigned int s = 0; s < p_sphereActors.size(); s++ )
   {
      p_isoSphereActors[ p_path->getID() ].push_back( p_sphereActors[s] );
   }
}

//
//  loop over all the paths and rebuild their iso surfaces
//
void SectionEditor::rebuildPaths()
{
   cout << "rebuilding paths" << endl ;
   for( unsigned int p = 0; p < p_listOfPaths.size(); p++ )
   {
      string vtkfile;
      Path P = p_listOfPaths.get( p );
      cout << "rebuilding " << P.getID() << endl ;
      //if( P.getID() == 320 ){
      vtkfile = processPath( &P, P.getID() );
      //exit(1); }
      p_listOfPaths.getRef(p).setVTKFilename( vtkfile );
      p_listOfPaths.getRef(p).print();
      //
      // write this out every time so that if rebuildPaths fails, we
      // don't have to edit the xml file to add in all the vtk filenames.
      //
      p_listOfPaths.writeXMLPaths( p_pathFile.c_str() );
   }
   // write out the file again with the vtk filename
   //string pathfn = p_vtkPath + "/paths.xml";
   //p_listOfPaths.writeXMLPaths( pathfn.c_str() );
   p_listOfPaths.writeXMLPaths( p_pathFile.c_str() );
}

void SectionEditor::processPath()
{
   string vtkfile = processPath( p_path );
   p_path->setVTKFilename( vtkfile );
   p_path->setID( p_pathInc );
   p_listOfPaths.addPath( new Path( p_path ) );


   delete p_path;
   p_path = NULL;

   // write out the file
   if( p_pathFile.size() == 0 )
      p_pathFile = "tmpPath.xml";
   //string pathfn = p_vtkPath + "/paths.xml";
   //p_listOfPaths.writeXMLPaths( pathfn.c_str() );
   p_listOfPaths.writeXMLPaths( p_pathFile.c_str() );

}

string SectionEditor::processPath( Path* path, int pathid ) 
{
   cout << "spacing test: " 
      << p_spacing[0] << ", " << p_spacing[1]
      << ", " << p_spacing[2] << endl;
   
   if( pathid != -1 )
      p_pathInc = pathid;
   else
      p_pathInc++;
   path->print();

   int minslice = p_minSection;
   int maxslice = p_maxSection;
   path->getMinMaxSlice( minslice, maxslice );
   cout << "min, max: " << minslice << ", " << maxslice << endl ;

   int ds = p_downSample; // downsample factor

   int region[6];
   region[0] = p_dims[0];
   region[1] = p_dims[1];
   region[2] = p_dims[2];
   region[3] = p_dims[3];
   region[4] = minslice;
   //region[5] = maxslice; // default for now
   region[5] = maxslice; // default for now

   p_pathImageData = vtkImageData::New();
   p_pathImageData->SetSpacing( p_spacing[0]*ds, p_spacing[1]*ds, p_spacing[2] );
   p_pathImageData->SetExtent( region[0], region[1]/ds, 
                               region[2], region[3]/ds,
                               region[4], region[5] );
   p_pathImageData->SetScalarTypeToFloat();
   p_pathImageData->SetNumberOfScalarComponents(1);
   p_pathImageData->AllocateScalars();
   p_pathImageData->GetPointData()->GetScalars()->FillComponent(0, 0 );
   p_pathImageData->SetOrigin(0, 0, 0 ); //minslice);
   int* regiontst = p_pathImageData->GetExtent();
   cout << "extent: " ;
   for( unsigned int i = 0; i < 6; i++ )
      cout << regiontst[i] << " " ;
   cout << endl ;

   for( int s = minslice; s <= maxslice; s++ )
   {
      if( path->sliceExists( s ) ) // some paths skip slices!
      {
         region[4] = region[5] = s;
         p_waterSliceExtractor->SetOutputWholeExtent( region );
         p_waterSliceExtractor->Modified();
         p_waterSliceExtractor->UpdateWholeExtent();

         //vtkImageData* img = p_waterSliceExtractor->GetOutput();
         //vtkIdType id = img->ComputePointId( xyz );
         //vtkDataArray* data = img->GetPointData()->GetScalars();
         vtkDataArray* data = 
            p_waterSliceExtractor->GetOutput()->GetPointData()->GetScalars();
         vtkUnsignedShortArray* udata = dynamic_cast< vtkUnsignedShortArray* >( data );

         // get the label at these locations
         int xyz[3];
         xyz[0] = path->getX( s ); 
         xyz[1] = region[3] - path->getY( s );  // new vtk version
         //xyz[1] = path.getY( s ); 
         xyz[2] = s; //0 ;
         vtkIdType id  = p_waterSliceExtractor->GetOutput()->ComputePointId( xyz );
         unsigned short val = udata->GetValue( id );
         cout << "id: " << id << ", slice: " << s << ", label: " << val << 
            " x,y: " << xyz[0] << "," << xyz[1] << "," << xyz[2] << endl ;
         //cout << "test: " << getPixelValue( xyz[0], xyz[1], xyz[2] ) << endl ;

         if ( val != 0 )
         {
            for(int y = region[2]; y < region[3]; y+=ds)
            {
               for(int x = region[0]; x < region[1]; x+=ds)
               {
                  // get the pixel value from the image
                  int ijk[3];
                  ijk[0] = x; ijk[1] = y; ijk[2] = s;
                  vtkIdType id = p_waterSliceExtractor->GetOutput()->ComputePointId( ijk );
                  unsigned short pval = udata->GetValue( id );
                  //cout << pval << " " ;
                  if ( pval == val )
                     p_pathImageData->SetScalarComponentFromFloat
                        ( x/ds, y/ds, s, 0, 1.0 );
                  //else
                  //   p_pathImageData->SetScalarComponentFromFloat
                  //                       ( x/ds, y/ds, s, 0, 0.0 );
               }
            }
         } // end if
      }

   }

   // put the extractor back to where it was supposed to be set
   p_waterSliceExtractor->SetOutputWholeExtent( p_dims );
   p_waterSliceExtractor->Modified();


   //
   // write the data to a couple of files for use later
   //
   /*
   vtkMetaImageWriter *vtkVolWriter = vtkMetaImageWriter::New();
   vtkVolWriter->SetInput( p_pathImageData );//vtkSlice );
   char tmp[50];
   sprintf( tmp, "%03i", p_pathInc );
   string incstr( tmp );
   string wf = "vtkPathImageData" + incstr + ".mha";
   vtkVolWriter->SetFileName( wf.c_str() );
   vtkVolWriter->Write();
   cout << "binary volume written to " << wf << endl ;
   vtkVolWriter->Delete();
   */

   string vtk_filename = isosurfaceSeg( p_pathImageData, 1.0 );

   p_pathImageData->Delete();

   removeSpheres();

   return vtk_filename;
}

void SectionEditor::printPaths()
{
   p_listOfPaths.print( "nerv_paths.txt" );
}

void SectionEditor::setWindow( int w, int o )
{
   p_window = w;
   p_offset = o;
}

// ---------------------------------
//  load in the sequences of images
//

void SectionEditor::loadRawImage( vector< string > filenames )
{
  p_rawImageReader = vtkPNGReader::New();
  p_rawImageStreamer = vtkImageDataStreamer::New();
  p_rawSliceExtractor = vtkImageClip::New();
  bool calcDims = true;
  loadImageData( filenames, 
                 p_rawImageReader,
                 p_rawImageStreamer,
                 p_rawSliceExtractor,
                 calcDims );
}

void SectionEditor::loadSegImage( vector< string > filenames )
{
  p_segImageReader = vtkPNGReader::New();
  p_segImageStreamer = vtkImageDataStreamer::New();
  p_segSliceExtractor = vtkImageClip::New();
  loadImageData( filenames, 
                 p_segImageReader,
                 p_segImageStreamer,
                 p_segSliceExtractor );

  // add a little color
  vtkLookupTable* satLut = vtkLookupTable::New();
  satLut->SetTableRange (0, 2);
  satLut->SetHueRange (.4, .5);
  satLut->SetAlphaRange( 0, 1 );
  satLut->SetSaturationRange (0, 1);
  satLut->SetValueRange (0, 1);
  satLut->Build(); //effective built

  p_segColorMap = vtkImageMapToColors::New();
  p_segColorMap->SetLookupTable( satLut );

}

void SectionEditor::loadSegEditImage( vector< string > filenames )
{
  p_segEditImageReader = vtkPNGReader::New();
  p_segEditImageStreamer = vtkImageDataStreamer::New();
  p_segEditSliceExtractor = vtkImageClip::New();
  //cout << "loading seg edit image data" << endl ;
  loadImageData( filenames, 
                 p_segEditImageReader,
                 p_segEditImageStreamer,
                 p_segEditSliceExtractor );

  // add a little color
  vtkLookupTable* satLut = vtkLookupTable::New();
  satLut->SetTableRange (0, 2);
  satLut->SetHueRange (.4, .5);
  satLut->SetSaturationRange (0, 1);
  satLut->SetValueRange (0, 1);
  satLut->Build(); //effective built

  p_segEditColorMap = vtkImageMapToColors::New();
  p_segEditColorMap->SetInputConnection( p_segEditSliceExtractor->GetOutputPort() );
  p_segEditColorMap->SetLookupTable( satLut );
}


void SectionEditor::loadWaterImage( vector< string > filenames )
{
  //p_waterImageReader = vtkMetaImageReader::New();
  p_waterImageReader = vtkPNGReader::New();
  p_waterImageStreamer = vtkImageDataStreamer::New();
  p_waterSliceExtractor = vtkImageClip::New();

  loadImageData( filenames, 
                 p_waterImageReader,
                 p_waterImageStreamer,
                 p_waterSliceExtractor );
/*
  //
  //  read in the files
  //
  cout << "watershed images: " << endl ;
  vtkStringArray *allfilenames = vtkStringArray::New();
  for ( unsigned int i = 0; i < filenames.size(); i++ )
  {
     allfilenames->InsertNextValue( filenames[i].c_str() );
     cout << allfilenames->GetValue( i ) << endl;
  }
  cout << allfilenames->GetNumberOfValues() << endl;
  p_waterImageReader->SetFileNames( allfilenames );
  p_waterImageReader->SetDataExtent( p_dims[0], p_dims[1], 
                                     p_dims[2], p_dims[3], 
                                     p_dims[4], p_dims[5] );
  //p_waterImageReader->Update();  // this will force a read of all the slices

  //
  // setup the streamer
  //
  p_waterImageStreamer->SetInputConnection( p_waterImageReader->GetOutputPort() );
  p_waterImageStreamer->GetExtentTranslator()->SetSplitModeToZSlab();
  p_waterImageStreamer->SetNumberOfStreamDivisions( 4 );
  p_waterImageStreamer->UpdateInformation();

  p_waterSliceExtractor->SetInputConnection( p_waterImageStreamer->GetOutputPort() );
  p_waterSliceExtractor->SetOutputWholeExtent( p_dims[0], p_dims[1], 
                                               p_dims[2], p_dims[3], 0, 0 );
*/
  vtkSmartPointer< vtkImageAccumulate > accum = vtkSmartPointer< vtkImageAccumulate >:: New();
  accum->SetInput( p_waterSliceExtractor->GetOutput() );
  accum->Update();
  double max[3];  accum->GetMax( max );
  cout << "max label: " << max[0] << endl ; 
  p_numLabels = max[0]*2;
  if( p_randomTransferFx ) p_randomTransferFx->Delete();
  p_randomTransferFx = vtkColorTransferFunction::New();
  srand( time(NULL) ); //1 );
  double randH = 1;
  double randR, randG, randB = 1;
  //cout << "randH: " ;
  for (int i = 0; i < p_numLabels; i++)
  {
     randH = (double)rand() / (double)RAND_MAX;
     //cout << randH << " " ;
     p_randomTransferFx->AddHSVPoint(i, randH, 1, 1 );
     //randR = (double)rand() / (double)RAND_MAX;
     //randG = (double)rand() / (double)RAND_MAX;
     //randB = (double)rand() / (double)RAND_MAX;
     //p_randomTransferFx->AddHSVPoint(i, randR, randG, randB);
  }
  //cout << endl;

  if ( p_waterColorMap ) p_waterColorMap->Delete();
  p_waterColorMap = vtkImageMapToColors::New();
  p_waterColorMap->SetInputConnection( p_waterSliceExtractor->GetOutputPort() );
  p_waterColorMap->SetLookupTable( p_randomTransferFx );
  p_waterColorMap->Update();

}


void SectionEditor::loadImageData( vector< string > filenames, 
                                   vtkPNGReader* reader,
                                   vtkImageDataStreamer* streamer,
                                   vtkImageClip* extractor,
                                   bool calcDims )
{
  //
  //  read in the file names
  //
  vtkStringArray *allfilenames = vtkStringArray::New();
  for ( unsigned int i = 0; i < filenames.size(); i++ )
  {
     allfilenames->InsertNextValue( filenames[i].c_str() );
     cout << allfilenames->GetValue( i ) << endl;
  }

  //
  //  get the image extent, without reading the image - not sure how to
  //  do this in vtk.  this works for libpng 1.2.5
  //
  if( calcDims )  // only need to do this once
  {
     FILE* fp = fopen( allfilenames->GetValue( 0 ), "r" );
     png_structp png_ptr;
     png_infop info_ptr;
     png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
     info_ptr = png_create_info_struct(png_ptr);
     png_init_io(png_ptr, fp);
     png_read_info(png_ptr, info_ptr);
     int width = info_ptr->width;
     int height = info_ptr->height;
     //long width, height;
     //int t = readpng_init( fp, &width, &height );
     //cout << width << " " << height << endl ;
     p_dims[0] = p_dims[2] = p_dims[4] = 0;
     p_dims[1] = width - 1;
     p_dims[3] = height - 1;
     p_dims[5] = filenames.size() - 1;
     fclose( fp );
     //delete fp;
     p_minSection = 0;
     p_maxSection = p_dims[5];
  }

  //
  //  read in the files
  //
  reader->SetFileNames( allfilenames );
  reader->SetDataExtent( p_dims[0], p_dims[1], 
                         p_dims[2], p_dims[3], 
                         p_dims[4], p_dims[5] );
  reader->SetDataSpacing( p_spacing );
  //reader->Update();  // this will force a read of all the slices!!!!

  //
  // setup the streamer
  //
  streamer->SetInputConnection( reader->GetOutputPort() );
  streamer->GetExtentTranslator()->SetSplitModeToZSlab();
  streamer->SetNumberOfStreamDivisions( 4 );
  streamer->UpdateInformation();

  extractor->SetInputConnection( streamer->GetOutputPort() );
  extractor->SetOutputWholeExtent( p_dims[0], p_dims[1], 
                                   p_dims[2], p_dims[3], 0, 0 );

}

//
//  load in the isosurfaces
//
void SectionEditor::loadVTKFiles( string fn, bool addActors )
{
   // let's just see if we have a file to open?
   //string wholepathfile = p_vtkPath + "/paths.xml";
   p_pathFile = fn;
   cout << "whole: " << p_pathFile << endl ;
   FILE* fpr = fopen( fn.c_str(), "r" );
   //ifstream pathfile( wholepathfile.c_str() );
   if( fpr == NULL )
   {
      cout << "FYI: no path file to open" << endl ;
   }
   else
   {
      fclose( fpr );
      // read in the XML file
      //p_listOfPaths.readXMLPaths( p_pathFile.c_str() );
      ListOfPaths tmpLOP;
      tmpLOP.readXMLPaths( p_pathFile.c_str() );

      // read in only paths with vtk files.
      for( unsigned int ii = 0; ii < tmpLOP.size(); ii++ )
      {
         Path* PP = tmpLOP.getPathByIndex( ii ); 
         string wholevtkfile = p_vtkPath + "/" + PP->getVTKFilenames()[0];
         ifstream fptest; fptest.open( wholevtkfile.c_str() );
         if( fptest ) 
         {  cout << "loading " << PP->getID() << endl;
            p_listOfPaths.addPath( *PP );
         }
         fptest.close();
      }

      if( p_listOfPaths.size() == 0 )
      {
         cout << "no paths loaded, returning " << endl ;
         return;
      }

      // get the last pathid
      //p_pathInc = p_listOfPaths.get( p_listOfPaths.size()-1 ).getID() ;
      vector< int > path_ids;

      // read in all the vtk files  - we only have one now :-P
      for( unsigned int i = 0; i < p_listOfPaths.size(); i++ )
      {
         string vtkfile = p_listOfPaths.get( i ).getVTKFilenames()[0];
         int cur_id = p_listOfPaths.get( i ).getID();
         path_ids.push_back( cur_id );

         //if( addActors )
         if( addActors && ( vtkfile.size() > 0 ) )
         {
            string wholevtkfile = p_vtkPath + "/" + vtkfile;
            cout << "reading " << cur_id << ": " << wholevtkfile << endl ;

            vtkPolyDataReader* reader = vtkPolyDataReader::New();
            reader->SetFileName( wholevtkfile.c_str() );

            //vtkPolyDataNormals* norms = vtkPolyDataNormals::New();
            //norms->SetInput( reader->GetOutput() );

            vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
            //mapper->SetInputConnection( norms->GetOutputPort() );
            mapper->SetInputConnection( reader->GetOutputPort() );
            mapper->ScalarVisibilityOff();

            vtkActor* actor = vtkActor::New();
            actor->SetMapper( mapper );
            // get a color
            double r, g, b;
            randColor( r, g, b);
            //cout << "color: " << r << " " << g << " " << b << endl ;
            actor->GetProperty()->SetColor( r, g, b );
            actor->SetVisibility( false );

            // add sphere actors for each path
            vector< int > all_slices = p_listOfPaths.get( i).getSlices();
            vector< int > all_x = p_listOfPaths.get( i).getXs();
            vector< int > all_y = p_listOfPaths.get( i).getYs();
            vector< vtkActor* > sphereActors;
            for( unsigned int s = 0; s < all_slices.size(); s++ )
            {
               vtkSphereSource* sphere = vtkSphereSource::New();
               //cout << "adding " << all_x[s] << " " <<  all_y[s] << " "
               //   << all_slices[s] << endl ;
               sphere->SetCenter( all_x[s], p_dims[3]-all_y[s], all_slices[s]*p_spacing[2] );
               sphere->SetRadius( 10 );

               vtkPolyDataMapper* polySphere = vtkPolyDataMapper::New();
               polySphere->SetInputConnection( sphere->GetOutputPort() );

               vtkActor* sphereActor = vtkActor::New();
               sphereActor->SetMapper( polySphere );
               sphereActor->GetProperty()->SetColor( r, g, b );
               sphereActor->SetVisibility( false );

               p_ren->AddActor( sphereActor );

               sphereActors.push_back( sphereActor );

            }
            p_isoSphereActors[ cur_id ] = sphereActors ;
            

            p_renWater->AddActor( actor );
            //p_ren->AddActor( actor ); // debug
            p_isoActors[ cur_id ] = actor;
         }
      }

      // figure out the new p_pathInc
      sort( path_ids.begin(), path_ids.end() );
      //int last_id = path_ids[ path_ids.size() -1];
      int max_id = *(max_element( path_ids.begin(), path_ids.end() ));
      p_pathInc = max_id;

   }
}

//
//  end of image reading methods
// ------------------------------

void SectionEditor::updateSegColorData( vtkImageData* image )
{
   /*
   if( !p_segImageData )
      p_segImageData = vtkImageData::New();

   p_segImageData->DeepCopy( image );
   p_segImageData->SetNumberOfScalarComponents(4);
   p_segImageData->Update();


   //  change the color of the segdata before the overlay
   vtkUnsignedCharArray *colorData = vtkUnsignedCharArray::New();
   vtkDataArray *tmpData = p_segImageData->GetPointData()->GetScalars();
   vtkUnsignedCharArray *origData = dynamic_cast<vtkUnsignedCharArray *>( tmpData );
   if( origData != NULL )
   {
     colorData->SetNumberOfComponents( 4 );
     colorData->SetNumberOfTuples( origData->GetNumberOfTuples() );
     unsigned char pixelData[4] = { 0, 255, 0, 0 };
     for( vtkIdType i=0;i < colorData->GetNumberOfTuples(); i++ )
     {
       unsigned char scalarVal = origData->GetValue(i);
       pixelData[3] = scalarVal;
       //cout << "scalarVal: " << tmpData->GetValue(i) << endl ;
       colorData->SetTupleValue(i, pixelData);
     }
     p_segImageData->GetPointData()->SetScalars(colorData);
     //origData->Delete();
   }else{
     std::cerr << "Error, original data not in unsigned char format after cast??" << std::endl;
   }

   p_segImageData->Update();
   */
}


void SectionEditor::createVTKPipeline()
{
   p_rawSectionActor = vtkImageActor::New();
   p_rawSectionActor->SetInput( p_rawSliceExtractor->GetOutput() );

   p_rawSectionActor2 = vtkImageActor::New();
   p_rawSectionActor2->SetInput( p_rawSliceExtractor->GetOutput() );

   if( p_segEditSliceExtractor )
   {
      blend();
      p_segColorMap->SetInputConnection( p_segImageBlend->GetOutputPort() );
   }
   else
      p_segColorMap->SetInputConnection( p_segSliceExtractor->GetOutputPort() );


   p_segSectionActor = vtkImageActor::New();
   p_segSectionActor->SetInput( p_segColorMap->GetOutput() );
   p_segSectionActor->SetOpacity( .3 );
   //p_segSectionActor->AddPosition( 0.0, 0.0, 10 );
   //double pos[3];
   //p_segSectionActor->GetPosition(pos);
   ////pos[2] += 10.0;
   //p_segSectionActor->SetPosition(pos);

   //p_segEditSectionActor = vtkImageActor::New();
   //p_segEditSectionActor->SetInput( p_segEditSliceExtractor->GetOutput() );
   //p_segEditSectionActor->SetInput( p_segEditColorMap->GetOutput() );
   //p_segEditSectionActor->SetOpacity( .3 );

   p_ren->AddActor( p_rawSectionActor );
   p_renWater->AddActor( p_rawSectionActor2 );
   p_ren->AddActor( p_segSectionActor );
   p_ren->AddActor( p_segEditSectionActor );

   if ( p_waterSliceExtractor )
   {
      if ( p_waterSectionActor ) p_waterSectionActor->Delete();
      p_waterSectionActor = vtkImageActor::New();
      //p_waterSectionActor->SetInput( p_waterSliceExtractor->GetOutput() );
      p_waterSectionActor->SetInput( p_waterColorMap->GetOutput() );
      //double wpos[3];
      //p_waterSectionActor->GetPosition( wpos );
      //p_waterSectionActor->SetPosition( wpos );
      //p_waterSectionActor->AddPosition( 0.0, 0.0, 20 );
      p_waterSectionActor->SetOpacity(.5);  

      p_ren->AddActor( p_waterSectionActor );
   }


   p_ren->ResetCamera();
   p_renWater->ResetCamera();
}

void SectionEditor::initializeLiveWire()
{
   //if( !p_liveWireEdit )
   p_liveWireEdit = new vtkImageEditor( p_ren ); 
}

void SectionEditor::updateLiveWire()
{
   vector< int > extent(6);
   int ex[6];
   //p_rawSliceExtractor->GetOutputWholeExtent( ex );
   p_segEditSliceExtractor->GetOutputWholeExtent( ex );
   cout << "extent: " ;
   for(int i = 0; i < 6; i++ )
   {
      extent[i] = ex[i];
      cout << extent[i] << " ";
   }
   cout << endl ;

   string filename = p_segEditImageReader->GetFileNames()->GetValue( extent[5] );
   cout << "filename: " << filename << endl ;
/*
   vtkImageData* img = vtkImageData::New();
   img->DeepCopy( p_rawSliceExtractor->GetOutput() );
   img->Update();
   vtkImageData* img_edit = vtkImageData::New();
   img_edit->DeepCopy( p_segEditSliceExtractor->GetOutput() );
   img_edit->Update();
 */  
   //p_liveWireEdit->loadRawImage( img, img_edit, extent );
   //p_liveWireEdit->loadRawImage( p_rawSliceExtractor->GetOutput(), 
   //                              p_segEditSliceExtractor->GetOutput(), 
   //                              extent );
   p_liveWireEdit->loadData( extent, p_spacing[2] );
   p_liveWireEdit->setFilename( filename );

   p_liveWireEdit->setLiveWireBuffer( 500 );

//   img->Delete();
//   img_edit->Delete();
//   allfilenames->Delete();
   
}

#define ADD_ONLY 1
//#define SUB_ALSO 1
void SectionEditor::blend()
{
   cout << "blending" << endl ;
#ifdef ADD_ONLY
   p_segImageBlend = vtkImageMathematics::New();
   p_segImageBlend->SetOperationToAdd();
   //p_segImageBlend->SetOperationToMax();
   p_segImageBlend->SetInput1( p_segSliceExtractor->GetOutput() );
   p_segImageBlend->SetInput2( p_segEditSliceExtractor->GetOutput() );
#endif
#ifdef SUB_ALSO
   //
   // isolate the pixels we want to erase
   //
   //vtkImageThreshold* p_thresh_sub = vtkImageThreshold::New();
   p_thresh_sub = vtkImageThreshold::New();
   p_thresh_sub->SetInputConnection( p_segEditSliceExtractor->GetOutputPort() );
   p_thresh_sub->ThresholdByLower( 129.0 );
   p_thresh_sub->SetReplaceIn( 0 );
   p_thresh_sub->SetReplaceOut( 255 );

   //
   // subtract the pixels we want to erase from the main image
   //
   //vtkImageMathematics* segImageBlend = vtkImageMathematics::New();
   p_segImageBlend1 = vtkImageMathematics::New();
   p_segImageBlend1->SetOperationToSubtract();
   p_segImageBlend1->SetInput1( p_segSliceExtractor->GetOutput() );
   p_segImageBlend1->SetInput2( p_thresh_sub->GetOutput() );

   //
   // threshold this output to get rid of outliers
   //
   //vtkImageThreshold* thresh = vtkImageThreshold::New();
   p_thresh = vtkImageThreshold::New();
   p_thresh->SetInputConnection( p_segImageBlend1->GetOutputPort() );
   //thresh_sub->ThresholdBetween( 127.0, 129.0 );
   p_thresh->ThresholdByLower( 250 );
   p_thresh->SetReplaceIn( 255 );
   p_thresh->SetReplaceOut( 0 );

   //
   // isolate the pixels we want to add back to the image
   //
   //vtkImageThreshold* thresh_add = vtkImageThreshold::New();
   p_thresh_add = vtkImageThreshold::New();
   p_thresh_add->SetInputConnection( p_segEditSliceExtractor->GetOutputPort() );
   p_thresh_add->ThresholdByLower( 129.0 );
   p_thresh_add->SetReplaceIn( 255 );
   p_thresh_add->SetReplaceOut( 0 );

   p_segImageBlend = vtkImageMathematics::New();
   p_segImageBlend->SetOperationToMax();
   p_segImageBlend->SetInput1( p_thresh_add->GetOutput() );
   p_segImageBlend->SetInput2( p_thresh->GetOutput() );

      cout << "saving test.png" << endl ;
      vtkPNGWriter* Writer = vtkPNGWriter::New();
      Writer->SetFileName( "test.png" );
      Writer->SetInputConnection( p_thresh_add->GetOutputPort() );
      Writer->Write();
#endif
//exit(1);
}

void SectionEditor::createVTKPipelineForWatershed()
{
   // Setup the watershed data volume pipeline
   /*
   if( p_waterImageData )
   {
      p_ren->RemoveActor( p_waterSectionActor );

      if( p_waterSliceExtractor ) p_waterSliceExtractor->Delete();
      p_waterSliceExtractor = vtkExtractVOI::New();
      p_waterSliceExtractor->SetInput(p_waterImageData);
      p_waterSliceExtractor->SetVOI(p_dims);
      p_waterSliceExtractor->Update();

      if( p_satLut ) p_satLut->Delete();
      p_satLut = vtkLookupTable::New();
      p_satLut->SetTableRange (0, p_numLabels);
      p_satLut->SetHueRange (0, 1);
      p_satLut->SetSaturationRange (1, 1);
      p_satLut->SetValueRange (1, 1);
      p_satLut->Build(); //effective built

      if( p_randomTransferFx ) p_randomTransferFx->Delete();
      p_randomTransferFx =
         vtkColorTransferFunction::New();
      srand( 1 );
      double randH = 1;
      //cout << "randH: " ;
      for (int i = 0; i < p_numLabels; i++)
      {
         randH = (double)rand() / (double)RAND_MAX;
         //cout << randH << " " ;
         p_randomTransferFx->AddHSVPoint(i, randH, 1, 1 );
      }
      cout << endl;

      if ( p_waterColorMap ) p_waterColorMap->Delete();
      p_waterColorMap = vtkImageMapToColors::New();
      p_waterColorMap->SetInputConnection( p_waterSliceExtractor->GetOutputPort() );
      p_waterColorMap->SetLookupTable( p_randomTransferFx );
      //p_waterColorMap->SetLookupTable( p_satLut );
      p_waterColorMap->Update();

      //p_waterShiftScale = vtkImageShiftScale::New();
      //p_waterShiftScale->SetInputConnection( p_rawSliceExtractor->GetOutputPort() );
      //p_waterShiftScale->SetOutputScalarTypeToUnsignedChar();
      //p_waterShiftScale->Update();

      if ( p_waterSectionActor ) p_waterSectionActor->Delete();
      p_waterSectionActor = vtkImageActor::New();
      p_waterSectionActor->SetInput( p_waterColorMap->GetOutput() );
      double wpos[3];
      p_waterSectionActor->GetPosition( wpos );
      //wpos[2] += 100.0;
      p_waterSectionActor->SetPosition( wpos );
      p_waterSectionActor->SetOpacity(.5);  

      p_ren->AddActor( p_waterSectionActor );
      p_ren->ResetCamera();
   }

   p_ren->ResetCamera();
   updateRenderWindow();
   */
}

void SectionEditor::applyThreshold( int s )
{
   applyFilter( THRESHOLD, s );
}

void SectionEditor::applyBlur( int s )
{
   applyFilter( BLUR, s );
}


void SectionEditor::applyFilter( int type, int s )
{
   /*
   int mins = p_minSection;
   int maxs = p_maxSection;
   if ( s != -1 ) // do just one slice
   {
      mins = s;
      maxs = s+1;
   }

   int region[6];
   region[0] = p_dims[0];
   region[1] = p_dims[1];
   region[2] = p_dims[2];
   region[3] = p_dims[3];
   region[4] = region[5] = 0; // default for now


   vtkImageData* threshImage = vtkImageData::New();
   threshImage->DeepCopy( p_segImageDataFloat );
   threshImage->Update();

   float maxInt = 255;
   float threshold = maxInt * p_threshold;
   float blurRadius = 2* p_blurSigma;
   cout << "threshold: " << threshold << endl ;
   cout << "blur radius: " << blurRadius << endl ; 

   vtkImageShiftScale *shiftScale = vtkImageShiftScale::New();
   shiftScale->SetInput( threshImage );
   shiftScale->SetScale( maxInt );
   shiftScale->SetOutputScalarTypeToUnsignedChar();
   shiftScale->Update();
   threshImage = shiftScale->GetOutput();

   vtkExtractVOI* extractSlice = vtkExtractVOI::New();
   extractSlice->SetInput( threshImage );
   //extractSlice->SetInputConnection( shiftScale->GetOutputPort() );

   //cout << "dims: " << threshImage->GetDataDimension() << endl ;

   for( int s = mins; s < maxs; s++ )
   {
      cout << "s: " << s << endl ;
      region[4] = region[5] = s; 
      extractSlice->SetVOI( region );
      extractSlice->Update();

      vtkImageThreshold* ti = vtkImageThreshold::New();
      vtkImageGaussianSmooth* si = vtkImageGaussianSmooth::New();

      switch ( type )
      {
         case THRESHOLD:
            ti->SetInputConnection( extractSlice->GetOutputPort() );
            ti->ThresholdBetween( threshold, 255 );
            ti->SetReplaceIn( true );
            ti->SetReplaceOut( true );
            ti->SetInValue( 255 );
            ti->SetOutValue( 0 );
            ti->Update();
            break;
         case BLUR:
            si->SetInputConnection( extractSlice->GetOutputPort() );
            si->SetRadiusFactors( blurRadius, blurRadius);
            si->Update();
            break;
      }

      for(int y = region[2]; y < region[3]; y++)
      {
         for(int x = region[0]; x < region[1]; x++)
         {
            if( type == THRESHOLD )
               threshImage->SetScalarComponentFromFloat(x, y, s, 0,
                     ti->GetOutput()->GetScalarComponentAsFloat( x, y, s, 0 ) );
            else if( type == BLUR )
               threshImage->SetScalarComponentFromFloat(x, y, s, 0,
                     si->GetOutput()->GetScalarComponentAsFloat( x, y, s, 0 ) );
            else
               cout << "ERROR: No filter type selected in applyFilter." << endl ;
         }
      }

      ti->Delete();

   }

   extractSlice->Delete();

   cout << "done" << endl ;
   updateSegColorData( threshImage );

   //threshImage->Delete();
   //shiftScale->Delete();


   p_ren->GetRenderWindow()->Render();
   */

}

void SectionEditor::createFloodFill( int s )
{
   cout << "TODO!" << endl ;
}

void SectionEditor::createWatershedSegmentation( int s )
{
   /*
   //if ( p_waterImageData )
   //   p_waterImageData->Delete();

   int mins = p_minSection;
   int maxs = p_maxSection;
   if ( s != -1 ) // do just one slice
   {
      mins = s;
      maxs = s+1;
   }

   int region[6];
   region[0] = p_dims[0];
   region[1] = p_dims[1];
   region[2] = p_dims[2];
   region[3] = p_dims[3];
   region[4] = region[5] = 0; // default for now

   vtkExtractVOI* extractSlice = vtkExtractVOI::New();
   extractSlice->SetInput( p_segImageDataFloat );

   //int labelsum = 0;

   // make a new volume
   if( !p_waterImageData )
   {
      p_waterImageData = vtkImageData::New();
      p_waterImageData->SetSpacing( p_segImageDataFloat->GetSpacing() );
      p_waterImageData->SetExtent( p_segImageDataFloat->GetWholeExtent() );
      p_waterImageData->SetScalarTypeToFloat();
      p_waterImageData->SetNumberOfScalarComponents(1);
      p_waterImageData->AllocateScalars();
   }

   for( int s = mins; s < maxs; s++ )
   //for( int s = 0; s < 1; s++ )
   {
      cout << "watershed " << s << endl ;
      region[4] = region[5] = s;  // default for now
      extractSlice->SetVOI( region );
      extractSlice->Update();

      // not sure why, but I need to flip the image
      //cout << "flip" << endl ;
      //vtkImageFlip* flipy = vtkImageFlip::New();
      //flipy->SetInput( extractSlice->GetOutput() );
      //flipy->SetFilteredAxis( 1 );
      //flipy->FlipAboutOriginOn();
      //flipy->Update();

      cout << "converting" << endl ;

      itkImageType::Pointer itkSlice = ItkVtkConvert::ConvertToItkImage
                                               ( extractSlice->GetOutput());

      cout << "watering" << endl ;
      WatershedFilter watershedFilter( itkSlice, p_watershedLevel, p_watershedThreshold );
      int numlabels = watershedFilter.relabel();

      //labelsum += numlabels;
      if ( numlabels > p_numLabels )
         p_numLabels = numlabels;
      
      //cout << "watering" << endl;
      //WatershedFilterType::Pointer watershed = WatershedFilterType::New();
      //watershed->SetLevel( .5 );
      //watershed->SetThreshold( .2 );
      //watershed->SetInput( itkSlice );
      //watershed->Update();

      //CastImageFilter::Pointer castImage = CastImageFilter::New();
      //castImage->SetInput( watershed->GetOutput() );
      //castImage->Update();
      //vtkImageData* vtkSlice = ItkVtkConvert::ConvertToVtkImage( castImage->GetOutput() );

      vtkImageData* vtkSlice = ItkVtkConvert::ConvertToVtkImage( watershedFilter.getImage() );

      cout << "copying" << endl ;
      // copy slice into final volume
      for(int y = region[2]; y < region[3]; y++)
      {
         for(int x = region[0]; x < region[1]; x++)
         {
            p_waterImageData->SetScalarComponentFromFloat(x, y, s, 0,
                  vtkSlice->GetScalarComponentAsFloat( x, y, 0, 0 ) );
         }
      }

      //p_numLabels = labesum / p_maxSection;
      //flipy->Delete();
   }

   extractSlice->Delete();

   vtkMetaImageWriter *vtkVolWriter = vtkMetaImageWriter::New();
   vtkVolWriter->SetInput( p_waterImageData );//vtkSlice );
   string wf = "vtkWaterImageData.mha";
   vtkVolWriter->SetFileName( wf.c_str() );
   vtkVolWriter->Write();
   cout << "watershed written to " << wf << endl ;
   vtkVolWriter->Delete();

   createVTKPipelineForWatershed();
   */
}

string SectionEditor::isosurfaceSeg( vtkImageData* imgdata, float val )
{
   cout << "smoothing" << endl ;
   vtkSmartPointer< vtkImageGaussianSmooth > smooth = 
      vtkSmartPointer< vtkImageGaussianSmooth >::New();
   smooth->SetInput( imgdata );
   //smooth->SetRadiusFactors( 1.5, 1.5, 1.5 );
   smooth->SetRadiusFactors( 2, 2, 4 );
   smooth->SetStandardDeviation( 1, 1, 1.5 );
   smooth->Update();

   //vtkMetaImageWriter *vtkVolWriter = vtkMetaImageWriter::New();
   //vtkVolWriter->SetInput( smooth->GetOutput() );
   //vtkVolWriter->SetFileName( "smooth.mha" );
   //vtkVolWriter->Write();

   //for( int p = 1; p <= numpaths; p++ )
   //{
      //double isoval = p - .5; 
      double isoval = .5;//20; 
      cout << "marching " << endl;
      vtkSmartPointer< vtkImageMarchingCubes > iso = 
         vtkSmartPointer< vtkImageMarchingCubes >::New();
      iso->SetInput( smooth->GetOutput() );
      iso->SetValue( 0, isoval );
      iso->Update();

      vtkSmartPointer< vtkPolyData > polydata = vtkSmartPointer< vtkPolyData >::New();
      polydata->DeepCopy( iso->GetOutput() );

      cout << "mapping" << endl ;
      vtkSmartPointer< vtkPolyDataMapper > mapper = 
         vtkSmartPointer< vtkPolyDataMapper >::New();
      //mapper->SetInput( iso->GetOutput() );
      mapper->SetInput( polydata );
      mapper->SetScalarVisibility( false );

      //
      //  write it to a vtk file for use later and in other applications
      //
      vtkSmartPointer<vtkTriangleFilter> tri =
         vtkSmartPointer<vtkTriangleFilter>::New();
      tri->SetInput( iso->GetOutput() );

      vtkSmartPointer<vtkPolyDataNormals> norms =
         vtkSmartPointer<vtkPolyDataNormals>::New();
      norms->SetInput( tri->GetOutput() );
      norms->SetFeatureAngle( 180.0 );

      vtkSmartPointer<vtkPolyDataWriter> vtkVolWriter =
         vtkSmartPointer<vtkPolyDataWriter>::New();
      //vtkVolWriter->SetInput( mapper->GetInput() );//vtkSlice );
      vtkVolWriter->SetInput( norms->GetOutput() );//vtkSlice );
      char tmp[50];
      sprintf( tmp, "%03i", p_pathInc );
      string incstr( tmp );
      string vtkfile = incstr + ".vtk";
      string vtkpath = p_vtkPath + "/" + vtkfile;
      vtkVolWriter->SetFileName( vtkpath.c_str() );
      vtkVolWriter->Write();
      cout << "vtk volume written to " << vtkpath << endl ;
      //vtkVolWriter->Delete();

      double randR = (double)rand() / (double)RAND_MAX;
      double randG = (double)rand() / (double)RAND_MAX;
      double randB = (double)rand() / (double)RAND_MAX;
      //cout << "acting " << randR << " " << randG << " " << randB << endl ;
      vtkActor* isoActor = vtkActor::New();
      isoActor->SetMapper( mapper );
      isoActor->GetProperty()->SetColor( randR, randG, randB );
      //isoActor->GetProperty()->SetOpacity( .5 );  // need to sort for
      //this to work properly!

      p_renWater->AddActor( isoActor ) ;
      isoActor->SetVisibility( false );
      p_isoActors[ p_pathInc ] = isoActor;
   //}

   cout << "rendering " << endl ;
   p_renWater->ResetCameraClippingRange();
   p_renWater->GetRenderWindow()->Render();
   cout << "done" << endl ;

   return vtkfile;

}

void SectionEditor::randColor( double& r, double& g, double& b )
{
   r = (double)rand() / (double)RAND_MAX;
   g = (double)rand() / (double)RAND_MAX;
   b = (double)rand() / (double)RAND_MAX;
}

/*
void SectionEditor::goToSection( int mx, int my )
{
      vtkWorldPointPicker* picker = vtkWorldPointPicker::New();
      picker->Pick( mx, my, 0, p_renWater );
      double pickCoords[3];
      picker->GetPickPosition( pickCoords);
      cout << "picked: " ;
      for( int p = 0; p < 3; p++ )
         cout << pickCoords[p] << " " ;
      cout << endl;
     
      int slice = pickCoords[2]/p_spacing[2] ;
      cout << "going to slice " << slice << endl ;
      showSection( slice );
}
*/


void SectionEditor::setWatershedSegmentation( int mx, int my )
{
   if( p_waterSliceExtractor )
   {
      // do we want to record this picked point in our path???
      //cout << "p_pick: " << p_pick << endl ;
      if ( !p_pick )
         return;

      vtkWorldPointPicker* picker = vtkWorldPointPicker::New();
      picker->Pick( mx, my, 0, p_ren );
      double pickCoords[3];
      picker->GetPickPosition( pickCoords);
      cout << "picked: " ;
      for( int p = 0; p < 3; p++ )
         cout << pickCoords[p] << " " ;
      cout << endl;

      addSphere( pickCoords[0], pickCoords[1], pickCoords[2] );

      double* spacing = getSpacing();
      int* extent = getExtent();
      cout << "extent: " ;
      for( int p = 0; p < 6; p++ )
         cout << extent[p] << " " ;
      cout << endl;
      cout << "spacing: " ;
      for( int p = 0; p < 3; p++ )
         cout << spacing[p] << " " ;
      cout << endl;

      int imgCoords[3]; // TODO: subtract origin, also
      imgCoords[0] = pickCoords[0] / spacing[0];
      imgCoords[1] = pickCoords[1] / spacing[1];
      imgCoords[2] = p_curSection;
      cout << "image coords: " ;
      for( int p = 0; p < 3; p++ )
         cout << imgCoords[p] << " " ;
      cout << endl;

      unsigned short label = getPixelValue( imgCoords[0], imgCoords[1], imgCoords[2]);

      if ( label != -1 )
         cout << "label: " << label << endl;


      int x = imgCoords[0];
      int y = p_dims[3] - imgCoords[1]; // need to flip the Y.  why? I dont know why...
      //int y = imgCoords[1]; 
      if ( p_path == NULL ) //p_path->size() == 0 )
      {
         p_path = new Path();
         p_path->initPath( p_curSection, label, 0, 0, x, y );
      }
      else
      {
         p_path->add( p_curSection, label, 0, 0, x, y );
      }

      p_path->print();

      cout << "---- " << endl ;
   }
}

void SectionEditor::addSphere( double x, double y, double z )
{
   cout << "adding sphere " << endl ;
   vtkSmartPointer< vtkSphereSource > sphere = vtkSmartPointer< vtkSphereSource >::New();
   sphere->SetCenter( x, y, z );
   sphere->SetRadius( 10 );

   vtkSmartPointer< vtkPolyDataMapper > polyDataMapper = vtkSmartPointer< vtkPolyDataMapper >::New();
   polyDataMapper->SetInputConnection( sphere->GetOutputPort() );

   vtkSmartPointer< vtkActor > sphereActor = vtkSmartPointer< vtkActor >::New();
   sphereActor->SetMapper( polyDataMapper );

   p_ren->AddActor( sphereActor );

   // keep track of these to delete them later
   p_sphereSources.push_back( sphere );
   p_sphereMappers.push_back( polyDataMapper );
   p_sphereActors.push_back( sphereActor );

}

void SectionEditor::removeSpheres()
{
   cout << "removing spheres" << endl;

   for( unsigned int i = 0; i < p_sphereActors.size(); i++ )
      p_ren->RemoveActor( p_sphereActors[i] );

   // removing spheres
   p_sphereSources.clear();
   p_sphereMappers.clear();
   p_sphereActors.clear();

   p_ren->GetRenderWindow()->Render();

}

unsigned short SectionEditor::getPixelValue( int x, int y, int z )
{
   unsigned short pixel = -1;
   if( ( x >= p_dims[0] ) && ( x < p_dims[1] ) &&
       ( y >= p_dims[2] ) && ( y < p_dims[3] ) &&
       ( z == p_curSection ) )
   {
      int ijk[3];
      ijk[0] = x;
      ijk[1] = y;
      ijk[2] = z;

      vtkIdType id = p_waterSliceExtractor->GetOutput()->ComputePointId( ijk );

      vtkDataArray* data = p_waterSliceExtractor->GetOutput()->GetPointData()->GetScalars();

      vtkUnsignedShortArray* udata = dynamic_cast< vtkUnsignedShortArray* >( data );
      if( udata != NULL )
      {
         pixel = udata->GetValue( id );
      }
      cout << "pixel: " << pixel << endl;
   }
   return pixel;
}
  
void SectionEditor::setupInteractor()
{
   //vtkInteractorStyleImage *imageStyle = vtkInteractorStyleImage::New();
   vtkInteractorStyleTrackballCamera *imageStyle = vtkInteractorStyleTrackballCamera::New();
   vtkInteractorStyleTrackballCamera *imageStyleWater = vtkInteractorStyleTrackballCamera::New();

   //
   // key press event
   //

   vtkCallbackCommand *myCallbackCommand = vtkCallbackCommand::New();
   myCallbackCommand->SetCallback( charCallback );
   myCallbackCommand->SetClientData( (void*)this );

   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::New();
   interactor->SetInteractorStyle(imageStyle);
   interactor->AddObserver( vtkCommand::KeyPressEvent, myCallbackCommand );

   vtkRenderWindowInteractor *interactorWater = vtkRenderWindowInteractor::New();
   interactorWater->SetInteractorStyle(imageStyleWater);

   //
   //  pick callback
   //
   addPickCallback( interactor );

   p_ren->GetRenderWindow()->SetInteractor(interactor);

   p_renWater->GetRenderWindow()->SetInteractor(interactorWater);

}

void SectionEditor::addPickCallback( vtkRenderWindowInteractor *interactor )
{
   cout << "adding callbacks " << endl ;
   //
   //  pick callback for waterhsed
   //
   vtkCallbackCommand *pickCallbackCommand = vtkCallbackCommand::New();
   pickCallbackCommand->SetCallback( pickWatershedSegmentation );
   pickCallbackCommand->SetClientData( (void*)this );

   interactor->AddObserver( vtkCommand::LeftButtonPressEvent, pickCallbackCommand );

   /* 
   vtkCallbackCommand *pickSliceCommand = vtkCallbackCommand::New();
   pickSliceCommand->SetCallback( pickSliceCallback );
   pickSliceCommand->SetClientData( (void*)this );

   vtkRenderWindowInteractor* waterInteractor =
      p_renWater->GetRenderWindow()->GetInteractor();
   waterInteractor->AddObserver( vtkCommand::LeftButtonPressEvent, pickSliceCommand );
   */ 


}

void SectionEditor::show()
{
   // set up the main window for viewing the segmentation
   vtkRenderWindow *window = vtkRenderWindow::New();
   window->AddRenderer( p_ren );

   // main window for looking at the watershed segmentation
   vtkRenderWindow *windowWater = vtkRenderWindow::New();
   windowWater->AddRenderer( p_renWater );

   setupInteractor();

   window->Render();
   windowWater->Render();

   p_ren->GetRenderWindow()->GetInteractor()->Start();
   p_renWater->GetRenderWindow()->GetInteractor()->Start();

}

// -------------------------------
//
//       tracking methods
//
// -------------------------------

void SectionEditor::setSTrANDImageData()
{
   //
   //  convert vtk images into itk images to process
   //
   /*
   cout << "converting data" << endl ;
   itkImage3DType::Pointer itkWaterImageData = 
      ItkVtkConvert::ConvertToItkImage3D( p_waterImageData );
   itkImage3DType::Pointer itkRawImageData = 
      ItkVtkConvert::ConvertToItkImage3D( p_rawImageData );

   cout << "done converting data" << endl ;

   p_strand->setImageData( itkWaterImageData, itkRawImageData );
   */
}


//
//  merge region methods
//

void SectionEditor::openMergeRegions( const string& filename )
{
   //p_strand->readMergeRegions( filename );
}


void SectionEditor::mergeRegions( float thresh )
{
   cout << "Finding regions to merge " << endl ;

   setSTrANDImageData();

   //p_strand->findRegionsToMerge( thresh );

}

//
//  find region correlations 
//

void SectionEditor::correlateRegions()
{
   /*
   if( p_strand->doneMerging() )
   {
      if( !p_strand->gotImages() )
         setSTrANDImageData();

      int slices_to_skip = 1;
      p_strand->calcRegionCorrelations( slices_to_skip );

      //  write the files in case we need them later
      p_strand->printCorrelations( "corr/matrix_" );
      p_strand->printLabels( "corr/slice" );
   }
   else
      cout << "Need to merge regions first! " << endl ;
      */
}

void SectionEditor::openMatrixFiles( const vector< string >& matrices,
                                     const vector< string >& labels )
{
   //p_strand->readCorrelations( matrices, labels );
 
}

void SectionEditor::openPolyDataFiles( const vector< string >& files )
{
   cout << "Opening...." << endl ;
   for( unsigned int f = 0; f < files.size(); f++ )
   {
      cout << files[f] << endl ;
      vtkSmartPointer<vtkPolyDataReader> vtkDataReader = 
         vtkSmartPointer<vtkPolyDataReader>::New();
      vtkDataReader->SetFileName( files[f].c_str() );

      vtkPolyDataMapper* vtkDataMapper = vtkPolyDataMapper::New();
      vtkDataMapper->SetInput( vtkDataReader->GetOutput() );
      vtkDataMapper->SetScalarVisibility( false );

      vtkActor* actor = vtkActor::New();
      actor->SetMapper( vtkDataMapper );
      double randR = (double)rand() / (double)RAND_MAX;
      double randG = (double)rand() / (double)RAND_MAX;
      double randB = (double)rand() / (double)RAND_MAX;
      cout << "acting " << randR << " " << randG << " " << randB << endl ;
      actor->GetProperty()->SetColor( randR, randG, randB );

      p_renWater->AddActor( actor );
   }

   p_renWater->ResetCameraClippingRange();
   p_renWater->GetRenderWindow()->Render();
}

//
//  compute paths - do tracking
//

void SectionEditor::computePaths()
{
   /*
   bool use_dijkstra = true,
        skip_slices = false,
        be_greedy = true,
        use_minspanning = false;

   //p_strand->setLabels( ); //use all the labels?

   // get regions to start with
   p_path.print();
   vector< long int > labels;
   for( unsigned int p = 0; p < p_path.size(); ++p )
   {
      labels.push_back( p_path.getIndex( p ) );
      cout << p_path.getIndex( p ) << " ";
   }
   p_strand->setLabels( labels );
   p_path.clear();

   p_strand->findPaths( use_dijkstra,
                        skip_slices,
                        be_greedy,
                        use_minspanning );

   ListOfPaths paths = p_strand->getPaths();
   for( unsigned int p = 0; p < paths.size(); p++ )
   {
      cout << "processing path " << endl ;
      paths.getRef( p ).print();
      processPath( paths.getRef( p ) );
   }
   */
}

void SectionEditor::openPathFile( string file )
{
   cout << "opening " << file << endl ;
   p_listOfPaths.readPaths( file );
   for( unsigned int p = 0; p < p_listOfPaths.size(); p++ )
   {
      cout << "processing path " << endl ;
      p_listOfPaths.getRef( p ).print();
      processPath( p_listOfPaths.getPathByIndex( p ) );
   }
}

void SectionEditor::calcNextPath()
{
   cout << "calculating path" << endl ;
   int p = 0;
   //ListOfPaths listOfPaths;
   map< float, Path > allPaths;

   //
   // calculate all possibel paths through the graph
   //
   CalcPaths::calcPaths( p_allNodes, allPaths );

   if( allPaths.size() < 1 ) // no more paths to find, exit
   {
      cout << "no paths found" << endl ;
      return;
   }

   cout << "best path: " << allPaths.begin()->first << "... " ; 
   allPaths.begin()->second.print();
   allPaths.begin()->second.setID( p );

   //listOfPaths.addPath( allPaths.begin()->second );

   //
   // now, remove that path from the graph.
   //
   CalcPaths::removePath( allPaths.begin()->second, p_allNodes );

   //
   // render the path in the viewer
   //
   p_path = new Path( allPaths.begin()->second );
   processPath();

   //listOfPaths.writeXMLPaths( pathFilename.c_str() );

}

void SectionEditor::initCorrelation( const vector< string >& correlation_filenames )
{
   unsigned int s = (int) correlation_filenames.size();
   vector< unsigned int > inc(1); inc[0] = 1;
   cout << "got " << s << endl ;

   //
   //  read in all the correlations
   //
   CalcCorrelations::readCorrelations( p_allCorrespondences,
                                       correlation_filenames,
                                       0, 10, inc );
                                       //start, stop, inc );

   //
   //  build the graph
   //
   cout << "building the graph" << endl ;
   float dist_sigma = 50; // needs to be set by user
   CalcPaths::buildGraph( p_allCorrespondences, p_allNodes, dist_sigma );

   // get the minum number of paths to compute
   p_maxnumpaths = numeric_limits< unsigned int >::max();
   for( unsigned int q = 0; q < p_allNodes.size(); q++ )
      if( p_allNodes[q].size() < p_maxnumpaths )
         p_maxnumpaths = p_allNodes[q].size();

}

// ---------------------------------
//
//    END of Tracking methods
//
// ---------------------------------
/*
void pickSliceCallback(vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);
   if( interactor->GetRepeatCount() > 0 )
   {
      //cout << "count: " << interactor->GetRepeatCount() << endl ;
      //cout << "pick slice callback called" << endl ;
      SectionEditor* se = static_cast< SectionEditor* >( clientdata );

      if( interactor )
      {
         int mouseX, mouseY;
         interactor->GetEventPosition( mouseX, mouseY );
         //cout << "mouse: " << mouseX << ", " << mouseY << endl;
         se->goToSection( mouseX, mouseY );
      } 
   }

}
*/

void pickWatershedSegmentation(vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   //cout << "test pick callback called" << endl ;
   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);
   //char key = interactor->GetKeyCode();
   //cout << "key test: " << key << endl ;
   //if ( key == '^G' )
   //   cout << "got ctrl" << endl ;
   SectionEditor* se = static_cast< SectionEditor* >( clientdata );
   if( !se->getShowWaterFlag() )
      return;

   if( interactor )
   {
      int mouseX, mouseY;
      interactor->GetEventPosition( mouseX, mouseY );
      //cout << "mouse: " << mouseX << ", " << mouseY << endl;

      se->setWatershedSegmentation( mouseX, mouseY );
   } 

}

void charCallback(vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);
   SectionEditor* se = static_cast< SectionEditor* >( clientdata );
   if( interactor )
   {
      char key = interactor->GetKeyCode();
      cout << "key: " << key << endl ;
      switch( key )
      {
         case ',':  // left arrow
            se->showPrevSection();
            break;
         case '.':  // right arrow
            se->showNextSection();
            break;
         case 'p':
            se->processPath();
         case 'a':
            se->showRaw();
            break;
         case 's':
            se->showSeg();
            break;
         case 'w':
            se->showWater();
            break;
         default:
            break;
      }

   } 
}

