#ifndef SECTIONEDITOR_H
#define SECTIONEDITOR_H

#include <vector>
#include <string.h>

#include "vtkWindowLevelLookupTable.h"
#include "vtkImageData.h"
#include "Path.h"
#include "MetaReader.h"
#include "vtkImageEditor.h"
//#include "STrAND.h"
#include "Correspondence.h"
#include "Dijkstra.h"

// Forward class declarations
class vtkExtractVOI;
class vtkImageActor;
class vtkImageData;
class vtkImageShiftScale;
class vtkInteractorStyleImage;
class vtkMetaImageReader;
class vtkRenderWindowInteractor;
class vtkRenderer;
class vtkImageMapToWindowLevelColors;
class vtkImageMapToColors;
class vtkColorTransferFunction;
class vtkPNGReader;
class vtkImageClip;
class vtkImageDataStreamer;
class vtkImageMathematics;
class vtkSphereSource;
class vtkPolyDataMapper;
class vtkImageThreshold;

//class STrAND;

using namespace std;

void pickWatershedSegmentation(vtkObject* caller, unsigned long event, void *calldata, void *);
//void pickSliceCallback(vtkObject* caller, unsigned long event, void *calldata, void *);
void charCallback(vtkObject* caller, unsigned long event, void *calldata, void *);

class SectionEditor 
{

   public:

      // Constructor/Destructor
      SectionEditor();
      ~SectionEditor();

      //
      //  read in all the data
      //
      void loadRawImage( vector< string > filename );
      void loadSegImage( vector< string > filename );
      void loadSegEditImage( vector< string > filename );
      void loadWaterImage( vector< string > filename );
      void loadImageData( vector< string > filenames,
                          vtkPNGReader* reader,
                          vtkImageDataStreamer* streamer,
                          vtkImageClip* extractor,
                          bool calcDims = false);

      void setVTKPath( string path ) { p_vtkPath = path; };
      void loadVTKFiles( string file, bool addActors = true );

      void updateSegColorData( vtkImageData* image );

      void setWindow( int w, int o );

      vtkRenderer* getRenderer();
      vtkRenderer* get3DRenderer();

      void addPickCallback( vtkRenderWindowInteractor *interactor);
      void updateRenderWindow();
      //void goToSection( int x, int y );
      bool showSection( int i );
      void showNextSection() { showSection( p_curSection + 1 ); };
      void showPrevSection() { showSection( p_curSection - 1 ); };
      void processPath();
      string processPath( Path* path, int pathid = -1 ); 
      void rebuildPaths();
      void rebuildPath();
      void deletePaths( vector< int > deleteIds );
      void deletePathGeometry( int i );
      void mergePaths( int p1, int p2 );

      int getMinSection() { return p_minSection; };
      int getMaxSection() { return p_maxSection; };
      void createVTKPipeline();
      void initializeLiveWire();
      void updateLiveWire();
      string isosurfaceSeg( vtkImageData*, float iso = 1 );
      void show();
      void showRaw();
      void showSeg();
      void showWater();
      void showIso( vector< int > ids );
      void setPick() ;
      void setPickPath( int id ) ;
      void autoFillPath( int id );
      bool getShowWaterFlag() { return p_showWater; };
      void calcNextPath();
      void initCorrelation( const vector< string >& files );

      void setSpacing( vector< float > spacing )
      {
         if ( spacing.size() < 3 ) return;
         p_spacing[0] = spacing[0]; 
         p_spacing[1] = spacing[1];
         p_spacing[2] = spacing[2];
         cout << "spacing: " 
            << p_spacing[0] << ", " << p_spacing[1]
            << ", " << p_spacing[2] << endl;
      };
      unsigned short getPixelValue( int x, int y, int z );
      //double* getOrigin() { return p_waterImageData->GetOrigin(); };
      double* getSpacing() { return p_spacing ; };
      int* getExtent() { return p_dims; } ;

      void setDraw( bool draw = true ) 
      { p_liveWireEdit->setDraw( draw ); };

      //
      //  tracking methods
      //
      void mergeRegions( float edge_threshold );
      void correlateRegions();
      void computePaths();
      void openMergeRegions( const string& filename );
      void openMatrixFiles( const vector< string >& matrices,
                            const vector< string >& labels );
      void printPaths();
      void openPathFile( string file );
      void openPolyDataFiles( const vector< string >& files );

      const ListOfPaths* getPaths() { return &p_listOfPaths; } ;

      //
      //  flood fill methods
      //
      void createFloodFill( int s = -1 );

      //
      //  all watershed methods
      //
      void createWatershedSegmentation( int s = -1 );
      void createVTKPipelineForWatershed();
      float getWatershedLevel() { return p_watershedLevel; };
      float getWatershedThreshold() { return p_watershedThreshold; };
      string getWatershedLevelStr() { 
         stringstream s;  s << p_watershedLevel; return s.str(); };
      string getWatershedThresholdStr() {
         stringstream s;  s << p_watershedThreshold; return s.str(); };
      void setWatershedLevel( float level ) 
        { p_watershedLevel = level; };
      void setWatershedThreshold( float thresh ) 
        { p_watershedThreshold = thresh; };
      void setWatershedSegmentation( int mouseX, int mouseY );
      float getThreshold() { return p_threshold; };
      float getBlurSigma() { return p_blurSigma; };
      string getThresholdStr() { 
         stringstream s;  s << p_threshold; return s.str(); };
      string getBlurSigmaStr() { 
         stringstream s;  s << p_blurSigma; return s.str(); };
      void setThreshold( float t ) { p_threshold = t ; };
      void setBlurSigma( float t ) { p_blurSigma = t ; };
      void applyThreshold( int s = -1 ); 
      void applyBlur( int s = -1 );
      void applyFilter( int type, int s );
      void setDownSampleValue( int ds ) { p_downSample = ds; };

      void setupInteractor();


   private:

      enum FILTER { THRESHOLD, BLUR };

      // data extents
      int p_dims[6];
      double p_spacing[3];
      int p_minSection;
      int p_maxSection;
      int p_curSection;
      int p_numLabels;
      bool p_showRaw;
      bool p_showSeg;
      bool p_showWater;
      bool p_pick;

      Path* p_path;
      ListOfPaths p_listOfPaths;
      int p_pathInc;
      string p_pathFile; //xml file with path info
      map< int, vector< CorrespondenceMatrix* > > p_allCorrespondences;
      vector< vector< Node* > > p_allNodes;
      int p_maxnumpaths;


      string p_vtkPath;
    
      //MetaReader p_rawMetaReader;
      int p_window;
      int p_offset;

      float p_watershedLevel;
      float p_watershedThreshold;
      float p_threshold;
      float p_blurSigma;

      int p_downSample;

      // live wire stuff
      vtkImageEditor* p_liveWireEdit;
      
      // rendering objects
      vtkRenderer* p_ren;
      vtkRenderWindowInteractor* p_renWinInteractor;
      vtkInteractorStyleImage* p_interactor;

      // rendering objects
      vtkRenderer* p_renWater;
      vtkRenderWindowInteractor* p_renWinInteractorWater;
      vtkInteractorStyleImage* p_interactorWater;
      map< int, vtkActor* > p_isoActors;
      map< int, vector< vtkActor* > > p_isoSphereActors;

      // raw section rendering pipeline
      vtkPNGReader* p_rawImageReader;
      vtkImageDataStreamer* p_rawImageStreamer;
      vtkImageClip* p_rawSliceExtractor;
      vtkImageActor* p_rawSectionActor;
      vtkImageActor* p_rawSectionActor2;

      // membrane segmentation section rendering pipeline
      vtkPNGReader* p_segImageReader;
      vtkImageDataStreamer* p_segImageStreamer;
      vtkImageClip* p_segSliceExtractor;
      vtkImageActor* p_segSectionActor;
      vtkImageMapToColors* p_segColorMap;
      vtkImageMathematics* p_segImageBlend;

      // membrane segmentation editing section rendering pipeline
      vtkPNGReader* p_segEditImageReader;
      vtkImageDataStreamer* p_segEditImageStreamer;
      vtkImageClip* p_segEditSliceExtractor;
      vtkImageActor* p_segEditSectionActor;
      vtkImageMapToColors* p_segEditColorMap;
           
      vtkImageThreshold* p_thresh_sub;
      vtkImageMathematics* p_segImageBlend1;
      vtkImageThreshold* p_thresh;
      vtkImageThreshold* p_thresh_add;


      // watershed segmentation section rendering pipeline
      //vtkMetaImageReader* p_waterImageReader;
      vtkPNGReader* p_waterImageReader;
      vtkImageDataStreamer* p_waterImageStreamer;
      vtkImageClip* p_waterSliceExtractor;
      vtkImageActor* p_waterSectionActor;
      vtkImageMapToColors* p_waterColorMap;
      vtkColorTransferFunction* p_randomTransferFx;

      vtkImageData* p_pathImageData;
      int num_paths;

      // STrAND stuff
      //STrAND* p_strand;

      void setSTrANDImageData();

      void blend();
      
      void addSphere( double x, double y, double z );
      void removeSpheres();
      vector< vtkSphereSource* > p_sphereSources;
      vector< vtkPolyDataMapper* > p_sphereMappers;
      vector< vtkActor* > p_sphereActors;

      void randColor( double& r, double& g, double& b );


};

#endif // SIMPLEVIEW_H

