#include "WatershedFilter.h"

#include "itkWatershedImageFilter.h"
#include "itkCastImageFilter.h"
#include "itkImageRegionIterator.h"
#include <vector>
#include <map>

typedef itk::Image< float, 2 > ImageType;
typedef itk::Image< unsigned short, 2 > ShortImageType;
typedef itk::Image< unsigned long, 2 > LongImageType;
typedef itk::WatershedImageFilter< ImageType > WatershedFilterType;
typedef itk::CastImageFilter< LongImageType, ImageType > CastImageFilter;
typedef itk::ImageRegionIterator< ImageType > ImageRegionIterator;

//typedef itk::ConnectedComponentImageFilter< ImageType, ShortImageType >
//                      ConnectedComponentImageFilter;


WatershedFilter::WatershedFilter( ImageType::Pointer image,
                                  float level, float threshold )
{
   p_numComponents = 0;
   
   WatershedFilterType::Pointer watershed = WatershedFilterType::New();
   watershed->SetLevel( level );
   watershed->SetThreshold( threshold );
   watershed->SetInput( image );
   watershed->Update();

   CastImageFilter::Pointer castImage = CastImageFilter::New();
   castImage->SetInput( watershed->GetOutput() );
   castImage->Update();

   p_image = castImage->GetOutput();

}

WatershedFilter::~WatershedFilter()
{

}

//
// relabel the slice - itk gives us some silly labels for the slice,
// lets come up with something more reasonable
//
int WatershedFilter::relabel()
{
   std::vector< float > labels;
   // get the number of labels in the image
   ImageRegionIterator img_iter ( p_image, 
         p_image->GetLargestPossibleRegion() );
   for( img_iter.GoToBegin(); !img_iter.IsAtEnd(); ++img_iter)
   {
      float val = img_iter.Get();
      if ( find( labels.begin(), labels.end(), val ) == labels.end() )
      {
         //std::cout << "adding " << val << std::endl;
         labels.push_back( val );
      }
   }
   p_numComponents = labels.size();
   std::cout << labels.size() << " labels found." << std::endl;

   // make new labels
   std::vector< float > new_labels;
   for( unsigned int n = 1; n < labels.size(); n++ )
      new_labels.push_back( n );
   random_shuffle( new_labels.begin(), new_labels.end() );

   // map the old labels to the new labels
   std::map< float, float > label_map;
   label_map[1] = 0;
   typedef std::vector< float >::iterator iter;
   iter one_iter = std::find( labels.begin(), labels.end(), 1.0 );
   if (one_iter != labels.end() )
      labels.erase( one_iter );
   else
      std::cout << "warning: 1 was not found as a label!" << std::endl;
   for( unsigned int i = 0; i < labels.size(); i++ )
   {
      label_map[ labels[i] ] = new_labels[i];
   }

   // replace current labels with new labels
   for( img_iter.GoToBegin(); !img_iter.IsAtEnd(); ++img_iter)
   {
      img_iter.Set( label_map[ img_iter.Get() ] );
   }

   return p_numComponents;

}

ImageType::Pointer WatershedFilter::getImage()
{
   return p_image;
}
