#ifndef __itkLiveWireImageFunction_txx
#define __itkLiveWireImageFunction_txx

#include "itkLiveWireImageFunction.h"

#include "itkCastImageFilter.h"
#include "itkConstNeighborhoodIterator.h"
#include "itkNeighborhood.h"
#include "itkRescaleIntensityImageFilter.h"
#include "itkZeroCrossingBasedEdgeDetectionImageFilter.h"
#include "itkInvertIntensityImageFilter.h"
#include "itkRegionOfInterestImageFilter.h"

#include "vnl/vnl_math.h"

#include "itkImageFileWriter.h"

namespace itk
{

// Constructor
   template<class TInputImage>
      LiveWireImageFunction<TInputImage>
      ::LiveWireImageFunction()
      {
         ///*
         this->m_GradientMagnitudeWeight = 0.43;
         this->m_ZeroCrossingWeight = 0.43;
         this->m_GradientDirectionWeight = 0.14; 
         this->m_ZeroCrossingImage = NULL;
         //*/

         this->m_UseFaceConnectedness = true;
         this->m_UseImageSpacing = true;

         this->m_MaskImage = NULL;

         this->m_InsidePixelValue = NumericTraits<MaskPixelType>::One;

         this->m_AnchorSeed.Fill( 0 );

         this->m_originKeep.Fill( 0 );
      }

   // Destructor
   template<class TInputImage>
      LiveWireImageFunction<TInputImage>
      ::~LiveWireImageFunction()
      {
      }

   /*
   template <class TInputImage>
   void
   LiveWireImageFunction<TInputImage>
   ::SetInputImage( const data * ptr )
   {

   // convert data * ptr to an itk image ptr

   // SetInputImage( newptr );

   }
    */

   /**
    * Initialize by setting the input image
    */
   template <class TInputImage>
      void
      LiveWireImageFunction<TInputImage>
      ::SetInputImage( const InputImageType * ptr )
      {
         //cout << "Requested region test: " << 
         //      ptr->GetRequestedRegion()
         //      << endl ; 
         //typename InputImageType::RegionType requestedRegionKeep = ptr->GetRequestedRegion();
         m_requestedRegionKeep = ptr->GetRequestedRegion();
         // lets copy the region into a new image to process
         typedef itk::RegionOfInterestImageFilter
            <InputImageType, InputImageType> ExtractScalarRegionFilterType;
         typename ExtractScalarRegionFilterType::Pointer scalarRegionFilter =
                                           ExtractScalarRegionFilterType::New();
         scalarRegionFilter->SetRegionOfInterest( m_requestedRegionKeep );
         scalarRegionFilter->SetInput( ptr );
         scalarRegionFilter->Update();
         typename InputImageType::Pointer regionImage = scalarRegionFilter->GetOutput();
         m_originKeep = regionImage->GetOrigin();
         cout << "origin keep: " << m_originKeep << endl ;
         //cout << "anchor: " << m_AnchorSeed << endl ;
         //m_AnchorSeed[0] = m_AnchorSeed[0] - m_originKeep[0]; 
         //m_AnchorSeed[1] = m_AnchorSeed[1] - m_originKeep[1];
         cout << "regionImage: " << regionImage << endl;
         //cout << "anchor: " << m_AnchorSeed << endl ;

         //Superclass::SetInputImage( ptr );
         Superclass::SetInputImage( regionImage );

         //cout << "Requested region: " << 
         //      this->GetInputImage()->GetRequestedRegion()
         //      << endl ; 

         typedef itk::ImageFileWriter<ImageType> ImageWriterType;
         ImageWriterType::Pointer sliceWriter = ImageWriterType::New();
         sliceWriter->SetInput( regionImage );
         sliceWriter->SetFileName( "ri.mha" );
         sliceWriter->Write();

         typedef itk::InvertIntensityImageFilter< InputImageType, InputImageType > InvertImageFilter;
         typename InvertImageFilter::Pointer invertFilter = InvertImageFilter::New();
         invertFilter->SetInput( this->GetInputImage() );
         invertFilter->SetMaximum( 255 );
         invertFilter->Update();
         this->m_InvertImage = invertFilter->GetOutput();


         /**
          * Generate images for generating the weights from the input image
          */
         ///*
         //typedef CastImageFilter<InputImageType, RealImageType> CasterType;
         //typename CasterType::Pointer caster = CasterType::New();
         //caster->SetInput( this->GetInputImage() );
         //caster->Update();

         typename GradientFilterType::Pointer gradient = GradientFilterType::New();
         gradient->SetInput( this->GetInputImage() );
         gradient->SetUseImageSpacing( this->m_UseImageSpacing );
         gradient->Update();
         this->m_GradientImage = gradient->GetOutput();

         //cout << "Requested region 2: " << 
         //      this->m_Image->GetRequestedRegion()
         //      << endl ; 

         this->m_GradientMagnitudeImage = RealImageType::New();
         this->m_GradientMagnitudeImage->SetOrigin( 
               this->GetInputImage()->GetOrigin() );
         this->m_GradientMagnitudeImage->SetSpacing( 
               this->GetInputImage()->GetSpacing() );
         this->m_GradientMagnitudeImage->SetRegions( 
               this->GetInputImage()->GetRequestedRegion() );
         this->m_GradientMagnitudeImage->Allocate();
         //cout << "grad mag Requested region: " << 
         //      this->m_GradientMagnitudeImage->GetRequestedRegion()
         //      << endl ; 

         ImageRegionIterator<GradientImageType> ItG( this->m_GradientImage, 
               this->m_GradientImage->GetRequestedRegion() );
         ImageRegionIterator<RealImageType> ItM( this->m_GradientMagnitudeImage, 
               this->m_GradientMagnitudeImage->GetRequestedRegion() );
         for ( ItG.GoToBegin(), ItM.GoToBegin(); !ItG.IsAtEnd(); ++ItG, ++ItM )
         {
            ItM.Set( ( ItG.Get() ).GetNorm() );
         } 

         typedef RescaleIntensityImageFilter<RealImageType, RealImageType> RescalerType;
         typename RescalerType::Pointer rescaler = RescalerType::New();
         rescaler->SetInput( this->m_GradientMagnitudeImage );
         rescaler->SetOutputMinimum( NumericTraits<RealType>::Zero );
         rescaler->SetOutputMaximum( NumericTraits<RealType>::One );
         rescaler->Update();  
         this->m_RescaledGradientMagnitudeImage = rescaler->GetOutput();

         if ( !this->m_ZeroCrossingImage )
         {
            typedef ZeroCrossingBasedEdgeDetectionImageFilter<RealImageType, RealImageType> 
               ZeroCrossingFilterType;
            typename ZeroCrossingFilterType::Pointer 
               zeroCrossing = ZeroCrossingFilterType::New();
            zeroCrossing->SetInput( this->GetInputImage() );
            //zeroCrossing->SetInput( caster->GetOutput() );
            zeroCrossing->SetForegroundValue( NumericTraits<RealType>::One ); 
            zeroCrossing->SetBackgroundValue( NumericTraits<RealType>::Zero ); 
            zeroCrossing->Update();
            this->m_ZeroCrossingImage = zeroCrossing->GetOutput();
         }
         //*/

//         this->GeneratePathDirectionImage();
      }

   template <class TInputImage>
      void
      LiveWireImageFunction<TInputImage>
      ::GeneratePathDirectionImage()
      {
         //IndexType test_anchorSeed;
         //test_anchorSeed[0] = this->m_AnchorSeed[0] - m_originKeep[0]; 
         //test_anchorSeed[1] = this->m_AnchorSeed[1] - m_originKeep[1];
         if ( !this->IsInsideBuffer( this->m_AnchorSeed ) )
         //if ( !this->IsInsideBuffer( test_anchorSeed ) )
         {
            cout << "buffer error: " <<
               this->GetInputImage()->GetBufferedRegion() << endl ;
            cout << "anchor error: " << this->m_AnchorSeed << endl ; 
            itkExceptionMacro( "m_AnchorSeed is not inside buffer" );
         }

         /**
          * Initialize data structures
          */
         typedef Image<bool, ImageDimension> BooleanImageType;
         typename BooleanImageType::Pointer expanded = BooleanImageType::New();
         expanded->SetOrigin( this->GetInputImage()->GetOrigin() );
         expanded->SetSpacing( this->GetInputImage()->GetSpacing() );
         expanded->SetRegions( this->GetInputImage()->GetRequestedRegion() );
         expanded->Allocate();
         expanded->FillBuffer( false );
         expanded->SetPixel( this->m_AnchorSeed, true );

         this->m_PathDirectionImage = OffsetImageType::New();
         this->m_PathDirectionImage->SetOrigin( this->GetInputImage()->GetOrigin() );
         this->m_PathDirectionImage->SetSpacing( this->GetInputImage()->GetSpacing() );
         this->m_PathDirectionImage->SetRegions( this->GetInputImage()->GetRequestedRegion() );
         this->m_PathDirectionImage->Allocate();

         typename ConstNeighborhoodIterator<InputImageType>::RadiusType radius;
         radius.Fill( 1 );
         ConstNeighborhoodIterator<BooleanImageType> It( radius, expanded, 
               expanded->GetRequestedRegion() );
         unsigned int numberOfNeighbors = 1;
         for ( unsigned int d = 0; d < ImageDimension; d++ )
         {
            numberOfNeighbors *= ( 2*radius[d] + 1 ); 
         }
         Neighborhood<RealType, ImageDimension> scaleFactors;
         scaleFactors.SetRadius( radius );
         for ( unsigned int n = 0; n < numberOfNeighbors; n++ )
         {
            scaleFactors[n] = 0;
            if ( n == static_cast<unsigned int>( 0.5 * numberOfNeighbors ) )
            {
               continue; 
            } 
            typename Neighborhood<RealType, ImageDimension>::OffsetType offset 
               = scaleFactors.GetOffset( n );

            bool isFaceConnected = true;
            unsigned int sumOffset = 0; 
            for ( unsigned int d = 0; d < ImageDimension; d++ )
            {
               sumOffset += vnl_math_abs( offset[d] );
               if ( this->m_UseImageSpacing )
               {
                  scaleFactors[n] += ( static_cast<RealType>( offset[d] * offset[d] ) 
                        * this->GetInputImage()->GetSpacing()[d] 
                        * this->GetInputImage()->GetSpacing()[d] );  
               }
               else
               {
                  scaleFactors[n] += static_cast<RealType>( offset[d] * offset[d] );  
               }
               if ( sumOffset > 1 && this->m_UseFaceConnectedness )
               {
                  isFaceConnected = false;
                  break;
               }
            }      
            if ( !isFaceConnected )
            {
               scaleFactors[n] = 0;
            }
            if ( scaleFactors[n] > 0 )
            {
               scaleFactors[n] = vcl_sqrt( scaleFactors[n] );
            } 
         }   

         /**
          * Generate the PathDirectionImage with Dijkstra's algorithm
          */

         PriorityQueueElementType anchorElement( this->m_AnchorSeed, 0.0 );

         if ( this->m_MaskImage && 
               this->m_MaskImage->GetPixel( this->m_AnchorSeed ) 
               != this->m_InsidePixelValue )
         {
            itkWarningMacro( "The anchor seed is outside the user-defined mask region." ); 
            return;
         }    

         typename PriorityQueueType::Pointer Q = PriorityQueueType::New();
         Q->Initialize();
         Q->Push( anchorElement );

         while ( !Q->Empty() )
         {
            PriorityQueueElementType centerElement = Q->Peek(); 
            Q->Pop();

            expanded->SetPixel( centerElement.m_Element, true );
            It.SetLocation( centerElement.m_Element ); 
            PointType centerPoint;
            this->GetInputImage()->TransformIndexToPhysicalPoint( 
                  centerElement.m_Element, centerPoint );

            ///*
            typename GradientImageType::PixelType centerGradient 
               = this->m_GradientImage->GetPixel( centerElement.m_Element );
            RealType centerNorm 
               = this->m_GradientMagnitudeImage->GetPixel( centerElement.m_Element );
            //   */

            for ( unsigned int n = 0; n < numberOfNeighbors; n++ )
            {
               if ( scaleFactors[n] == 0 )
               {
                  continue;
               } 
               bool inBounds;
               bool isExpanded = It.GetPixel( n, inBounds );
               if ( isExpanded || !inBounds )
               {
                  continue;
               }  

               IndexType neighborIndex = It.GetIndex( n );

               if ( this->m_MaskImage && this->m_MaskImage->GetPixel( 
                        neighborIndex ) != this->m_InsidePixelValue )
               {
                  continue;
               }    

               ///*
               typename GradientImageType::PixelType neighborGradient 
                  = this->m_GradientImage->GetPixel( neighborIndex );
               RealType neighborNorm 
                  = this->m_GradientMagnitudeImage->GetPixel( neighborIndex );

               RealType fz = 0.0;
               RealType fg = 0.0;
               RealType fd = 0.0;

               if ( this->m_ZeroCrossingWeight > 0 && this->m_ZeroCrossingImage )
               {
                  fz = 1.0 - this->m_ZeroCrossingImage->GetPixel( neighborIndex );
               } 

               if ( this->m_GradientMagnitudeWeight > 0.0 )
               {
                  fg = 1.0 - 
                     this->m_RescaledGradientMagnitudeImage->GetPixel( neighborIndex );
               }
               if ( this->m_GradientDirectionWeight > 0.0 
                     && neighborNorm > 0 && centerNorm > 0 )
               {
                  PointType neighborPoint;
                  this->GetInputImage()->TransformIndexToPhysicalPoint( 
                        neighborIndex, neighborPoint );
                  typename PointType::VectorType vector 
                     = neighborPoint - centerPoint;
                  RealType vectorNorm = vector.GetNorm();            

                  RealType centerMin = vnl_math_min( centerGradient * vector, 
                        centerGradient * -vector );            
                  RealType neighborMin = vnl_math_min( neighborGradient * vector, 
                        neighborGradient * -vector );            

                  fd = 1.0 - ( vcl_acos( centerMin / ( centerNorm * vectorNorm ) ) +  
                        vcl_acos( neighborMin / ( neighborNorm * vectorNorm ) ) )
                     / vnl_math::pi;
               }

               RealType neighborCost 
                  = centerElement.m_Priority + scaleFactors[n] *
                  ( fg * this->m_GradientMagnitudeWeight
                    + fz * this->m_ZeroCrossingWeight
                    + fd * this->m_GradientDirectionWeight );
               //     */

               //RealType neighborCost 
               //   = centerElement.m_Priority + scaleFactors[n] * 
               //          this->m_InvertImage->GetPixel( neighborIndex );
               //          //this->GetInputImage()->GetPixel( neighborIndex );
               //cout << "neighborCost: " << neighborCost << endl ;

               PriorityQueueElementType neighborElement( neighborIndex, neighborCost );

               Q->Push( neighborElement ); 

               this->m_PathDirectionImage->SetPixel( neighborIndex, 
                     centerElement.m_Element - neighborElement.m_Element ); 

               PriorityQueueElementType element = Q->Peek();
               while ( expanded->GetPixel( element.m_Element ) )
               {
                  Q->Pop();
                  element = Q->Peek();
               } 
            }  
         }  
      }

   /*
   template <class TInputImage>
   int array_of_some_sort
   LiveWireImageFunction<TInputImage>
   ::EvaluateAtIndex( const int x, const int y ) const
   {
   ImageType::IndexType free;
   free[0] = x - m_originKeep; free[1] = y - m_originKeep;

   int array;
   if ( PathType::Pointer path = livewire->EvaluateAtIndex( free ) )
   {
   typedef itk::PathIterator<LabelImageType, PathType> IteratorType;
   IteratorType It( labelImage, path );
   It.GoToBegin();
   while ( !It.IsAtEnd() )
   {
   array[index]  = It.getIndex()
   ++It;
   }
   }   

   return array;
   }
   */

   template <class TInputImage>
      typename LiveWireImageFunction<TInputImage>::OutputType::Pointer 
      LiveWireImageFunction<TInputImage>
      ::EvaluateAtIndex( const IndexType &index ) const
      { 
         typename OutputType::Pointer output = OutputType::New();
         output->Initialize();

         typename InputImageType::IndexType new_index;
         new_index[0] = index[0] - m_originKeep[0];
         new_index[1] = index[1] - m_originKeep[1];
         if ( !this->IsInsideBuffer( new_index ) )
         {
            itkWarningMacro( "Requested index is not inside buffer." );
            //return NULL;
            return output;
         }

         if ( this->m_MaskImage && 
               this->m_MaskImage->GetPixel( new_index ) 
               != this->m_InsidePixelValue )
         {
            itkWarningMacro( "The index is outside the user-defined mask region." ); 
            return NULL;
         }    


         IndexType currentIndex = new_index;
         while ( currentIndex != this->m_AnchorSeed )
         {
            output->AddVertex( VertexType( currentIndex ) );
            currentIndex += this->m_PathDirectionImage->GetPixel( currentIndex );
         } 
         output->AddVertex( VertexType( currentIndex ) );

         return output;
      }

   /**
    * Standard "PrintSelf" method
    */
   template <class TInputImage>
      void
      LiveWireImageFunction<TInputImage>
      ::PrintSelf(std::ostream& os, Indent indent) const
      {
         Superclass::PrintSelf( os, indent );

         os << indent << "AnchorSeed: " 
            << this->m_AnchorSeed << std::endl;
         os << indent << "GradientMagnitudeWeight: " 
            << this->m_GradientMagnitudeWeight << std::endl;
         os << indent << "GradientDirectionWeight: " 
            << this->m_GradientDirectionWeight << std::endl;
         os << indent << "ZeroCrossingWeight: " 
            << this->m_ZeroCrossingWeight << std::endl;
         os << indent << "UseImageSpacing: " 
            << this->m_UseImageSpacing << std::endl;
         os << indent << "UseFaceConnectedness: " 
            << this->m_UseFaceConnectedness << std::endl;

         os << indent << "MaskImage"
            << this->m_MaskImage << std::endl;
         os << indent << "InsidePixelValue"
            << this->m_InsidePixelValue << std::endl;

      }

} // end namespace itk

#endif
