#include <qapplication.h>
#include "ReconstructionViewer.h"
#include "CmdLineOpts.h"

using namespace std;

int main( int argc, char** argv )
{
   
   // 
   // QT Stuff
   //
   QApplication app( argc, argv );

   //
   // process my cmd line options 
   //
   CmdLineOpts mycmd( argc, argv );

   //
   // read all the dat files
   //
   vector< string > rawFile = mycmd.getRawFileNames();
   vector< string > segFile = mycmd.getSegFileNames();
   vector< string > segEditFile = mycmd.getSegEditFileNames();
   vector< string > watFile = mycmd.getWaterFileNames();
   vector< string > pathFiles = mycmd.getPathFileName();
   vector< string > corrFiles = mycmd.getCorrelationDataFileNames();
   //cout << "params: " << rawFile << " " << segFile << " " << watFile << endl ;


   //  create our main data structure
   ReconstructionViewer mainwindow;

   // set buffering for large data
   int window = mycmd.getWindow();
   int overlap = mycmd.getOverlap();
   mainwindow.setWindow( window, overlap );

   if( watFile.size() > 0 )
      mainwindow.setFileNames(rawFile, segFile, watFile);
   else
      mainwindow.setFileNames(rawFile, segFile);

   if( segEditFile.size() > 0 )
      mainwindow.setSegEditFileNames( segEditFile );

   vector< float > spacing = mycmd.getSpacing();
   if( ( spacing.size() > 0 ) && ( spacing.size() < 4 ) )
   {
      mainwindow.setSpacing( spacing );
   }
   else
   {
      cout << "ERROR: not enough parameters for the spacing: " ;
      for( unsigned int i = 0; i < spacing.size(); i++ )
        cout << spacing[i] << " ";
      cout << endl;
      exit(1);
   }

   // do we want to rebuild the neurons in 3d?
   mainwindow.setBuildPaths( mycmd.getRebuildFlag() );

   string vtkpath = mycmd.getVTKPath();
   mainwindow.setVTKPath( vtkpath );

   // get the downsample value for rendering
   int ds = mycmd.getDownSampleValue();
   mainwindow.setDownSampleValue( ds );

   mainwindow.fileOpen();

   //
   // read in tracking info, if it exists
   //
   string mergeFile = mycmd.getMergeFileName();
   if( mergeFile.size() > 0 )
      mainwindow.openMergeRegions( mergeFile );

   vector< string > allMatrixFiles = mycmd.getMatrixFileNames();
   if( allMatrixFiles.size() > 0 )
      mainwindow.openMatrixFiles( allMatrixFiles );

   if( pathFiles.size() > 0 )
      mainwindow.openPathFile( pathFiles );

   if( corrFiles.size() > 0 )
      mainwindow.setCorrFiles( corrFiles );

   //
   //  read in the polydata, if it exists
   //
   vector< string > allPolyDataFiles = mycmd.getPolyDataFileNames();
   if( allPolyDataFiles.size() > 0 )
      mainwindow.openPolyDataFiles( allPolyDataFiles );

   //
   //  start the viewer
   //
   mainwindow.show();

   return app.exec();
}

