#include "vtkImageEditor.h"

#include "vtkExtractVOI.h"
#include "vtkImageActor.h"
#include "vtkPointData.h"
#include "vtkDataArray.h"
#include "vtkUnsignedCharArray.h"
#include "vtkImageData.h"
#include "vtkImageFlip.h"
#include "vtkImageShiftScale.h"
#include "vtkImageThreshold.h"
#include "vtkImageGaussianSmooth.h"
#include "vtkInteractorStyleImage.h"
#include "vtkInteractorStyleTrackballCamera.h"
#include "vtkRenderWindowInteractor.h"
#include <vtkMetaImageReader.h>
#include <vtkMetaImageWriter.h>
#include <vtkPolyDataWriter.h>
#include <vtkPolyDataReader.h>
#include <vtkRenderWindow.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>
#include <vtkImageMarchingCubes.h>
#include <vtkPolyDataMapper.h>
#include <vtkProperty.h>
#include <vtkImageGaussianSmooth.h>
#include <vtkImageMapToWindowLevelColors.h>
#include <vtkImageMapToColors.h>
#include <vtkCallbackCommand.h>
#include <vtkColorTransferFunction.h>
#include <vtkCommand.h>
#include "vtkWorldPointPicker.h"
#include <vtkVolume.h>
#include <vtkVolumeProperty.h>
#include <vtkVolumeRayCastMapper.h>
#include <vtkVolumeRayCastCompositeFunction.h>
#include <vtkColorTransferFunction.h>
#include <vtkPiecewiseFunction.h>
#include <vtkImageCast.h>
#include <vtkLineSource.h>
#include <vtkPoints.h>
#include <vtkCellArray.h>
#include <vtkAppendPolyData.h>
#include <vtkImageCanvasSource2D.h>
#include <vtkPNGWriter.h>
#include <vtkPNGReader.h>
#include <vtkImageCast.h>
#include <vtkImageClip.h>
#include <vtkImageBlend.h>
#include <vtkImageMathematics.h>

#include <vtkActor.h>
#include <vtkRenderer.h>
#include <vtkRenderWindow.h>
#include "vtkCylinderSource.h"
#include <vtkPolyDataMapper.h>

#include <itkImageFileWriter.h>
#include <itkImage.h>


// Constructor
vtkImageEditor::vtkImageEditor( vtkRenderer* ren  )   :
  p_minSection(0),
  p_maxSection(0),
  p_curSection(0),
  p_ren(NULL),
  p_renWinInteractor(NULL),
  p_interactor(NULL),

  p_rawImageData(NULL),
  p_rawSliceExtractor(NULL),
  p_rawShiftScale(NULL),
  p_rawSectionActor(NULL),

  p_drawImage(NULL),
  p_drawImageCanvas(NULL),
  p_canvasActor(NULL)
{
  p_ren = ren;
  
  //p_showRaw = true;
  //p_showIso = true;
  p_pick = false;
  p_liveWireOn = false;
  p_draw = true;
  p_needToSave = false;

  //
  // simple line data structure
  //
  p_points = vtkPoints::New();
  p_polyLineData = vtkPolyData::New();
  p_polyLineData->Allocate();
  p_polyLineData->SetPoints( p_points );

  vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
  mapper->SetInput( p_polyLineData );
  //mapper->SetInput( line->GetOutput() );
  p_lastLineIndex = 0;

  vtkActor* lineActor = vtkActor::New();
  lineActor->SetMapper( mapper );
  lineActor->GetProperty()->SetColor( 0, 0, 1 );
  lineActor->GetProperty()->SetLineWidth( 3 );

  p_ren->AddActor( lineActor );

  //
  // simple points data structure
  //
  /*
  p_pointsOnly = vtkPoints::New();
  p_polyPointsData = vtkPolyData::New();
  p_polyPointsData->Allocate();
  p_polyPointsData->SetPoints( p_pointsOnly );

  vtkPolyDataMapper* mapperPts = vtkPolyDataMapper::New();
  mapperPts->SetInput( p_polyPointsData );
  //mapper->SetInput( line->GetOutput() );
  p_pointlastLineIndex = 0;

  vtkActor* pointsActor = vtkActor::New();
  pointsActor->SetMapper( mapperPts );
  pointsActor->GetProperty()->SetColor( 0, 0, 1 );
  pointsActor->GetProperty()->SetLineWidth( 3 );

  p_ren->AddActor( pointsActor );
  */
  p_drawImageCanvas = vtkImageCanvasSource2D::New();
  p_drawImageCanvas->SetScalarTypeToUnsignedChar();
  p_drawImageCanvas->SetNumberOfScalarComponents(3);
  //p_drawImageCanvas->SetDrawColor(255, 255, 255, 255);
  //p_drawImageCanvas->FillBox( 0, 100, 0, 100 );

  p_canvasActor = vtkImageActor::New();
  p_drawImageCanvas->SetNumberOfScalarComponents(1);
  p_canvasActor->SetInput( p_drawImageCanvas->GetOutput() );
  //p_canvasActor->SetOpacity( .5 );
  //p_canvasActor->SetPosition( 0, 0, -10 );

  p_ren->AddActor( p_canvasActor );

  //
  //  fancy, live wire implemnetation
  //
  p_liveWirePoints = vtkPoints::New();
  p_polyLiveWireData = vtkPolyData::New();
  p_polyLiveWireData->Allocate();
  p_polyLiveWireData->SetPoints( p_liveWirePoints );
  //cout << p_polyLiveWireData->GetNumberOfPoints() << endl ;

  vtkPolyDataMapper* liveWireMapper = vtkPolyDataMapper::New();
  liveWireMapper->SetInput( p_polyLiveWireData );
  //mapper->SetInput( line->GetOutput() );
  p_liveWireLastLineIndex = 0;

  vtkActor* liveWireLineActor = vtkActor::New();
  liveWireLineActor->SetMapper( liveWireMapper );
  liveWireLineActor->GetProperty()->SetColor( 0, 1, 0 );
  liveWireLineActor->GetProperty()->SetLineWidth( 3 );

  p_ren->AddActor( liveWireLineActor );

  p_liveWireTmpPoints = vtkPoints::New();
  p_polyLiveWireTmpData = vtkPolyData::New();
  p_polyLiveWireTmpData->Allocate();
  p_polyLiveWireTmpData->SetPoints( p_liveWireTmpPoints );

  vtkPolyDataMapper* liveWireTmpMapper = vtkPolyDataMapper::New();
  liveWireTmpMapper->SetInput( p_polyLiveWireTmpData );
  //mapper->SetInput( line->GetOutput() );
  p_liveWireLastLineTmpIndex = 0;

  vtkActor* liveWireLineTmpActor = vtkActor::New();
  liveWireLineTmpActor->SetMapper( liveWireTmpMapper );
  liveWireLineTmpActor->GetProperty()->SetColor( 1, 0, 0 );
  liveWireLineTmpActor->GetProperty()->SetLineWidth( 3 );

  p_liveWireFunction = LiveWireFunction::New();

  p_ren->AddActor( liveWireLineTmpActor );

  setupInteractor();

}

vtkImageEditor::~vtkImageEditor()
{
   if( p_rawImageData )
      p_rawImageData->Delete();
   if( p_rawSectionActor )
      p_rawSectionActor->Delete();
}

vtkRenderer* vtkImageEditor::getRenderer()
{
   return p_ren;
}

void vtkImageEditor::updateRenderWindow()
{
   p_ren->GetRenderWindow()->Render();
}

/*
void vtkImageEditor::showRaw()
{
   if( p_rawSectionActor )
   {
      p_showRaw = !p_showRaw;
      p_rawSectionActor->SetVisibility( p_showRaw );
      p_ren->GetRenderWindow()->Render();
   }
}

void vtkImageEditor::showIso()
{
   if( p_isoActor )
   {
      p_showIso = !p_showIso;
      p_isoActor->SetVisibility( p_showIso );
      p_ren->GetRenderWindow()->Render();
   }
}
*/

void vtkImageEditor::setPick()
{
   p_pick = !p_pick;
}

/*
void vtkImageEditor::showSection( int s )
{
   if( ( s < p_maxSection ) && ( s >= p_minSection ) )
   {
      p_curSection = s;
      p_dims[4] = p_dims[5] = p_curSection;
      p_rawSliceExtractor->SetVOI(p_dims);
   }
   p_ren->ResetCameraClippingRange();
   p_ren->GetRenderWindow()->Render();
}
*/

void vtkImageEditor::save( vtkImageClip* clip )
{
   if( p_needToSave )
   {
      //p_drawImageCanvas->Update();

      //vtkImageData* id = vtkImageData::New();
      //id->DeepCopy( p_drawImageCanvas->GetOutput() );
      //id->Update();

      vtkImageCast* castFilter = vtkImageCast::New();
      castFilter->SetOutputScalarTypeToUnsignedChar ();
      //castFilter->SetInput( id );
      castFilter->SetInputConnection( p_drawImageCanvas->GetOutputPort());
      castFilter->Update();

      // open the original file
      cout << "reading " << p_drawImageFilename << endl;
      vtkPNGReader* reader = vtkPNGReader::New();
      reader->SetFileName( p_drawImageFilename.c_str() );

      vtkImageMathematics* segImageBlend = vtkImageMathematics::New();
      segImageBlend->SetOperationToAdd();
      segImageBlend->SetInput1( reader->GetOutput() );
      segImageBlend->SetInput2( p_drawImageCanvas->GetOutput() );

      cout << "saving " << p_drawImageFilename << endl ;
      vtkPNGWriter* Writer = vtkPNGWriter::New();
      Writer->SetFileName( p_drawImageFilename.c_str() );
      Writer->SetInputConnection( segImageBlend->GetOutputPort() );
      Writer->Write();
      cout << p_drawImageFilename << " saved. "  << endl ;

      p_needToSave = false;

      castFilter->Delete();
      Writer->Delete();
      segImageBlend->Delete();

   }
}

void vtkImageEditor::saveLineData( string filename )
{
   cout << "saving " << filename << endl;
   vtkPolyDataWriter* lineWriter = vtkPolyDataWriter::New();
   lineWriter->SetInput( p_polyLineData );
   lineWriter->SetFileName( filename.c_str() );
   lineWriter->Write();
   lineWriter->Delete();
}

void vtkImageEditor::saveLiveWireData( string filename )
{
   cout << "saving " << filename << endl;
   vtkPolyDataWriter* lineWriter = vtkPolyDataWriter::New();
   lineWriter->SetInput( p_polyLiveWireData );
   lineWriter->SetFileName( filename.c_str() );
   lineWriter->Write();
   lineWriter->Delete();
}

void vtkImageEditor::openLineData( string filename )
{
   cout << "reading " << filename << endl;
   vtkPolyDataReader* lineReader = vtkPolyDataReader::New();
   lineReader->SetFileName( filename.c_str() );
   lineReader->Update();

   // this overwrites current poly data
   //p_polyLineData->DeepCopy( lineReader->GetOutput() );

   vtkAppendPolyData* appendPolyData = vtkAppendPolyData::New();
   appendPolyData->AddInput( lineReader->GetOutput() );
   appendPolyData->AddInput( p_polyLineData );
   appendPolyData->Update();

   p_polyLineData->DeepCopy( appendPolyData->GetOutput() );
   p_polyLineData->Modified();
   lineReader->Delete();

   // update the display
   updateRenderWindow();

   // update the pt index count
   p_lastLineIndex = p_polyLineData->GetNumberOfPoints() ;

   //
   // how to read the data from polydata
   //
   /*
   double* spacing = getSpacing();  // get these from the image data
   double* origin = getOrigin();    //
   unsigned int numPoints = p_polyLineData->GetNumberOfPoints();
   cout << "numPoints: " << numPoints << endl ;
   vtkPoints* allPts = p_polyLineData->GetPoints();
   cout << "printing points" << endl ;
   for( unsigned int p = 0; p < numPoints; p++ )
   {
      double* pt = allPts->GetPoint( p );

      // these are the coordinates of the points in the image space
      // (need to cast to an int)
      cout << round( origin[0] + ( pt[0] / spacing[0] ) ) << ", " 
           << round( origin[1] + ( pt[1] / spacing[1] ) ) << ", " 
           << round( origin[2] + ( pt[2] / spacing[2] ) ) << endl ;
   }
   */
}

/*
void vtkImageEditor::openIsoData( string filename )
{
   cout << "reading " << filename << endl;
   vtkPolyDataReader* isoReader = vtkPolyDataReader::New();
   isoReader->SetFileName( filename.c_str() );
   isoReader->Update();

   p_isoPolyData = isoReader->GetOutput();

   vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
   mapper->SetInput( p_isoPolyData );
   mapper->SetScalarVisibility( false );

   double randR = (double)rand() / (double)RAND_MAX;
   double randG = (double)rand() / (double)RAND_MAX;
   double randB = (double)rand() / (double)RAND_MAX;
   p_isoActor = vtkActor::New();
   p_isoActor->SetMapper( mapper );
   p_isoActor->GetProperty()->SetColor( randR, randG, randB );
   p_isoActor->GetProperty()->SetOpacity( .5 );

   p_ren->AddActor( p_isoActor ) ;

   // update the display
   updateRenderWindow();
}

void vtkImageEditor::saveIsoData( string filename )
{
   cout << "saving " << filename << endl;
   vtkPolyDataWriter* isoWriter = vtkPolyDataWriter::New();
   isoWriter->SetInput( p_isoPolyData );
   isoWriter->SetFileName( filename.c_str() );
   isoWriter->Write();
   isoWriter->Delete();
}
*/

void vtkImageEditor::convertToImageSpace( double* coords )
{
   for( unsigned int i = 0; i < 3; i++ )
   {
      coords[i] = round( coords[i] / p_imageSpacing[i] );
   }
}

void vtkImageEditor::convertToWorldSpace( double* coords )
{
   for( unsigned int i = 0; i < 3; i++ )
   {
      coords[i] = coords[i] * p_imageSpacing[i] ;
   }
}

void vtkImageEditor::loadData( vector< int > extent, double offset )
{
   // make whole image white
   p_drawImageCanvas->SetExtent( extent[0], extent[1], extent[2], extent[3], 0, 0 ); 
   p_drawImageCanvas->SetDrawColor( 0, 0, 0 );
   p_drawImageCanvas->FillBox( extent[0], extent[1], extent[2], extent[3]);
   p_drawImageCanvas->SetDrawColor( 255, 255, 255, 255 );
   p_canvasActor->SetPosition( 0, 0, (extent[4]+.1) * offset);
   p_canvasActor->SetOpacity(.2);
   p_needToSave = false;
}

void vtkImageEditor::loadRawImage( vtkImageData* image, 
                                   vtkImageData* image_edit, 
                                   vector<int> extent )
{
   cout << "vtkImageEditor - extent: " ;
   for(int i = 0; i < 6; i++ )
      cout << extent[i] << " ";
   cout << endl ;

   p_imageSpacing = image->GetSpacing(); 

   p_liveWireBuffer = 0;

   //
   //  handle the draw image canvas
   //
   /*
   if( p_drawImage )
      p_drawImage->Delete();
   p_drawImage = vtkImageData::New();
   p_drawImage->DeepCopy( image_edit );
   p_drawImage->SetExtent( extent[0], extent[1], extent[2], extent[3], 0, 0); 
   p_drawImage->Update();
   */
   //cout << "drawimage: ";
   //p_drawImage->Print( cout );
   //cout << "0,0,0: " << p_drawImage->FindPoint( 0, 0, 0 ) << endl ;
   //p_drawImageCanvas->SetExtent( extent[0], extent[1], extent[2], extent[3], extent[4], extent[5] ); 
   p_drawImageCanvas->SetExtent( extent[0], extent[1], extent[2], extent[3], 0, 0 ); 
   p_drawImageCanvas->SetDrawColor( 255, 255, 255 );
   // make whole image white
   //p_drawImageCanvas->FillBox( extent[0], extent[1], extent[2], extent[3]);
   //p_drawImageCanvas->SetDefaultZ( extent[4] );
   //p_drawImageCanvas->DrawImage( 0, 0, p_drawImage );
   p_canvasActor->SetPosition( 0, 0, extent[4] );
   

   //
   //  set up the live wire processing
   //
   /*
   cout << "processing itk slice" << endl ;
   if( p_rawImageData )
      p_rawImageData->Delete();
   p_rawImageData = vtkImageData::New();
   p_rawImageData->DeepCopy( image );
   p_rawImageData->SetExtent( extent[0], extent[1], extent[2], extent[3], 0, 0);
   p_rawImageData->Update();

   p_itkImage = ItkVtkConvert::ConvertToItkImage( p_rawImageData );
   p_largestPossibleRegion = p_itkImage->GetLargestPossibleRegion();
   p_requestedRegion = p_itkImage->GetRequestedRegion();

   p_rawImageData->GetExtent( p_dims );
   // save the image size
   p_wholeDims[0] = p_dims[ 1 ];
   p_wholeDims[1] = p_dims[ 3 ];


   //p_liveWireFunction->SetInputImage( p_itkImage );
*/
   cout << "done" << endl ;

   /*
   vtkMetaImageReader* image_reader = vtkMetaImageReader::New();
   image_reader->SetFileName( filename.data() );
   image_reader->Update();

   //vtkImageFlip* flipy;
   //flipy = vtkImageFlip::New();
   //flipy->SetInput( image_reader->GetOutput() );
   //flipy->SetFilteredAxis(1);
   //flipy->FlipAboutOriginOn();
   //flipy->Update();
   
   p_rawImageData = vtkImageData::New();
   p_rawImageData->DeepCopy( image_reader->GetOutput() );

   p_imageSpacing = p_rawImageData->GetSpacing(); 

   image_reader->Delete();
   //flipy->Delete();

   p_rawImageData->GetExtent( p_dims );


   // keep track of the slice we're viewing
   //int num_slices_to_render = p_dims[5]-p_dims[4];
  // p_minSection = p_dims[4];
   //p_curSection = num_slices_to_render / 2; //p_minSection;
   //p_maxSection = p_dims[4] + num_slices_to_render ;
   //p_dims[4] = p_dims[5] = p_curSection;

   //  send the raw data to the livewire function
#ifdef USE_2D
      cout << "processing itk slice" << endl ;
      vtkExtractVOI* extractSlice = vtkExtractVOI::New();
      extractSlice->SetInput( p_rawImageData );
      int region[6];
      region[0] = p_dims[0];
      region[1] = p_dims[1];
      region[2] = p_dims[2];
      region[3] = p_dims[3];
      region[4] = region[5] = p_curSection;
      extractSlice->SetVOI( region );
      extractSlice->Update();
      p_itkImage = ItkVtkConvert::ConvertToItkImage( extractSlice->GetOutput() );
      //cout << "slice: " << p_itkImage << endl ;
      p_liveWireFunction->SetInputImage( p_itkImage );
#else
      cout << "processing itk volume" << endl ;
      ImageReader::Pointer itk_reader = ImageReader::New();
      itk_reader->SetFileName( filename );
      itk_reader->Update();
      p_itkImage = itk_reader->GetOutput(); 
      //   ItkVtkConvert::ConvertToItkImage3D( p_rawImageData );
      p_liveWireFunction->SetInputImage( p_itkImage );
#endif

   p_liveWireFunction->SetUseImageSpacing( true );
   cout << "done" << endl ;

   //if( calcIso ) 
   //   isoSurface();
   //rayCast();
   //createVTKPipeline();
   //test();
   */
}

/*
void vtkImageEditor::isoSurface()
{
   cout << "smoothing" << endl ;
   vtkImageGaussianSmooth* smooth = vtkImageGaussianSmooth::New();
   smooth->SetInput( p_rawImageData  );
   //smooth->SetRadiusFactors( 1.5, 1.5, 1.5 );
   //smooth->SetRadiusFactors( .5, .5, .5 );
   double rad = p_isoBlur*3;
   smooth->SetRadiusFactors( rad, rad, rad );
   smooth->SetStandardDeviations( p_isoBlur, p_isoBlur, p_isoBlur );
   smooth->Update();

   //vtkMetaImageWriter *vtkVolWriter = vtkMetaImageWriter::New();
   //vtkVolWriter->SetInput( smooth->GetOutput() );
   //vtkVolWriter->SetFileName( "smooth.mha" );
   //vtkVolWriter->Write();

   //double isoval = 100;
   cout << "marching " << p_isoVal << endl;
   //vtkImageMarchingCubes* iso = vtkImageMarchingCubes::New();
   p_iso = vtkImageMarchingCubes::New();
   p_iso->SetInput( smooth->GetOutput() );
   p_iso->SetValue( 0, p_isoVal );
   p_iso->Update();

   p_isoPolyData = p_iso->GetOutput();

   //cout << "mapping" << endl ;
   vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
   mapper->SetInput( p_isoPolyData );
   mapper->SetScalarVisibility( false );

   //
   //  write it to a vtk file for use in other applications
   //
   //vtkPolyDataWriter *vtkVolWriter = vtkPolyDataWriter::New();
   //vtkVolWriter->SetInput( mapper->GetInput() );//vtkSlice );
   //char tmp[50];
   //sprintf( tmp, "%03i", p_pathInc );
   //string incstr( tmp );
   //string wf = "vtkPolyData" + incstr + ".mha";
   //vtkVolWriter->SetFileName( wf.c_str() );
   //vtkVolWriter->Write();
   //cout << "vtk volume written to " << wf << endl ;
   //vtkVolWriter->Delete();

   double randR = (double)rand() / (double)RAND_MAX;
   double randG = (double)rand() / (double)RAND_MAX;
   double randB = (double)rand() / (double)RAND_MAX;
   //cout << "acting " << randR << " " << randG << " " << randB << endl ;
   p_isoActor = vtkActor::New();
   p_isoActor->SetMapper( mapper );
   p_isoActor->GetProperty()->SetColor( randR, randG, randB );
   p_isoActor->GetProperty()->SetOpacity( .5 );

   p_ren->AddActor( p_isoActor ) ;

   cout << "rendering " << endl ;
   p_ren->ResetCameraClippingRange();

}
*/

/*
void vtkImageEditor::rayCast()
{
   vtkImageCast* cast = vtkImageCast::New();
   cast->SetInput( p_rawImageData );
   cast->SetOutputScalarTypeToUnsignedChar();

   vtkPiecewiseFunction* opacityTF = vtkPiecewiseFunction::New();
   opacityTF->AddPoint( 0, 0 );
   opacityTF->AddPoint( 100, 0 );
   opacityTF->AddPoint( 200, 1 );

   vtkVolumeProperty* volProp = vtkVolumeProperty::New();
   //volProp->SetColor( colorTF );
   volProp->SetScalarOpacity( opacityTF );
   volProp->ShadeOn();
   volProp->SetInterpolationTypeToLinear();

   vtkVolumeRayCastCompositeFunction* comp =
      vtkVolumeRayCastCompositeFunction::New();
   vtkVolumeRayCastMapper* mapper = vtkVolumeRayCastMapper::New();
   mapper->SetVolumeRayCastFunction( comp );
   mapper->SetInputConnection( cast->GetOutputPort() );

   vtkVolume* vol = vtkVolume::New();
   vol->SetMapper( mapper );
   vol->SetProperty( volProp );

   p_ren->AddVolume( vol );

   p_ren->ResetCamera();

}
*/

void vtkImageEditor::createVTKPipeline()
{
   // Setup the raw data volume pipeline
   p_rawSliceExtractor = vtkExtractVOI::New();
   p_rawSliceExtractor->SetInput(p_rawImageData);
   p_rawSliceExtractor->SetVOI(p_dims);
   p_rawSliceExtractor->Update();

   p_rawShiftScale = vtkImageShiftScale::New();
   p_rawShiftScale->SetInputConnection( p_rawSliceExtractor->GetOutputPort() );
   p_rawShiftScale->SetOutputScalarTypeToUnsignedChar();
   p_rawShiftScale->Update();

   //vtkTransform* transform = vtkTransform::New();
   //transform->Translate(0, 0, -.5);

   p_rawSectionActor = vtkImageActor::New();
   p_rawSectionActor->SetInput( p_rawShiftScale->GetOutput() );
   p_rawSectionActor->SetOpacity( .75 );
   //double pos[3];
   //p_rawSectionActor->GetPosition( pos );
   //pos[2] -= .5;
   //p_rawSectionActor->SetPosition( pos );

   p_ren->AddActor( p_rawSectionActor );

   p_ren->ResetCamera();
}


void vtkImageEditor::setupInteractor()
{
   //vtkInteractorStyleImage *imageStyle = vtkInteractorStyleImage::New();
//   vtkInteractorStyleTrackballCamera *imageStyle = vtkInteractorStyleTrackballCamera::New();

   //
   //  charachter PRESS callback
   //
   /*
   vtkCallbackCommand *charPressCallbackCommand = vtkCallbackCommand::New();
   charPressCallbackCommand->SetCallback( charPressCallback );
   charPressCallbackCommand->SetClientData( (void*)this );

   //
   //  charachter RELEASE callback (not used right now)
   //
   vtkCallbackCommand *charReleaseCallbackCommand = vtkCallbackCommand::New();
   charReleaseCallbackCommand->SetCallback( charReleaseCallback );
   charReleaseCallbackCommand->SetClientData( (void*)this );
   */

   //
   //  LEFT mouse picking
   //
   vtkCallbackCommand *pickCallbackCommand = vtkCallbackCommand::New();
   pickCallbackCommand->SetCallback( pickCallback );
   pickCallbackCommand->SetClientData( (void*)this );

   //
   //  RIGHT mouse picking
   //
   vtkCallbackCommand *rightPressCallbackCommand = vtkCallbackCommand::New();
   rightPressCallbackCommand->SetCallback( rightPressCallback );
   rightPressCallbackCommand->SetClientData( (void*)this );

   //
   //  mouse MOVE
   //
   vtkCallbackCommand *mouseMoveCallbackCommand = vtkCallbackCommand::New();
   mouseMoveCallbackCommand->SetCallback( mouseMoveCallback );
   mouseMoveCallbackCommand->SetClientData( (void*)this );

   //vtkCallbackCommand *doubleClickCallbackCommand = vtkCallbackCommand::New();
   //doubleClickCallbackCommand->SetCallback( doubleClickCallback );
   //doubleClickCallbackCommand->SetClientData( (void*)this );

   //
   //  add them all to the renderer
   //
   //p_ren->GetRenderWindow()->GetInteractor()->AddObserver
   //   ( vtkCommand::KeyPressEvent, charPressCallbackCommand );
   //p_ren->GetRenderWindow()->GetInteractor()->AddObserver
   //   ( vtkCommand::KeyReleaseEvent, charReleaseCallbackCommand );
   p_ren->GetRenderWindow()->GetInteractor()->AddObserver
      ( vtkCommand::LeftButtonPressEvent, pickCallbackCommand );
   p_ren->GetRenderWindow()->GetInteractor()->AddObserver
      ( vtkCommand::RightButtonPressEvent, rightPressCallbackCommand );
   p_ren->GetRenderWindow()->GetInteractor()->AddObserver
      ( vtkCommand::MouseMoveEvent, mouseMoveCallbackCommand );
   //p_ren->GetRenderWindow()->GetInteractor()->AddObserver
   //   ( vtkCommand::LeftButtonReleaseEvent, doubleClickCallbackCommand );

}


void vtkImageEditor::show()
{
   // set up the main window for viewing the segmentation
   vtkRenderWindow *window = vtkRenderWindow::New();
   window->AddRenderer( p_ren );

   setupInteractor();

   window->Render();

   p_ren->GetRenderWindow()->GetInteractor()->Start();

}

void vtkImageEditor::getPickPosition( int x, int y, double* pickCoords )
{
   vtkWorldPointPicker* picker = vtkWorldPointPicker::New();
   picker->Pick( x, y, 0, p_ren );
   //double pickCoords[3];
   picker->GetPickPosition( pickCoords);
   //cout << "picked: " ;
   //for( int p = 0; p < 3; p++ )
   //   cout << pickCoords[p] << " " ;
   //cout << endl;

   //return pickCoords;
}

void vtkImageEditor::setPickPoint( int x, int y, bool end )
{
   double* pickCoords = new double[3];
   getPickPosition( x, y, pickCoords );
   drawLine( pickCoords[0], pickCoords[1], round( pickCoords[2] ), end ); 
   delete pickCoords;

   /*  not using this right now - might need it later?
   double* spacing = getSpacing();
   int* extent = getExtent()//;
   //cout << "extent: " ;
   //for( int p = 0; p < 6; p++ )
   //   cout << extent[p] << " " ;
   //cout << endl;
   //cout << "spacing: " ;
   //for( int p = 0; p < 3; p++ )
   //   cout << spacing[p] << " " ;
   //cout << endl;

   int imgCoords[3];
   imgCoords[0] = pickCoords[0] / spacing[0];
   imgCoords[1] = pickCoords[1] / spacing[1];
   imgCoords[2] = p_curSection;
   //cout << "image coords: " ;
   //for( int p = 0; p < 3; p++ )
   //   cout << imgCoords[p] << " " ;
   //cout << endl;

   float pixel = getPixelValue( imgCoords[0], imgCoords[1], imgCoords[2]);

   if ( pixel != -1 )
      cout << "pixel: " << pixel << endl;
   */
}

void vtkImageEditor::updateLiveWireBufferedRegion( itkImageType::IndexType index )
{
   itkImageType::RegionType br = p_itkImage->GetBufferedRegion();
   //cout << "current buffered region: " << br << endl ;
   //cout << "largest poss: " << p_largestPossibleRegion << endl ;
   //cout << "requested: " << p_requestedRegion << endl ;
   //cout << "image before: " << p_itkImage << endl ;
   // only returns the buffered region - dont use it.
   //itkImageType::SizeType wholeSize =
   //   p_itkImage->GetLargestPossibleRegion().GetSize();
   //cout << "whole size: " << wholeSize;
   itkImageType::SizeType bufferedSize;
   //cout << "Dimension: " << Dimension << endl ;
   for( unsigned int i = 0; i < Dimension; i++ )
   {
      // move the anchor so it's the center of the buffer and the index
      // is in the corner
      index[ i ] = index[i] - ( p_liveWireBuffer / 2 );

      // copy the buffer into the size
      bufferedSize[ i ] = p_liveWireBuffer;

      // handle index edges
      if( index[i] < 0 )
         index[i] = 0; 

      // handle size edges 
      //cout << index[i] + bufferedSize[ i ] << " ? " << p_wholeDims[ i ] << endl ;
      if ( index[i] + bufferedSize[ i ] > 
            static_cast< unsigned int>( p_wholeDims[ i ] ) )
         bufferedSize[ i ] = p_wholeDims[ i ] - index[ i ] ;
   }
   //bufferedSize[2] = 2;
   float origin[2];
   origin[0] = index[0];
   origin[1] = index[1];

   itkImageType::RegionType bufferedRegion;
   bufferedRegion.SetIndex( index );
   bufferedRegion.SetSize( bufferedSize );
   //cout << "index: " << index << endl;
   //cout << "bufferedSize: " << bufferedSize << endl ;
   //cout << "bufferedRegion: " << bufferedRegion << endl ;
   
   //typedef itk::ImageFileWriter<ImageType> ImageWriterType;
   //ImageWriterType::Pointer sliceWriter = ImageWriterType::New();
   //sliceWriter->SetInput( p_itkImage );
   //sliceWriter->SetFileName( "ConvertToITKTest4.mhd" );
   //sliceWriter->Write();
   
   //p_itkImage->SetRegions( p_largestPossibleRegion );
   //p_itkImage->SetRegions( bufferedRegion );
   //p_itkImage->SetLargestPossibleRegion( p_largestPossibleRegion );
   p_itkImage->SetRequestedRegion( bufferedRegion );
   //p_itkImage->SetBufferedRegion( bufferedRegion );
   //p_itkImage->Allocate();

   //if( p_itkImage->

   /*
   IteratorType it( p_itkImage, bufferedRegion );
   it.GoToBegin();
   while( !it.IsAtEnd() )
   {
      it.Set(it);
      ++it;
   }
   */

   //ImageWriterType::Pointer sliceWriter2 = ImageWriterType::New();
   //sliceWriter2->SetInput( p_itkImage );
   //sliceWriter2->SetFileName( "ConvertToITKTest5.mhd" );
   //sliceWriter2->Write();
   cout << "image after: " << p_itkImage << endl ;

   //cout << "new buffer: " << bufferedRegion << endl ;
}

void vtkImageEditor::setLiveWirePickPoint( int x, int y, bool end )
{
   p_liveWireOn = true;
   if( end )
      p_liveWireOn = false;

   //
   //  copy the tmp points into the main data strucutre and delete the
   //  other tmp points
   //
   cout << p_polyLiveWireData->GetNumberOfPoints() << endl ;
   cout << p_polyLiveWireTmpData->GetNumberOfPoints() << endl ;
   vtkAppendPolyData* append = vtkAppendPolyData::New();
   append->AddInput( p_polyLiveWireData );
   append->AddInput( p_polyLiveWireTmpData );
   append->Update();
   p_polyLiveWireData->DeepCopy( append->GetOutput() );
   p_polyLiveWireData->Modified();
   p_liveWireLastLineIndex = p_polyLiveWireData->GetNumberOfPoints();
   p_polyLiveWireData->BuildLinks();
   // get rid of current poly data
   p_liveWireTmpPoints->Initialize();
   p_polyLiveWireTmpData->Initialize();
   p_polyLiveWireTmpData->Allocate();
   p_polyLiveWireTmpData->SetPoints( p_liveWireTmpPoints );

   //
   //  set the next picked point
   //
   double* coords = new double[3];
   getPickPosition( x, y, coords );
   cout << "picked: ";
   for( unsigned int i = 0; i < 3; i++ )
      cout << coords[i] << ", ";
   cout << endl ;

   //convertToImageSpace( coords );
   itkImageType::IndexType anchor;
   for( unsigned int i = 0; i < Dimension; i++ )
      anchor[i] = round( coords[i] / p_imageSpacing[i] );

   cout << "anchor: " << anchor << endl ;


   //
   // update the buffered area - live wire requires a setinput again to
   // update the buffered area.
   //
   if( p_liveWireBuffer > 0 )
   {
      updateLiveWireBufferedRegion( anchor );
      cout << "updating image" << endl ;
      cout << p_itkImage->GetRequestedRegion() << endl ;
      p_liveWireFunction->SetInputImage( p_itkImage );
   }

   if( !end )
      p_liveWireFunction->SetAnchorSeed( anchor );
/*
   int ptid = p_liveWirePoints->InsertNextPoint
      ( coords[0], coords[1], coords[2] );

   cout << "ptid: " << ptid << endl ;
   delete coords;

   vtkIdType line[2];
   line[0] = ptid - 1;
   line[1] = ptid;

   if( ptid > p_liveWireLastLineIndex )
      p_polyLiveWireData->InsertNextCell( VTK_LINE, 2, line );

   if( ptid > p_lastLineIndex )
      p_polyLiveWireData->InsertNextCell( VTK_LINE, 2, line );

   if( end )
      p_liveWireLastLineIndex = ptid+1;

   p_polyLiveWireData->BuildLinks();
*/   p_polyLiveWireData->Modified();


   // update the display
   updateRenderWindow();
   if( end )
      cout << "done" << endl ;
   else
      cout << "ready" << endl ;
}

void vtkImageEditor::setAltPickPoint( int x, int y, bool draw )
{
   p_needToSave = true;
   //cout << "hovvering" << endl ;
   double* pc = new double[3];
   getPickPosition( x, y, pc );
   if (draw)
   {
      p_drawImageCanvas->SetDrawColor( 255, 255, 255, 255 );
      p_drawImageCanvas->FillBox( pc[0]-3, pc[0]+3, pc[1]-3, pc[1]+3 );
   }
   else // erase
   {
      p_drawImageCanvas->SetDrawColor( 128, 128, 128, 0 );
      p_drawImageCanvas->FillBox( pc[0]-3, pc[0]+3, pc[1]-3, pc[1]+3 );
   }
   //drawPoints( pickCoords[0], pickCoords[1], round( pickCoords[2] ), end ); 
   delete pc;
   // update the display
   updateRenderWindow();
}

void vtkImageEditor::setLiveWireHoverPoint( int x, int y )
{
   if( p_liveWireOn )
   {
      double* coords = new double[3];
      getPickPosition( x, y, coords );
      convertToImageSpace( coords );
      //cout << "Hovered: ";
      //for( unsigned int i = 0; i < 3; i++ )
      //   cout << coords[i] << ", ";
      //cout << endl ;

      itkImageType::IndexType free;
      for( unsigned int i = 0; i < Dimension; i++ )
         free[i] = coords[i];

      // get rid of current poly data
      p_liveWireTmpPoints->Initialize();
      p_polyLiveWireTmpData->Initialize();
      p_polyLiveWireTmpData->Allocate();
      p_polyLiveWireTmpData->SetPoints( p_liveWireTmpPoints );

      // what is our region of interest?
      itkImageType::RegionType bufferedRegion;
      bufferedRegion = p_itkImage->GetRequestedRegion();
      //itkImageType::PointType bufIndex = p_itkImage->GetOrigin(); 
      itkImageType::IndexType bufIndex = bufferedRegion.GetIndex(); 

      PathType::Pointer path = p_liveWireFunction->EvaluateAtIndex( free );
      const PathType::VertexListType* vertexList =  path->GetVertexList();
      PathType::VertexListType::ConstIterator vi = vertexList->Begin();
      while( vi != vertexList->End() )
      {
         PathType::VertexType V = vi.Value();
         //cout << V[0] << ", " << V[1] << endl ;
         ++vi;

         int ptid = p_liveWireTmpPoints->InsertNextPoint
            ( V[0] * p_imageSpacing[0] + bufIndex[0], 
              V[1] * p_imageSpacing[1] + bufIndex[1],
              0 );
              //V[2] * p_imageSpacing[2] );
              //coords[2]*p_imageSpacing[2] );

         vtkIdType line[2];
         line[0] = ptid - 1;
         line[1] = ptid;

         if( ptid > 0 )
            p_polyLiveWireTmpData->InsertNextCell( VTK_LINE, 2, line );

         //if( ptid > p_lastLineIndex )
         //   p_polyLiveWireData->InsertNextCell( VTK_LINE, 2, line );
         //int celldx = p_polyLineData->InsertNextCell( VTK_LINE, 2, line );

      }
      //cout << endl ;

      p_polyLiveWireTmpData->BuildLinks();
      p_polyLiveWireTmpData->Modified();

      //if( end )
      //   p_liveWireLastLineIndex = ptid+1;

      // update the display
      updateRenderWindow();

      delete coords;
   }
}

void vtkImageEditor::drawPoints( int x, int y, int z )
{
   //vtkLineSource* line = vtkLineSource::New();
   //line->SetPoint1( 0, 0, 0 );
   //line->SetPoint2( 512*6, 512*6, 23*10 );

   int ptid = p_points->InsertNextPoint( x, y, z );

   vtkIdType line[2];
   line[0] = ptid - 1;
   line[1] = ptid;

   if( ptid > p_lastLineIndex )
      p_polyLineData->InsertNextCell( VTK_LINE, 2, line );
   //int celldx = p_polyLineData->InsertNextCell( VTK_LINE, 2, line );

   // to make whole lines...  not just pairs of points like above do
   // something like this:
   /*
   p_line[100000];
   line[i] = ptid - 1;
   line[j] = ptid;
      p_polyLineData->InsertNextCell( VTK_LINE, 100000, line );
   */

   if( ptid > p_lastLineIndex )
      p_polyLineData->InsertNextCell( VTK_LINE, 2, line );
   //int celldx = p_polyLineData->InsertNextCell( VTK_LINE, 2, line );

   p_polyLineData->BuildLinks();
   p_polyLineData->Modified();

   //if( end )
   //   p_lastLineIndex = ptid+1;

}

void vtkImageEditor::drawLine( int x, int y, int z, bool end )
{
   //vtkLineSource* line = vtkLineSource::New();
   //line->SetPoint1( 0, 0, 0 );
   //line->SetPoint2( 512*6, 512*6, 23*10 );

   int ptid = p_points->InsertNextPoint( x, y, z );

   vtkIdType line[2];
   line[0] = ptid - 1;
   line[1] = ptid;

   if( ptid > p_lastLineIndex )
      p_polyLineData->InsertNextCell( VTK_LINE, 2, line );
   //int celldx = p_polyLineData->InsertNextCell( VTK_LINE, 2, line );

   // to make whole lines...  not just pairs of points like above do
   // something like this:
   /*
   p_line[100000];
   line[i] = ptid - 1;
   line[j] = ptid;
      p_polyLineData->InsertNextCell( VTK_LINE, 100000, line );
   */

   if( ptid > p_lastLineIndex )
      p_polyLineData->InsertNextCell( VTK_LINE, 2, line );
   //int celldx = p_polyLineData->InsertNextCell( VTK_LINE, 2, line );

   p_polyLineData->BuildLinks();
   p_polyLineData->Modified();

   if( end )
      p_lastLineIndex = ptid+1;

}

float vtkImageEditor::getPixelValue( int x, int y, int z )
{
   int pixel = -1;
   if( ( x >= p_dims[0] ) && ( x < p_dims[1] ) &&
       ( y >= p_dims[2] ) && ( y < p_dims[3] ) &&
       ( z == p_curSection ) )
         pixel = p_rawImageData->GetScalarComponentAsFloat( x, y, z, 0 );
   return pixel;

}

/*
void charPressCallback(vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);
   //cout << "char press: " << interactor->GetAltKey() << endl << endl ;

   vtkImageEditor* imageEditor = static_cast< vtkImageEditor* >( clientdata );
   if( interactor )
   {
      char key = interactor->GetKeyCode();
      //cout << "key: " << key << endl ;

      switch( key )
      {
         case ',':  // left arrow
            imageEditor->showPrevSection();
            break;
         case '.':  // right arrow
            imageEditor->showNextSection();
            break;
         case 'q':
            //cout << "Q" << endl ;
            //imageEditor->setPick( true );
            //cout << "press " << imageEditor->getPick() << endl ;
            break;
         case 'a':
            imageEditor->showRaw();
            break;
         case 's':
            imageEditor->showIso();
            break;
         case 'w':
            break;
         default:
            break;
      }

   } 
}
*/

/*
void charReleaseCallback(vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);
   //vtkImageEditor* imageEditor = static_cast< vtkImageEditor* >( clientdata );
   if( interactor )
   {
      char key = interactor->GetKeyCode();
      //cout << "key: " << key << endl ;
      switch( key )
      {
         case 'q':
            //cout << "Q" << endl ;
            //imageEditor->setPick( false );
            //cout << "release " << imageEditor->getPick() << endl ;
            break;
         default:
            break;
      }

   } 
}
*/

//  NOTE:  doesn't work.  the release callback doesnt work in Qt/VTK for
//  some reason.
void doubleClickCallback( vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   //vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);
   //cout << "repeat: " << interactor->GetRepeatCount() << endl ;

}

void pickCallback( vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);
   //cout << "blah" << endl ;
   
   //cout << "repeat: " << interactor->GetRepeatCount() << endl ;
   int mouseX, mouseY;
   bool end_point = false ;

   //  check for the control key
   int ctrlkey = interactor->GetControlKey();
   int shiftkey = interactor->GetShiftKey();
   //int altkey = interactor->GetShiftKey();
   //cout << "key test: " << ctrlkey << " " << shftkey << endl ;
   if ( ctrlkey )
   {
      vtkImageEditor* imageEditor = static_cast< vtkImageEditor* >( clientdata );
      if( interactor )
      {
         interactor->GetEventPosition( mouseX, mouseY );
         //cout << "mouse: " << mouseX << ", " << mouseY << endl;

         imageEditor->setPickPoint( mouseX, mouseY, end_point );
      }
   } // end if(key)
   else if ( shiftkey )  // really the shift key!
   {
      vtkImageEditor* imageEditor = static_cast< vtkImageEditor* >( clientdata );
      if( interactor )
      {
         interactor->GetEventPosition( mouseX, mouseY );
         //cout << "mouse: " << mouseX << ", " << mouseY << endl;

         imageEditor->setLiveWirePickPoint( mouseX, mouseY, end_point );
      }
   }
   /*
   else if ( altkey )  // really the alt key!
   {
      vtkImageEditor* imageEditor = static_cast< vtkImageEditor* >( clientdata );
      if( interactor )
      {
         interactor->GetEventPosition( mouseX, mouseY );
         cout << "mouse: " << mouseX << ", " << mouseY << endl;

         imageEditor->setAltPickPoint( mouseX, mouseY, end_point );
      }
   }
   */
}


void mouseMoveCallback( vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);
   vtkImageEditor* imageEditor = static_cast< vtkImageEditor* >( clientdata );
   
   int mouseX, mouseY;
   //bool end_point = false ;

   //  check for the control key
   int shiftkey = interactor->GetShiftKey();
   bool livewireon = imageEditor->getLiveWireStatus();
   //int altkey = interactor->GetAltKey();
   //cout << "moving " << altkey << endl ;

   if ( shiftkey & livewireon )
   {
      if( interactor )
      {
         interactor->GetEventPosition( mouseX, mouseY );
         //cout << "mouse: " << mouseX << ", " << mouseY << endl;

         imageEditor->setLiveWireHoverPoint( mouseX, mouseY );
      }
   }
   else if ( shiftkey )
   {
      vtkImageEditor* imageEditor = static_cast< vtkImageEditor* >( clientdata );
      if( interactor )
      {
         interactor->GetEventPosition( mouseX, mouseY );
         //cout << "mouse: " << mouseX << ", " << mouseY << endl;

         // draw pixels
         if( imageEditor->getDraw() )
            imageEditor->setAltPickPoint( mouseX, mouseY );
         // erase pixels
         else
            imageEditor->setAltPickPoint( mouseX, mouseY, false );
      }
   }
}

void rightPressCallback( vtkObject* caller, unsigned long event, void *clientdata, void *calldata)
{
   vtkRenderWindowInteractor *interactor = vtkRenderWindowInteractor::SafeDownCast(caller);

   //  check for the control key
   int ctrlkey = interactor->GetControlKey();
   int shiftkey = interactor->GetShiftKey();
   bool end_point = true;
   int mouseX, mouseY;
   if ( ctrlkey )
   {
      vtkImageEditor* imageEditor = static_cast< vtkImageEditor* >( clientdata );
      if( interactor )
      {
         interactor->GetEventPosition( mouseX, mouseY );
         //cout << "mouse: " << mouseX << ", " << mouseY << endl;

         imageEditor->setPickPoint( mouseX, mouseY, end_point );
      }
   } // end if(key)
   else if ( shiftkey )
   {
      vtkImageEditor* imageEditor = static_cast< vtkImageEditor* >( clientdata );
      if( interactor )
      {
         int mouseX, mouseY;
         interactor->GetEventPosition( mouseX, mouseY );
         //cout << "mouse: " << mouseX << ", " << mouseY << endl;

         imageEditor->setLiveWirePickPoint( mouseX, mouseY, end_point );
      }
   }

}


void vtkImageEditor::test()
{
  // Geometry
  vtkCylinderSource* source = vtkCylinderSource::New();

  // Mapper
  vtkPolyDataMapper* mapper = vtkPolyDataMapper::New();
  mapper->ImmediateModeRenderingOn();
  mapper->SetInputConnection(source->GetOutputPort());

  // Actor in scene
  vtkActor* actor = vtkActor::New();
  actor->SetMapper(mapper);

  // Add Actor to renderer
  p_ren->AddActor(actor);

  // Reset camera
  p_ren->ResetCamera();

  p_ren->GetRenderWindow()->Render();
}


