#ifndef vtkImageEditor_H
#define vtkImageEditor_H

#include <vector>
#include <string.h>

#include "vtkWindowLevelLookupTable.h"
#include "vtkImageData.h"

#include "ItkVtkConvert.h"
#include "itkImage.h"
#include "itkLiveWireImageFunction.h"
#include "itkImageFileReader.h"

// Forward class declarations
class vtkExtractVOI;
class vtkImageActor;
class vtkActor;
class vtkImageData;
class vtkImageShiftScale;
class vtkInteractorStyleImage;
class vtkMetaImageReader;
class vtkRenderWindowInteractor;
class vtkRenderer;
class vtkImageMapToWindowLevelColors;
class vtkImageMapToColors;
class vtkColorTransferFunction;
class vtkPoints;
class vtkPolyData;
class vtkImageMarchingCubes;
class vtkImageCanvasSource2D;
class vtkImageClip;

using namespace std;

//void charPressCallback(vtkObject* caller, unsigned long event, void *calldata, void *);
//void charReleaseCallback(vtkObject* caller, unsigned long event, void *calldata, void *);
void pickCallback(vtkObject* caller, unsigned long event, void *calldata, void *);
void doubleClickCallback(vtkObject* caller, unsigned long event, void *calldata, void *);
void rightPressCallback(vtkObject* caller, unsigned long event, void *calldata, void *);
void mouseMoveCallback(vtkObject* caller, unsigned long event, void *calldata, void *);

#define USE_2D
#ifdef USE_2D
const unsigned int Dimension = 2;
#else
const unsigned int Dimension = 3;
#endif 

typedef itk::Image< float, Dimension > itkImageType;
typedef itk::LiveWireImageFunction< itkImageType > LiveWireFunction;
typedef LiveWireFunction::OutputType PathType;
//typedef itk::ImageFileReader< itkImageType > ImageReader;
//typedef itk::PathIterator< itkImageType, PathType > PathIterator;

class vtkImageEditor 
{

   public:

      // Constructor/Destructor
      vtkImageEditor( vtkRenderer* ren );
      ~vtkImageEditor();

      void loadRawImage( vtkImageData* image, 
                         vtkImageData* image_edit,
                         vector<int> extent );
      //void loadRawImage( string filename, bool calcIso = true );

      void loadData( vector< int > extent, double offset );
      vtkRenderer* getRenderer();

      void updateRenderWindow();

      void show();
      //void showRaw();
      //void showIso();
      void setPick() ;
      void setPick( bool p ) { p_pick = p; } ;
      bool getPick() { return p_pick; };

      float getPixelValue( int x, int y, int z );
      double* getOrigin() { return p_rawImageData->GetOrigin(); };
      double* getSpacing() { return p_rawImageData->GetSpacing(); };
      int* getExtent() { return p_rawImageData->GetExtent(); } ;

      void setPickPoint( int x, int y, bool end = false );
      void drawLine( int x, int y, int z, bool end = false );

      void setAltPickPoint( int x, int y, bool draw = true );
      void drawPoints( int x, int y, int z );

      void setLiveWirePickPoint( int x, int y, bool end = false );
      void setLiveWireHoverPoint( int x, int y );

      void setupInteractor();

      void showSection( int i );
      void showNextSection() { showSection( p_curSection + 1 ); };
      void showPrevSection() { showSection( p_curSection - 1 ); };

      void save( vtkImageClip* img );
      void saveLineData( string filename );
      void saveLiveWireData( string filename );
      void openLineData( string filename );
      //void openIsoData( string filename );
      //void saveIsoData( string filename );

      void setFilename( string fn ) { p_drawImageFilename = fn; };

      void setIsoValue( float isoval ) { p_isoVal = isoval; };
      void setIsoBlur( float stdv ) { p_isoBlur = stdv; };

      void setLiveWireBuffer( int buffer ) { p_liveWireBuffer = buffer; };
      bool getLiveWireStatus() { return p_liveWireOn; } ;

      void setDraw( bool draw = true ) { p_draw = draw; };
      bool getDraw() { return p_draw; };



   private:


      // data extents
      int p_dims[4];  // should also be relatd to dims
      int p_wholeDims[2];  // should be set to Dimensions
      int p_minSection;
      int p_maxSection;
      int p_curSection;
      int p_numLabels;

      itkImageType::RegionType p_largestPossibleRegion;
      itkImageType::RegionType p_requestedRegion;

      bool p_showRaw;
      bool p_showIso;
      bool p_pick;
      bool p_liveWireOn;

      float p_isoVal;
      float p_isoBlur;

      int p_liveWireBuffer;
      double* p_imageSpacing;

      vtkPoints* p_points;
      vtkPolyData* p_polyLineData;
      int p_lastLineIndex;

      // live wire stuff
      vtkPoints* p_liveWirePoints;
      vtkPolyData* p_polyLiveWireData;
      int p_liveWireLastLineIndex;
      vtkPoints* p_liveWireTmpPoints;
      vtkPolyData* p_polyLiveWireTmpData;
      int p_liveWireLastLineTmpIndex;
      LiveWireFunction::Pointer p_liveWireFunction;
      itkImageType::Pointer p_itkImage;

      // rendering objects
      vtkRenderer* p_ren;
      vtkRenderWindowInteractor* p_renWinInteractor;
      vtkInteractorStyleImage* p_interactor;

      // raw section rendering pipeline
      vtkImageData* p_rawImageData;
      vtkExtractVOI* p_rawSliceExtractor;
      vtkImageShiftScale* p_rawShiftScale;
      vtkImageActor* p_rawSectionActor;
      vtkActor* p_isoActor;
      vtkImageMarchingCubes* p_iso;
      vtkPolyData* p_isoPolyData;

      vtkImageData* p_pathImageData;

      // drawing stuff
      vtkImageData* p_drawImage;
      vtkImageCanvasSource2D* p_drawImageCanvas;
      vtkImageActor* p_canvasActor;
      string p_drawImageFilename;
      bool p_needToSave;
      bool p_draw;

      void test();
      void createVTKPipeline();
      //void rayCast();
      //void isoSurface();

      void getPickPosition( int x, int y, double* pickedCoords );

      void convertToImageSpace( double* coords );
      void convertToWorldSpace( double* coords );

      void updateLiveWireBufferedRegion( itkImageType::IndexType index );

};

#endif // SIMPLEVIEW_H

